#!/usr/bin/env python

from datetime import datetime, timedelta
import errno
import os
import re
import shutil
import socket
import subprocess
import sys
import time
try:
    import httplib as httpconn
except ImportError:
    try:
        import http.client as httpconn
    except:
        pass
import mimetypes
import ssl
import platform

PUBLIC_CLOUD_CONFIG_FILE = "/etc/vmware/nsx/public-cloud-config"
NSX_PIPE_CLIENT = "/opt/vmware/nsx-agent/bin/nsx_container_pipe_client.py"
NSX_CHROOT_SUPPORT = "/opt/vmware/nsx-agent-chroot/support_save.sh"
CMD_LESS = "/usr/bin/less -X -d -F -K"
SHOW_OUTPUT = False
AWS = "aws"
AZURE = "azure"
NSXACTL_CMD = "ovs-appctl -t /var/run/vmware/nsx-agent/nsxa-ctl"
NSXCLI = "/opt/vmware/nsx-cli/bin/scripts/nsxcli"

# User defined exit codes
OUT_OF_DISK_SPACE_ERROR = 63
OUT_OF_MEMORY_ERROR = 64
INDENTATION_ERROR = 65
UNDEFINED_VARIABLE_ERROR = 66
IMPORT_ERROR = 67
KEYBOARD_INTERRUPT_ERROR = 68
INDEX_OUT_OF_RANGE_ERROR = 69
IO_ERROR = 70
SYSTEM_ERROR = 71
SYNTAX_ERROR = 72
TYPE_ERROR = 73
GENERIC_ERROR = 74

# dict to keep track of system commands per Linux flavor.
sys_command = {
                "Ubuntu": ["dpkg -l", "service --status-all 2> /dev/null"],
                "redhat": ["yum list installed",
                           "systemctl list-units --type service --all 2> /dev/null"],
                "oracle": ["yum list installed",
                           "systemctl list-units --type service --all 2> /dev/null"]
              }

def _set_show_output():
    global SHOW_OUTPUT
    SHOW_OUTPUT = True

def is_public_cloud_mode():
    if (os.path.isfile(PUBLIC_CLOUD_CONFIG_FILE)):
        return True
    return False

def is_public_cloud_chroot_mode():
    if not is_public_cloud_mode():
        return False
    with open(PUBLIC_CLOUD_CONFIG_FILE, "r") as fd:
        lines = fd.readlines()
    for line in lines:
        line = line.strip()
        if not line.startswith("container"):
            continue
        is_container = line.split('=', 1)[1]
        if is_container.lower() == "true":
            return True
    return False

class Component(object):
    @property
    def name(self):
        return self.__class__.__name__.lower()

    @property
    def commands(self):
        return []

    @property
    def extras(self):
        return []


class System(Component):
    def add_public_cloud_namespace_commands(self):
        ret = []
        command_list = ['arp -n', 'netstat -nape', 'ifconfig -a', 'route -n']
        for command in command_list:
            ret += ["ip netns exec nsx %s" % command]
            if not is_public_cloud_chroot_mode():
                ret += ["ip netns exec nsx-root %s" % command]
        return ret

    @property
    def extras(self):
        tmp = ["/etc/hosts",
               "/etc/ntp.conf",
               "/etc/resolv.conf",
               "/etc/rsyslog.conf",
               "/proc/cpuinfo",
               "/proc/diskstats",
               "/proc/loadavg",
               "/proc/meminfo",
               "/proc/net/arp",
               "/proc/net/dev",
               "/proc/net/route",
               "/proc/net/snmp",
               "/proc/stat",
               "/proc/uptime",
               "/proc/version",
               "/var/log/dmesg",
               "/var/log/syslog",
               "/opt/vmware/nsx-opsagent/bin/syslog_filter.sh"]

        ret = ["/var/log",
               "/opt/vmware/nsx-opsagent/bin",
               "/etc/default",
               "/etc/vmware",
               "/etc/init.d",
               # for ubuntu kvm only
               "/etc/network",
               # for rhel kvm only
               "/etc/sysconfig/network-scripts",
               # nestdb data files
               "/var/lib/vmware/nsx/nestdb"]

        if SHOW_OUTPUT:
            return tmp
        else:
            return ret

    @property
    def commands(self):
        tmp = ["/usr/sbin/arp -n",
               "/bin/df -lT -x securityfs",  # -x securityfs needed for non-root user
               "/bin/df -lTi -x securityfs",
               "sudo ip netns exec nsx /sbin/ifconfig -a",
               "sudo ip netns exec nsx /bin/netstat -nae",
               "/bin/ps -eo pid,uid,user,time,pcpu,pmem,rss,vsz,cmd",
               "/sbin/route -n",
               "/sbin/sysctl -A",
               "/usr/bin/top -bn 1",
               "/bin/uname -a",
               "/usr/bin/uptime"]

        ret = ["ifconfig",
               "netstat -an",
               "netstat -rn",
               "df -alT",
               "df -alTi",
               "dmesg",
               "ip -s xfrm policy",
               "ip -s xfrm state",
               "ip netns list"]

        # depending on the linux distribution, get list of packages
        if sys.version_info >= (3, 7,):
            # Starting in Python 3.7, platform.dist() is deprecated,
            # and so use distro.linux_distribution() instead
            import distro
            linux_dist = distro.linux_distribution()[0]
        else:
            linux_dist = platform.dist()[0]
        if linux_dist in sys_command:
            ret.extend(sys_command[linux_dist])

        # if executing in public cloud env
        if is_public_cloud_mode():
            ret += self.add_public_cloud_namespace_commands()

        if SHOW_OUTPUT:
            return tmp
        else:
            return ret


class HyperBus(Component):
    @property
    def commands(self):
        ret = [NSXCLI + ' --cmd "get hyperbus cif-table"',
               NSXCLI + ' --cmd "get hyperbus vif-lip-table"',
               NSXCLI + ' --cmd "get hyperbus vif-table"',
               NSXCLI + ' --cmd "get hyperbus connection info"']
        ret += ["ip netns exec hyperbus ifconfig -a"]
        return ret


class NsxAgent(Component):

    @property
    def name(self):
        return "domain"

    # commands that have to be run per vni
    def vni_commands_helper(self, vni_list):
        ret = []
        command_list = ['vtep_list', 'arp-table', 'mac-vtep-label', 'replication-mode']
        for vni in vni_list:
          for command in command_list:
            ret += [NSXACTL_CMD + " vni/%s %s" % (command, vni)]
          # get vtep list and ip, mac of the vteps
          vtep_output = ""
          try:
              vtep_output = (
              subprocess.check_output(NSXACTL_CMD + " vni/vtep_list %s" %
                                      vni, shell=True))
              vtep_output = vtep_output.decode('utf-8')
              vtep_list = []
              for vtep in vtep_output.split("\n")[2:]:
                if vtep:
                    vtep_list.append(vtep.split()[0])
              # get vtep details
              cmd_list = ['ip', 'pmac']
              for vtep  in vtep_list:
                 for command in cmd_list:
                    ret += [NSXACTL_CMD + " vtep/%s  %s" % (command, vtep)]
          except subprocess.CalledProcessError as e:
            print(("Error while running ovs-appctl cmd : " + str(e.returncode)
                  + ", output:\n" + e.output))

        return ret

    def vdr_commands_helper(self):
        ret = []
        vdr_output = ""
        try:
            vdr_output = (
                subprocess.check_output(NSXACTL_CMD + " vdr/list", shell=True))
            vdr_output = vdr_output.decode('utf-8')
        except subprocess.CalledProcessError as e:
            print(("Error while running ovs-appctl cmd : " + str(e.returncode)
                  + ", output:\n" + e.output))
        vdr_list = []
        command_list = ['type', 'route', 'lif', 'dhcp-relay']
        vdr_trimmed_list = vdr_output.split("\n")[1:][:-1]
        for vdr in vdr_trimmed_list:
            if vdr:
                vdr_list.append(vdr.split()[0])
        for vdr in vdr_list:
            for command in command_list:
                ret += [NSXACTL_CMD + " vdr/%s %s" % (command, vdr)]
        return ret

    def lswitch_commands_helper(self):
        ret = []
        lswitch_output = ""
        try:
            lswitch_output = (
                subprocess.check_output(NSXACTL_CMD + " lswitch/list",
                                        shell=True))
            lswitch_output = lswitch_output.decode('utf-8')
        except subprocess.CalledProcessError as e:
            print(("Error while running ovs-appctl cmd : " + str(e.returncode)
                  + ", output:\n" + e.output))
        lswitch_list = []
        lswitch_trimmed_list = lswitch_output.split("\n")[1:][:-1]
        for lswitch in lswitch_trimmed_list:
            if lswitch:
                lswitch_list.append(lswitch.split()[0])
        ret = []
        #for the given logical switch, print out all commands
        command_list = ['local-mac-table', 'mac-table', 'vif-contexts', 'vtep-table']
        for lswitch in lswitch_list:
            for command in command_list:
                ret += [NSXACTL_CMD + " lswitch/%s %s" % (command, lswitch)]
        return ret

    def lport_commands_helper(self):
        ret = []
        lport_list = []
        lport_output = ""
        try:
            lport_output = (
                 subprocess.check_output(NSXACTL_CMD + " lport/list", shell=True))
            lport_output = lport_output.decode('utf-8')
        except subprocess.CalledProcessError as e:
            print(("Error while running ovs-appctl cmd : " + str(e.returncode)
                  + ", output:\n" + e.output))
        lport_trimmed_list = lport_output.split("\n")[1:][:-1]
        for lport in lport_trimmed_list:
            if lport:
                lport_list.append(lport.split()[0])
        #for the given logical switch, print out all commands
        for lport in lport_list:
            ret += [NSXACTL_CMD + " lport/snoop-mode %s" % lport]
        return ret

    def add_public_cloud_namespace_commands(self):
        ret = []
        ret += [NSXACTL_CMD + " gateway/show"]
        ret += [NSXACTL_CMD + " gateway-cert/show all"]
        return ret

    @property
    def commands(self):
        bridges = []
        try:
            bridges = (
               subprocess.check_output(
                    "ovs-vsctl --columns=name --format=csv list bridge",
                    shell=True).decode('utf-8').strip().split("\n")[1:])
        except subprocess.CalledProcessError as e:
           print("Failed to list OVS bridges")

        ret = ["ovs-vsctl show",
               "ovs-vsctl list interface",
               "ovs-vsctl list br",
               "ovs-vsctl list port",
               "ovs-vsctl list ipfix",
               "ovs-vsctl list flow_sample_collector_set",
               NSXACTL_CMD + " nlog/dump",
               "ovs-dpctl show",
               "ovs-dpctl dump-flows -m",
               "[ -f \"/tmp/mockvms_daemon.log\" ]  && cat \"/tmp/mockvms_daemon.log\" ",
               "[ -f \"/tmp/tcpdump_ovs\" ]  && cat \"/tmp/tcpdump_ovs\" ",
               NSXACTL_CMD + " flow/list duplicate",
               NSXACTL_CMD + " flow/list missing",
               NSXACTL_CMD + " flow/list extra",
               NSXACTL_CMD + " neigh/show",
               NSXACTL_CMD + " dfw/vif",
               NSXACTL_CMD + " dfw/rules",
               NSXACTL_CMD + " dfw/ovs-rules",
               NSXACTL_CMD + " dfw/exclude",
               NSXACTL_CMD + " dfw/addrsets",
               NSXACTL_CMD + " dfw/realization",
               NSXACTL_CMD + " monitor-pair/show",
               "tail -n +1 /proc/sys/net/netfilter/nf_conntrack*",
               "tail -n +1 /root/colibri_support* 2> /dev/null"
              ]
        ret += ["ovs-ofctl show %s" % br for br in bridges]
        ret += ["ovs-ofctl dump-flows %s | ovs-decode-note" % br for br in bridges]
        ret += ["ovs-ofctl dump-ipfix-flow %s" % br for br in bridges]
        ret += [NSXACTL_CMD + " vdr/list"]
        ret += [NSXACTL_CMD + " vni/list"]
        ret += [NSXACTL_CMD + " vlan/list"]
        ret += [NSXACTL_CMD + " lswitch/list"]
        ret += [NSXACTL_CMD + " lswitch/ipfix"]
        ret += [NSXACTL_CMD + " lswitch/routing-domain"]
        ret += [NSXACTL_CMD + " lport/list"]
        ret += [NSXACTL_CMD + " lport/ipfix"]
        ret += [NSXACTL_CMD + " lport/mac-learning"]
        ret += [NSXACTL_CMD + " lport/vif"]
        ret += [NSXACTL_CMD + " lport/snoop-mode"]
        ret += [NSXACTL_CMD + " lport/white-list"]
        ret += [NSXACTL_CMD + " mirror-session/list"]
        ret += [NSXCLI + ' --cmd "get version"']
        ret += [NSXCLI + ' --cmd "get manager"']
        ret += [NSXCLI + ' --cmd "get controllers"']
        ret += [NSXCLI + ' --cmd "get service nsx-proxy agent monitor status"']
        ret += [NSXCLI + ' --cmd "get mirror-sessions"']
        ret += [NSXCLI + ' --cmd "verify lcp-kvm state"']
        ret += [NSXCLI + ' --cmd "verify firewall consistency"']

        # vdr info
        ret += self.vdr_commands_helper()

        # get vni/vlan list and get data for each of them.
        vni_list=[]
        vni_output = ""
        try:
            vni_output = (
                subprocess.check_output(NSXACTL_CMD + " vni/list", shell=True))
            vni_output = vni_output.decode('utf-8')
        except subprocess.CalledProcessError as e:
            print(("Error while running ovs-appctl cmd : " + str(e.returncode)
                  + ", output:\n" + e.output))
        vni_trimmed_list = vni_output.split("\n")[1:][:-1] #trim out header line and last empty line
        for vni in vni_trimmed_list:
            if vni:
                vni_list.append(vni.split()[0])

        # for every vni get all information:
        if vni_list:
            ret += self.vni_commands_helper(vni_list)

        # get LS and LP information
        ret += self.lswitch_commands_helper()

        # get vtep-table per routing domain
        rd_list = []
        routing_domain_output = ""
        try:
            routing_domain_output = (
                    subprocess.check_output(NSXACTL_CMD +
                    " lswitch/routing-domain", shell=True))
            routing_domain_output = routing_domain_output.decode('utf-8')
        except subprocess.CalledProcessError as e:
            print(("Error while running ovs-appctl cmd : " + str(e.returncode)
                  + ", output:\n" + e.output))

        rd_trimmed_list = routing_domain_output.split("\n")[1:][:-1]
        for rd in rd_trimmed_list:
           if rd:
              rd_list.append(rd.split()[1]) # routing domain is the second field in the output.
        for rd in rd_list:
           ret += [NSXACTL_CMD + " routing-domain/vtep-table %s" % rd]

        ret += [NSXACTL_CMD + " ip-discovery/show-dhcp-pending-requests"]
        ret += [NSXACTL_CMD + " ip-discovery/show-entries"]
        ret += [NSXACTL_CMD + " dr-vtep/list"]
        ret += [NSXACTL_CMD + " full-sync/status"]

        # if executing in public cloud env
        if is_public_cloud_mode():
            ret += self.add_public_cloud_namespace_commands()
            ret += [NSXACTL_CMD + " pbr/rules"]
            ret += [NSXACTL_CMD + " pbr/addrsets"]

        if SHOW_OUTPUT:
            return []
        else:
            return ret


class Exporter(Component):

    @property
    def commands(self):
        cmds = [
           "get/filters",
           "get/rulestats",
           "get/profiles",
           "get/container",
           "get/vifmap",
           "get/ipfixstats",
           "get/allalarmcfg",
           "get/intelcfg",
           "get/intelkafkacfg",
           "get/intelstats",
           "get/intelstats ack"
        ]

        if is_public_cloud_mode():
            cmds += ["pbr/stats"]

        ret = []
        ret += ["ovs-appctl -t /var/run/vmware/exporter/common-exporter-cli %s" % cmd for cmd in cmds]
        return ret


class OpsAgent(Component):

    @property
    def commands(self):
        ret = []
        ret += ["nsx-appctl -t /var/run/vmware/nsx-opsagent/opsagent-cli get/mpaclient"]
        ret += ["nsx-appctl -t /var/run/vmware/nsx-opsagent/opsagent-cli lldpd/status"]
        ret += ["nsx-appctl -t /var/run/vmware/nsx-opsagent/opsagent-cli lldpd/show"]
        return ret

class NsxProxy(Component):

    @property
    def commands(self):
        ret = ["/opt/vmware/nsx-proxy/bin/cli.sh --cmd 'RpcInfo:Connections'"]
        return ret

class SupportSave(object):

    def __init__(self, tarfile, age_limit, content_filter,
                 pcg_url, pcg_upload_token):
        try:
            host_ip = socket.gethostbyname(socket.gethostname()).replace(".", "-")
        except:
            host_ip = "unknown"
        self._dir = "/tmp/" + host_ip + "-" + datetime.now().isoformat() + "/"
        self._dir = self._dir.replace(":", "-")
        self._tarfile = tarfile
        self._age_limit = age_limit
        self._content_filter = content_filter
        self._pcg_url = pcg_url;
        self._pcg_upload_token = pcg_upload_token;
        if SHOW_OUTPUT is False:
            print(("executing support save, tarfile={}, age_limit={}, filter={}".format(
                   self._tarfile, self._age_limit, self._content_filter)))
        try:
            if os.path.exists(self._dir):
                shutil.rmtree(self._dir)
        except Exception as e:
            self.handle_shutil_error(e)

        if SHOW_OUTPUT is False:
            os.mkdir(self._dir)
        if self._content_filter:
            self._core_loc = None
        else:
            p = subprocess.Popen("cat /proc/sys/kernel/core_pattern",
                                 stdout=subprocess.PIPE, shell=True)
            self._core_loc, _ = p.communicate()
            self._core_loc = self._core_loc.decode('utf-8')
            if self._core_loc[0] == "/":
                delm = self._core_loc.rfind("/")
                self._core_loc = str(self._core_loc[:delm+1])
            elif "apport" in self._core_loc:
                # If core_pattern is configured to pipe to apport,
                # apport saves the core dump in /var/crash
                self._core_loc = "/var/crash/"
            else:
                print(("Could not find core location from %s" % self._core_loc))
                self._core_loc = None
        self._comp = []

    def _filter_content(self, dir_name, file_names):
        content_filter_patterns = ["^/var/log/audit/.*"]

        # filter content based age_limit and filter requirement
        filtered_files = set()
        # insert '/' at end of dir name
        dir_name = os.path.join(dir_name, "")
        if self._content_filter:
            for ignored in content_filter_patterns:
                if re.match(ignored, dir_name) is not None:
                    return file_names
                    for f in file_names:
                        full_name = os.path.join(dir_name, f)
                        if re.match(ignored, full_name) is not None:
                            filtered_files.add(f)

        # log age limit to be applied to log files only
        age_filter_patterns = ["^/var/log/.*"]
        if self._age_limit > 0:
            for f in file_names:
                full_name = os.path.join(dir_name, f)
                if not os.path.exists(full_name) or os.path.isdir(full_name):
                    continue
                for pattern in age_filter_patterns:
                    if re.match(pattern, full_name) is not None:
                        if (timedelta(seconds = time.time() -
                                     os.path.getmtime(full_name)).days >=
                            self._age_limit):
                            filtered_files.add(f)

        # public cloud gateway currently collects the recent ovs and nsx-agent logs
        if self._pcg_url is not None:
            pcg_filter_patterns = ["^/var/log/.*"]
            for f in file_names:
                full_name = os.path.join(dir_name, f)
                if not os.path.isfile(full_name):
                    continue
                for pattern in pcg_filter_patterns:
                    if re.match(pattern, full_name) is not None:
                        if re.match("(^/var/log/vmware/nsx-agent/Openflow_controller_cpp.log.INFO)"
                                    "|(^/var/log/vmware/nsx-agent/Openflow_controller_cpp.log.WARNING)"
                                    "|(^/var/log/vmware/nsx-agent/Openflow_controller_cpp.log.ERROR)"
                                    "|(^/var/log/vmware/nsx-agent/nsx-bundle-upgrade.log)"
                                    "|(^/var/log/vmware/nsx-agent/nsx-bundle-install.log)"
                                    "|(^/var/log/openvswitch/.*)", full_name) is None:
                            filtered_files.add(f)
        return [f for f in filtered_files]

    @property
    def components(self):
        return self._comp

    @components.setter
    def components(self, comp):
        self._comp = comp

    def upload_to_gateway(self, files):
        # Retry upload up to 3 times
        for i in range(0,3):
            conn = None
            try:
                try:
                    from urlparse import urlparse
                except ImportError:
                    try:
                        from urllib.parse import urlparse
                    except:
                        pass

                url_parse_result = urlparse(self._pcg_url)
                host = url_parse_result.hostname
                port = url_parse_result.port
                path = url_parse_result.path

                # Validate Host Name
                pcg_cert = "/etc/vmware/nsx/pcg-cert"
                with open(pcg_cert, "r") as f:
                    if str(host) not in f.read():
                        return {"success": False, "response": "Unable to validate pcg"}

                # Create the body and header
                boundary = '----------NSX_File_Boundary_%s' % (datetime.now().isoformat())
                L = []
                for (key, filename, value) in files:
                    L.append('--' + boundary)
                    L.append('Content-Disposition: form-data; name="%s"; filename="%s"'
                             % (key, filename))
                    L.append('Content-Type: %s' % (mimetypes.guess_type(filename)[0]
                                                   or 'application/octet-stream'))
                    L.append('')
                    # for binary data decoding byte string to unicode
                    if sys.version_info >= (3,):
                        L.append(value.decode('latin-1'))
                    else:
                        L.append(value)
                L.append('--' + boundary + '--')
                L.append('')
                body = '\r\n'.join((str(entry) for entry in L))
                content_type = 'multipart/form-data; boundary=%s' % boundary

                if sys.version_info >= (2, 7, 9):
                    ctx = ssl._create_unverified_context()
                    if url_parse_result.scheme == "https":
                        conn = httpconn.HTTPSConnection(host, port, context=ctx,
                                                       timeout=30)
                    else:
                        conn = httpconn.HTTPConnection(host, port, timeout=30)
                else:
                    if url_parse_result.scheme == "https":
                        conn = httpconn.HTTPSConnection(host, port, timeout=30)
                    else:
                        conn = httpconn.HTTPConnection(host, port, timeout=30)
                headers = {
                    "Content-Length": "%s" % str(len(body)),
                    "Content-Type": content_type
                }
                if self._pcg_upload_token is not None:
                    headers["nsx-pcg-upload-token"] = self._pcg_upload_token

                conn.connect()
                conn.request("POST", path, body, headers)
                resp = conn.getresponse()
                response = resp.read()
                if conn:
                    conn.close()
                return {"success": True, "response": response}
            except Exception as err:
                if conn:
                    conn.close()
                print (err)
                continue
        return {"success": False, "response": "Unable to upload"}

    def get_cloud_type(self):
        _, version, _ = self.invoke_command("cat /sys/class/dmi/id/bios_version".split())
        _, bios_vendor, _ = self.invoke_command("cat /sys/class/dmi/id/bios_vendor".split())
        _, vendor, _ = self.invoke_command("cat /sys/class/dmi/id/sys_vendor".split())
        cloud = "unknown"
        if "amazon" in version.lower():
            cloud = AWS
        elif "amazon" in bios_vendor.lower():
            cloud = AWS
        elif "microsoft" in vendor.lower():
            cloud = AZURE
        return cloud

    def get_vm_instanceid(self, default_name):
        conn = httpconn.HTTPConnection("169.254.169.254")
        cloud = self.get_cloud_type()
        if cloud == AWS:
            conn.request("GET", "/latest/meta-data/instance-id")
        elif cloud == AZURE:
            conn.request("GET", "/metadata/instance/compute/vmId?api-version=2017-08-01&format=text", \
                         headers={"Metadata":"True"})
        resp = conn.getresponse()
        if resp.status is 200:
            instance_id = resp.read()
            bundle_name = instance_id.decode('utf-8') + "-" + datetime.now().isoformat() + ".tar.gz"
            bundle_name = bundle_name.replace(":", "-")
            return bundle_name
        else:
            return default_name + ".tar.gz"

    def collect(self):
        commands = []
        extras = []
        cores = []
        for comp in self._comp:
            commands += comp.commands
            extras += comp.extras
            cores += [comp.name]
        set(commands)
        set(extras)
        set(cores)

        # Support bundle related paths
        full_name = self._dir[:-1]
        dirname = os.path.dirname(full_name)
        src = os.path.basename(full_name)
        if self._tarfile is not None:
            dst = self._tarfile
        elif self._pcg_url is not None:
            dst = self.get_vm_instanceid(src)
        else:
            dst = src + ".tar.gz"

        # processing cores
        core_dir = self._dir + "cores/"
        os.mkdir(core_dir)
        if self._core_loc:
            core_dirs =  [p for p in self._core_loc if os.path.exists(p)]
            for core in core_dirs:
                if "core" in core or "crash" in core:
                    try:
                        shutil.copy(self._core_loc + core,
                                    core_dir)
                    except Exception as e:
                        self.handle_shutil_error(e)
        # processing cmds
        cmd_dir = self._dir + "commands/"
        os.mkdir(cmd_dir)
        for cmd in commands:
            fn = re.sub("[^\w-]", "_", cmd) + ".txt"
            with open(cmd_dir+fn, "w+") as f:
                f.write(cmd)
                subprocess.call(cmd, stdout=f, shell=True)

        # process public cloud specific additional bundle
        if is_public_cloud_chroot_mode():
            print("Running within chroot, collecting root support bundle")
            try:
                root_tar = os.path.join(dirname, src, "nsx-chroot-%s" % dst)
                cmd = ("%s \"%s --tar-file %s --from-chroot\""
                       % (NSX_PIPE_CLIENT, NSX_CHROOT_SUPPORT, root_tar))
                subprocess.check_call(cmd, shell=True)
            except Exception as e:
                print("Failed to initiate root support bundle collection")

        # process extras
        for e in extras:
            try:
                if os.path.isdir(e):
                    shutil.copytree(e, self._dir+e, ignore=self._filter_content)
                elif os.path.isfile(e) and len(self._filter_content("", [e])) == 0:
                    shutil.copy(e, self._dir+e)
            except Exception as e:
                self.handle_shutil_error(e)

        # create a OVS support bundle
        ovs_dir = self._dir + "ovs/"
        os.mkdir(ovs_dir)
        ovs_tgz_file = "%s/ovs-bugreport-%s.tgz" \
                    % (ovs_dir, time.strftime("%Y%m%d%H%M%S"))
        log_days = self._age_limit
        if not log_days:
            log_days = 20
        ovs_cmd = ["ovs-bugtool", "-y", "-s", "--ovs",
                   "--log-days=%s" % log_days,
                   "--outfile=%s" % ovs_tgz_file]
        child = subprocess.Popen(ovs_cmd, stdout=subprocess.PIPE)
        child.communicate()
        if child.returncode:
            print("failed to generate OVS bugtool bundle")

        # Create support bundle tar
        subprocess.call(
            "cd %s; tar czf %s %s" % (dirname, dst, src), shell=True)
        try:
            shutil.rmtree(self._dir)
        except Exception as e:
            self.handle_shutil_error(e)
        print(("support save in %s/%s" % (dirname, dst)))
        if self._pcg_url is not None:
            file_path = '%s/%s' % (dirname, dst)
            files = [("file", file_path, open(file_path, "rb").read())]
            upload_result = self.upload_to_gateway(files)
            if upload_result.get('success'):
                print(("Uploaded %s to %s" % (dst, self._pcg_url)))
            else:
                print((upload_result.get('response')))

            # After upload remove the bundle-file from VM
            try:
                os.remove(file_path)
            except Exception as e:
                self.handle_shutil_error(e)

    def handle_shutil_error(self, error):
        print (error)
        print(("Warning:- A file I/O error has occurred. ",
               "The directory structure ",
               "may be unstable currently and ",
               "it is advised to re-run the script in some time."))

    def get_text_pager(self):
        """
        Return pager object that pages (forwards/backwards) user supplied text
        specified via the pager.write() method.
        """
        os.putenv("LESSSECURE", "1")
        # Using os.popen here instead of subprocess.Popen because the
        # latter disable text highlighting when the user performs a
        # a search.
        try:
            return os.popen(CMD_LESS, "w")
        except Exception:
            print(("Error invoking: %s" % (CMD_LESS)))
        return None

    def invoke_command(self, cmd_list):
        """Invoke a command.

        Args:
            cmd_list: List containing command to run and arguments to pass,
                  (e.g ["ps", "aux"]).
        Retrun:
            returncode, output, error
        """

        out_stream = subprocess.PIPE
        err_stream = subprocess.PIPE

        p = subprocess.Popen(cmd_list,
                            stdout=out_stream,
                            stderr=err_stream)
        out, err = p.communicate()
        if sys.version_info >= (3,):
            # In Python3, communicate() returns byte strings, so convert to string
            if out is not None:
                out = out.decode("utf8")
            if err is not None:
                err = err.decode("utf8")
        rc = p.returncode
        return rc, out, err

    def display(self):
        """
        Displays the support-bundle content to stdout
        """
        commands = []
        extras = []
        for comp in self._comp:
            commands += comp.commands
            extras += comp.extras
        set(commands)
        set(extras)
        pager = self.get_text_pager()
        if pager is None:
            print ("Error, insufficient system resources to view support bundle")
        try:
            for cmd in commands:
                try:
                    rc, out, err = self.invoke_command(cmd.split())
                    if rc != 0:
                        print(("Error running command '%s', rc: %s, err: %s" \
                               % (cmd, rc, err)))
                        continue
                    pager.write("%s\n%s\n%s\n" % ("-" * 80, cmd, "-" * 80))
                    #out = out.decode("utf8")
                    pager.write(out)
                except (IOError, KeyboardInterrupt):
                    pass           # Show file output
            for fi in extras:
                try:
                    with open(fi) as fd:
                        pager.write("%s\n%s\n%s\n" % ("-" * 80, fi, "-" * 80))
                        pager.write(fd.read())
                except Exception:
                    pass  # Ignore errors so we can process next command
        except (IOError, KeyboardInterrupt):
            pass
        try:
            pager.close()
        except IOError:
            pass


if __name__ == "__main__":
    errorcode = 0
    try:
        import argparse
        parser = argparse.ArgumentParser(prog="__file__")
        parser.add_argument("-t", "--tar_file", nargs="?", default=None,
                            help="target tar file name")
        parser.add_argument("-l", "--age_limit", nargs="?", default=0,
                            help="age limit (in days) of files to be collected")
        parser.add_argument("-f", "--content_filter", nargs="?", default="ALL",
                            help="content filter to exclude sensitive files if set to 'DEFAULT' ")
        parser.add_argument("-pcgurl", "--pcg_url", nargs="?", default=None,
                            help="public cloud gateway url")
        parser.add_argument("-pt", "--pcg_upload_token", nargs="?", default=None,
                            help="public cloud gateway upload token")
        parser.add_argument("-o", "--show_output", type=bool, nargs="?", default=False,
                            help="shows output to console")

        args = parser.parse_args()
        tarfile = args.tar_file
        age_limit = int(args.age_limit)
        content_filter = (args.content_filter == "DEFAULT")
        pcg_url = args.pcg_url
        pcg_upload_token = args.pcg_upload_token
        if args.show_output is True:
            _set_show_output()

        if tarfile is not None:
            tarfile = os.path.abspath(tarfile)
            if not os.path.isdir(os.path.dirname(tarfile)):
                print(("Creating parent directory to tarfile {} ".format(tarfile)))
                os.mkdir(os.path.dirname(tarfile))
        save = SupportSave(tarfile=tarfile, age_limit=age_limit, content_filter=content_filter,
                           pcg_url=pcg_url, pcg_upload_token=pcg_upload_token)
        # Collect support-bundle logs from system, l3d and nsx-agent when pcg_url
        # is specified or if running on public cloud VMs.
        if pcg_url is not None or is_public_cloud_mode():
            save.components += [System(), NsxAgent(), Exporter(), NsxProxy()]
        else:
            save.components += [System(), NsxAgent(), HyperBus(), Exporter(), OpsAgent(), NsxProxy()]

        if SHOW_OUTPUT:
            # Display the support-bundle output to the stdout
            save.display()
        else:
            # Collect the support-bundle output and tar it in a file
            save.collect()
    except MemoryError as e:
        errorcode = OUT_OF_MEMORY_ERROR
        print (e)
    except IndentationError as e:
        print (e)
        errorcode = INDENTATION_ERROR
    except NameError as e:
        print (e)
        errorcode = UNDEFINED_VARIABLE_ERROR
    except ImportError as e:
        errorcode = IMPORT_ERROR
        print (e)
    except KeyboardInterrupt as e:
        errorcode = KEYBOARD_INTERRUPT_ERROR
        print (e)
    except IndexError as e:
        errorcode = INDEX_OUT_OF_RANGE_ERROR
        print (e)
    except IOError as e:
        print (e)
        errorcode = IO_ERROR
    except OSError as e:
        print (e)
        if e.errno == errno.ENOSPC:
            errorcode = OUT_OF_DISK_SPACE_ERROR
        else:
            errorcode = SYSTEM_ERROR
    except SyntaxError as e:
        print (e)
        errorcode = SYNTAX_ERROR
    except TypeError as e:
        print (e)
        errorcode = TYPE_ERROR
    except Exception as e:
        print (e)
        errorcode = GENERIC_ERROR
    sys.exit(errorcode)
