# **************************************************************************
# Copyright 2020 VMware, Inc.  All rights reserved. VMware Confidential.
# **************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **************************************************************************
import nsx_logging
import re
try:
    import simplejson
except ImportError:
    import json as simplejson
import sys
import os
import uuid

from cli.utils import rest_utils
from cli.utils.command_utils import invoke_command
from cli.utils.constants import NODE_TYPE_EDGE
from cli.utils.constants import NODE_TYPE_AUTONOMOUS_EDGE
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.edge_utils import invoke_edge_appctl_command
from cli.utils.node_utils import get_node_type
from cli.utils.node_utils import is_baremetal

from compat23 import compat23_httplib
from compat23.compat23_utils import compat23_string_types
from compat23.compat23_utils import compat23_urlparse


g_node_type = get_node_type()
# Only for edge node
if g_node_type in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE,
    NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
    from cli.utils.edge_utils import get_edge_dp_physical_interfaces
    from cli.utils.edge_utils import get_edge_interfaces
    from cli.utils.edge_utils import get_linux_nics
elif ((g_node_type == NODE_TYPE_HOST_SLESKVM) or
      (g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
      (g_node_type == NODE_TYPE_HOST_RHELKVM) or
      (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
      is_baremetal()):
    from cli.commands.host_shared.ovs_util import get_lport_list
    from cli.commands.host_shared.ovs_util import get_lif_list
    from cli.commands.host_shared.ovs_util import get_vif_list
elif g_node_type == NODE_TYPE_HOST_ESX:
    from esx_dvs_utils import get_interface_names_esx


_IFCONFIG_RE = re.compile("^([^:]+): .*$")


lg = nsx_logging.getLogger("cli.commands.common.arguments")


class IntegerArgument(object):
    """Verify arguemnt to validate an integer.

        Args:
            value: Value to validate.

        Return:
            True, if specified value is an integer in the
            range (-sys.maxzize -1) to (sys.maxsize)
            otherwise, False.
    """
    _MIN_VALUE = None
    _MAX_VALUE = None

    @classmethod
    def verify(cls, value, req_dict):
        try:
            n = int(value)
            if cls._MIN_VALUE is not None:
                if not n >= cls._MIN_VALUE:
                    return False
            if cls._MAX_VALUE is not None:
                if not n <= cls._MAX_VALUE:
                    return False
            if str(value)[0] == "+":  # Per PM, we should not allow
                return False          # integers to be prexied with +
            return True
        except ValueError:
            pass
        return False


class CaptureFileCountArgument(IntegerArgument):
    """Verify arguemnt to validate file count for rotating packet capture.

        Args:
            value: Value to validate.

        Return:
            True, if specified value is a valid file-count;
            otherwise, False.
    """
    _MIN_VALUE = 1
    _MAX_VALUE = 25


class PositiveIntegerArgument(IntegerArgument):
    """Verify arguemnt to validate a positive, non-zero integer.

        Args:
            value: Value to validate.

        Return:
            True, if specified value is a positive, non-zero integer;
            otherwise, False.
    """
    _MIN_VALUE = 1
    _MAX_VALUE = None


class NonNegativeIntegerArgument(IntegerArgument):
    """Verify arguemnt to validate a non-negative integer

        Args:
            value: Value to validate.

        Return:
            True, if specified value is not negative, otherwise, False.
    """
    _MIN_VALUE = 0
    _MAX_VALUE = None


class CidrArgument(object):
    """Verify argument to validate it is in CIDR notation."""
    @staticmethod
    def verify(value, req_dict):
        """Verify specified value is in CIDR notation.

        Args:
            value: Value to validate.

        Return:
            True, if specified value is in CIDR notation; otherwise, False.
        """
        parts = value.split("/")
        if len(parts) != 2:
            return False
        try:
            n = int(parts[1])
            if n < 0 or n > 32:
                return False
        except ValueError:
            return False
        return IpAddressArgument.verify(parts[0], req_dict)


def get_physical_interfaces_mac_name():
    interfaces = {}
    ignore_interfaces_prefixes = ["bond0", "lo", "keth-", "span-", "docker0", "kni"]
    cmd = ["ip", "-br", "link"]
    rc, out, err = invoke_command(cmd)
    if rc != 0:
        return interfaces
    lines = out.split("\n")
    for line in lines:
        tokens = line.split()
        if not tokens:
            continue
        name = tokens[0]
        if '.' in name or name.startswith(tuple(ignore_interfaces_prefixes)):
            continue
        mac = tokens[2]
        if mac not in interfaces:
            interfaces[mac] = name
    if g_node_type in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE,
        NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
        dp_interfaces = get_edge_dp_physical_interfaces()
        if dp_interfaces is not None:
            for intf in dp_interfaces:
                if intf["mac"] not in interfaces:
                    interfaces[intf["mac"]] = intf["name"]
    return interfaces


def get_interface_names(ignore_interfaces_list=[], include_down=True):
    """Get local interface names"""
    interfaces = []
    # TODO: get list of interfaces from GET /node/network/interfaces
    cmd = ["ifconfig"]
    if include_down:
        cmd.append("-a")
    rc, out, err = invoke_command(cmd)
    if rc == 0:
        lines = out.split("\n")
        for line in lines:
            m = _IFCONFIG_RE.match(line)
            if m is None:
                continue
            name = m.group(1)
            if not name.startswith(tuple(ignore_interfaces_list)):
                interfaces.append(name)
            interfaces.sort()
    lg.info("get_interface_names: List of interfaces: %s", interfaces) # log the interfaces after regex match
    return interfaces


class MacArgument(object):
    _IGNORE_INTERFACE_PREFIXES = []

    """Query argument to determine available network interfaces."""
    def __init__(self, descriptor):
        self._descriptor = descriptor

    def __call__(self, req_dict):
        """Retrieve available physcial network interface macs

        Return:
            List of available physical network interface macs
        """
        result = get_physical_interfaces_mac_name()
        return simplejson.dumps(list(result.keys()))


class ConfigurableMacArgumentDescription(object):
    @staticmethod
    def value_description(value, req_dict, factory):
        result = get_physical_interfaces_mac_name()
        return result.get(value, "Physical Network Interface")


class InterfaceArgument(object):
    _IGNORE_INTERFACE_PREFIXES = []

    """Query argument to determine available network interfaces."""
    def __init__(self, descriptor):
        self._descriptor = descriptor

    def __call__(self, req_dict):
        """Retrieve available network interfaces.

        Return:
            List of available network interfaces.
        """
        results = get_interface_names(self._IGNORE_INTERFACE_PREFIXES)
        # without bond0 in available network interfaces, caller cannot even
        # configure it.
        if g_node_type in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE]:
            results.append("bond0")
        return simplejson.dumps(results)


class VariableInterfaceArgument(InterfaceArgument):
    """Query argument to determine available network interfaces.

    On Edge nodes, it also returns logical port uuid and physical port name.
    On KVM nodes, it ONLY returns logical port uuid.
    On ESX nodes, it returns logical port uuid and vmknic port, portid, vmnic
    port.
    """
    def __call__(self, req_dict):
        """Retrieve available network interfaces or port uuid/name.

        Return:
            List of available network interfaces or port uuid/name.
        """
        if g_node_type == NODE_TYPE_HOST_ESX:
            interfaces = get_interface_names_esx()
            return simplejson.dumps(interfaces)
        interfaces = get_interface_names(self._IGNORE_INTERFACE_PREFIXES)
        if g_node_type in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE,
            NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
            interfaces += get_edge_interfaces()
        elif ((g_node_type == NODE_TYPE_HOST_SLESKVM) or
              (g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
              (g_node_type == NODE_TYPE_HOST_RHELKVM) or
              (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
              is_baremetal()):
            interfaces = self.get_kvm_interfaces()
        return simplejson.dumps(interfaces)

    def get_kvm_interfaces(self):
        interfaces = []

        # Get logical ports
        lports = get_lport_list()
        interfaces.extend(lports)
        # Get logical router ports
        lifs = get_lif_list()
        interfaces.extend(lifs)
        # Get vif
        vifs = get_vif_list()
        interfaces.extend(vifs)

        return [lst[0] for lst in interfaces]

class LogicalPortArgument(InterfaceArgument):
    def __call__(self, req_dict):
        if ((g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
              (g_node_type == NODE_TYPE_HOST_RHELKVM) or
              (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
              (g_node_type == NODE_TYPE_HOST_SLESKVM) or
              is_baremetal()):
            lports = self.get_kvm_logical_ports()
        return simplejson.dumps(lports)

    def get_kvm_logical_ports(self):
        logical_ports = []
        lports = get_lport_list()
        logical_ports.extend(lports)
        return [lst[0] for lst in logical_ports]


class ConfigurableInterfaceArgument(InterfaceArgument):
    _IGNORE_INTERFACE_PREFIXES = ["lo", "keth-", "span-", "docker0", "kni-"]


class ConfigurableMacArgument(MacArgument):
    _IGNORE_INTERFACE_PREFIXES = ["lo", "keth-", "span-", "docker0", "kni-"]


class BondSlavesArgument(object):
    @staticmethod
    def verify(value, req_dict):
        results = get_linux_nics()
        devs = value.split(",")
        for dev in devs:
            if dev not in results:
                return False
        return True


class BondPrimaryArgument(object):
    @staticmethod
    def verify(value, req_dict):
        results = get_linux_nics()
        if value not in results:
            return False
        return True

# Using the same hostname regex as defined by NAPI here:
# node/api/src/python/management_api/napi/types.yml: SystemHostname.pattern
HOSTNAME_RE = re.compile(
    "^(?=.{1,255}$)[0-9A-Za-z](?:(?:[0-9A-Za-z]|-){0,61}[0-9A-Za-z])?"
    "(?:\\.[0-9A-Za-z](?:(?:[0-9A-Za-z]|-){0,61}[0-9A-Za-z])?)*\\.?$")


class HostnameArgument(object):
    """Verify argument to validate a hostname per RFC 942/1123."""
    @staticmethod
    def verify(hostname, req_dict):
        """Validate specified hostname.

        Args:
            hostname: Hostname to validate.

        Return:
            Error message, if hostname length is too long, as per RFC 1034
            Otherwise, True, if specified hostname is valid; or, False.
        """
        if any([len(n) > 63 for n in hostname.split('.')]):
            err = "MSG_CLI_INVALID_ARGUMENT\n\n"
            err += "Each domain node can be of maximum length 63"
            return err
        else:
            return HOSTNAME_RE.search(hostname) is not None


class NetworkArgument(object):
    """Verify argument to validate a network address"""
    @staticmethod
    def verify(network, req_dict):
        """Validate specified network address.

        Args:
            network: network address to validate

        Return:
            True, if specified network address is valid; otherwise, False.
        """
        parts = network.split("/")
        if len(parts) != 2:
            return False
        return (IpAddressArgument.verify(parts[0], req_dict) and
                PrefixLengthArgument.verify(parts[1], req_dict))


class Network6Argument(object):

    @staticmethod
    def verify_prefix(prefix):
        try:
            n = int(prefix)
            if n < 0 or n > 128:
                return False
        except ValueError:
            return False
        return True

    """Verify argument to validate a network address"""
    @staticmethod
    def verify(network, req_dict):
        """Validate specified network address.

        Args:
            network: network address to validate

        Return:
            True, if specified network address is valid; otherwise, False.
        """
        parts = network.split("/")
        if len(parts) != 2:
            return False
        return (Ip6AddressArgument.verify(parts[0], req_dict) and
                Network6Argument.verify_prefix(parts[1]))


class Network46Argument(object):
    """Verify argument to validate a network address"""
    @staticmethod
    def verify(network, req_dict):
        """Validate specified network address.

        Args:
            network: network address to validate

        Return:
            True, if specified network address is valid; otherwise, False.
        """
        parts = network.split("/")
        if len(parts) != 2:
            return False
        return (NetworkArgument.verify(network, req_dict) or
                Network6Argument.verify(network, req_dict))


class IpAddressArgument(object):
    """Verify argument to validate an IP address."""
    @staticmethod
    def verify(ipaddr, req_dict):
        """Validate specified IP address.

        Args:
            ipaddr: IP address to validate.

        Return:
            True, if specified IP address is valid; otherwise, False.
        """
        if "." not in ipaddr:
            return False
        parts = ipaddr.split(".")
        if len(parts) != 4:  # TODO: assume IPv4 for now
            return False
        for part in parts:
            try:
                i = int(part)
                if i < 0 or i > 255:
                    return False
            except ValueError:
                return False
        return True


class Ip6AddressArgument(object):
    """Verify argument to validate an IPv6 address."""
    @staticmethod
    def verify(ipaddr, req_dict):
        if ipaddr.count('::') > 1:
            return False
        parts = ipaddr.split(":")
        if len(parts) < 3 or len(parts) > 8:
            return False
        if not ipaddr.startswith('::') and len(parts[0]) == 0:
            return False
        if not ipaddr.endswith('::') and len(parts[-1]) == 0:
            return False
        # Check if it's ipv4 transition address like xxx:xxx::yyy.yyy.yyy.yyy
        if parts[-1].find('.') != -1:
            # Must be a valid IPv4 address
            if not IpAddressArgument.verify(parts[-1], req_dict):
                return False
            # Remove the IPv4 part
            parts.pop()
        for p in parts:
            if len(p) == 0:
                continue
            try:
                n = int(p, 16)
                if n < 0 or n > 0xffff:
                    return False
            except ValueError:
                return False
        return True


class Ip46AddressArgument(object):
    """Verify argument to validate an IP address."""
    @staticmethod
    def verify(ipaddr, req_dict):
        """Validate specified IP address.

        Args:
            ipaddr: IP address to validate.

        Return:
            True, if specified IP address is valid; otherwise, False.
        """
        return (IpAddressArgument.verify(ipaddr, req_dict) or
                Ip6AddressArgument.verify(ipaddr, req_dict))


class IpOrMacAddressArgument(object):

    @staticmethod
    def verify_mac(mac):
        parts = mac.split(":")
        if len(parts) != 6:
            return False
        for p in parts:
            if len(p) != 2:
                return False
            try:
                _n = int(p, 16)
            except ValueError:
                return False
        return True

    """Verify argument to validate an IP or MAC address."""
    @staticmethod
    def verify(addr, req_dict):
        """Validate specified IP address.

        Args:
            ipaddr: IP address to validate.

        Return:
            True, if specified IP address is valid; otherwise, False.
        """
        return (Ip46AddressArgument.verify(addr, req_dict) or
                IpOrMacAddressArgument.verify_mac(addr))


class MulticastIpAddressArgument(object):
    """Verify argument to validate an multicast IP address."""
    @staticmethod
    def verify(ipaddr, req_dict):
        """Validate specified IP address.

        Args:
            ipaddr: Multicast IP address to validate.

        Return:
            True, if specified IP address is valid; otherwise, False.
        """
        if "." not in ipaddr:
            return False
        parts = ipaddr.split(".")
        if len(parts) != 4:  # TODO: assume IPv4 for now
            return False
        first_part = True
        for part in parts:
            try:
                i = int(part)
                if first_part:
                    if i < 224 or i > 239:
                        return False
                    first_part = False
                else:
                    if i < 0 or i > 255:
                        return False
            except ValueError:
                return False
        return True


class NetmaskArgument(object):
    """Verify argument to validate a netmask."""
    @staticmethod
    def verify(netmask, req_dict):
        """Validate specified netmask.

        Args:
            ipaddr: Netmask to validate.

        Return:
            True, if specified netmask is valid; otherwise, False.
        """
        if "." not in netmask:
            return False
        parts = netmask.split(".")
        if len(parts) != 4:
            return False
        bin_str = ""
        for part in parts:
            try:
                bin_str += NetmaskArgument._to_bin_str(int(part))
            except ValueError:
                return False
        allow_one = True
        for pos in range(len(bin_str)):
            if bin_str[pos] == "1" and not allow_one:
                return False
            elif bin_str[pos] == "0" and allow_one:
                allow_one = False
        return True

    @staticmethod
    def _to_bin_str(int_val):
        """Convert an integer value in range(0, 256) to binary string.

        Args:
            int_val: Integer value to convert.
        Returns:
            Binary string or raise ValueError if int_val is invalid.
        """
        if int(int_val) != int_val or int_val < 0 or int_val > 255:
            raise ValueError("Invalid value")
        bin_str = bin(int_val)[2:]  # strip "0b" prefix
        # bin(2) return "0b10" instead of "0b00000010", so prefix with 0s
        return "0" * (8 - len(bin_str)) + bin_str


class PrefixLengthArgument(object):
    """Verify argument to validate IP prefix length."""
    @staticmethod
    def verify(prefix_len, req_dict):
        """Validate IP address prefix length.

        Args:
            prefix_len: IP prefix length to validate.

        Return:
            True, if IP prefix length is valid; otherwise, False.
        """
        i = int(prefix_len)
        if i < 0 or i > 32:
            return False
        return True


class HostnameOrIpAddressArgument(object):
    """Verify argument to validate a hostname or IP address."""
    @staticmethod
    def verify(value, req_dict):
        """Verify specified value is a valid hostname or IP address.

        Args:
            value: Value to validate.

        Return:
            True, if specified value is valid; otherwise, False.
        """
        return (IpAddressArgument.verify(value, req_dict) or
                HostnameArgument.verify(value, req_dict))


def _verify_port(port):
    """Return True if valid port, else False."""
    try:
        port = int(port)
        if port < 1 or port > 65535:
            return False
    except ValueError:
        return False
    return True


class HostnameOrIpAddressOptionalPortArgument(object):
    """Verify argument to validate a hostname or IP address with optional
       port."""
    @staticmethod
    def verify(value, req_dict):
        """Verify specified value is a valid hostname or IP address with
           optional port delimited by a colon.

        Args:
            value: Value to validate.

        Return:
            True, if specified value is valid; otherwise, False.
        """
        parts = value.split(":")
        len_parts = len(parts)
        if len_parts > 2:
            return False
        elif len_parts == 2:
            if not _verify_port(parts[1]):
                return False
        return (IpAddressArgument.verify(parts[0], req_dict) or
                HostnameArgument.verify(parts[0], req_dict))


class IpAddressOptionalPortArgument(object):
    """Verify argument to validate an IP address with optional port."""
    @staticmethod
    def verify(value, req_dict):
        """Verify specified value is a valid IP address with
           optional port delimited by a colon.

        Args:
            value: Value to validate.

        Return:
            True, if specified value is valid; otherwise, False.
        """
        parts = value.split(":")
        len_parts = len(parts)
        if len_parts > 2:
            return False
        elif len_parts == 2:
            if not _verify_port(parts[1]):
                return False
        return IpAddressArgument.verify(parts[0], req_dict)


class HardwareAddressArgument(object):
    """Verify argument to validate a MAC address."""
    @staticmethod
    def verify(macaddr, req_dict):
        """Verify specified value is a valid MAC address.

        Args:
            macaddr: MAC address to validate.

        Return:
            True, if specified MAC address is valid; otherwise, False.
        """
        parts = macaddr.split(":")
        if len(parts) != 6:
            return False
        try:
            for part in parts:
                i = int(part, 16)
                if i < 0 or i > 255:
                    return False
        except ValueError:
            return False
        return True


TP_WITH_COLONS_RE = re.compile("^([0-9A-Fa-f]{2}:){31}([0-9A-Fa-f]{2}$)")
TP_NO_COLONS_RE = re.compile("^[0-9A-Fa-f]{64}$")


class ThumbprintSHA256HexDigitsArgument(object):
    """
    Verify argument to validate a SHA256 thumbprint. May be 64 hex digits,
    or 32 pairs of two-byte hex digits separated by colons.
    """

    @staticmethod
    def verify(thumbprint, req_dict):
        """Validate thumbprint

        Args:
            thumbprint: Thumbprint to validate.
        Return:
            True, if specified thumbprint is valid; otherwise, False.
        """
        return (TP_WITH_COLONS_RE.match(thumbprint) is not None or
                TP_NO_COLONS_RE.match(thumbprint) is not None)

    @staticmethod
    def normalize(thumbprint):
        """
        Normalize the thumbprint. Remove colons and coerce to lower case.
        """
        return thumbprint.replace(":", "").lower()


class ApiThumbprintArgument(ThumbprintSHA256HexDigitsArgument):
    """Verify argument to validate an API thumbprint."""
    pass


class PrefixOrIpAddressArgument(object):
    """Verify argument to validate a prefix or IP address"""
    @staticmethod
    def verify(network, req_dict):
        """Validate specified prefix or IP address.

        Args:
           network: network address to validate

        Return:
           True, if specified address is valid; otherwise, False.
        """
        if "/" not in network:
            network += "/32"
        parts = network.split("/")
        if len(parts) != 2:
            return False
        return (IpAddressArgument.verify(parts[0], req_dict) and
                PrefixLengthArgument.verify(parts[1], req_dict))


class PrefixOrIpv6AddressArgument(object):
    """Verify argument to validate a prefix or IPv6 address"""
    @staticmethod
    def verify(network, req_dict):
        """Validate specified prefix or IPv6 address.

        Args:
           network: network address to validate

        Return:
           True, if specified address is valid; otherwise, False.
        """
        if "/" not in network:
            network += "/128"
        parts = network.split("/")
        if len(parts) != 2:
            return False
        return (Ip6AddressArgument.verify(parts[0], req_dict) and
                Network6Argument.verify_prefix(parts[1]))


class ControllerNodeUuidArgument(object):
    """Query argument to determine existing controller node uuids."""
    def __init__(self, descriptor):
        self._descriptor = descriptor

    @staticmethod
    def _get_controller_uuids():
        results = []
        status, json, err, _ = rest_utils.local_node_request(
            "GET", "/cluster/nodes", proton_request=True)
        if status == compat23_httplib.OK:
            # Starting in Flash, the controller role is also on the
            # converged appliance, so use the manager_role node_id
            results = [node["id"] for node in json["results"] if
                       node.get("manager_role")]
        return results

    def __call__(self, req_dict):
        """Retrieve existing controller uuids

        Return:
            List of controller node uuids.
        """
        results = self._get_controller_uuids()
        return simplejson.dumps(results)


class VersionArgument(object):
    """Verify argument to validate a version number"""
    @staticmethod
    def verify(version, req_dict):
        """Validate the specified version number.

        Args:
           version: verison number to validate

        Return:
           True, if specified version is valid; otherwise, False.
        """
        res = False
        if re.search(r'^(\d+\.){5}(\d+){7,}$', version):
            res = True
        return res

class LogicalIdArgument(object):
    """Verify argument to validate Logical Port ID or Logical Switch ID"""
    @staticmethod
    def verify(id_str, req_dict):
        """Validate the specified Logical Port ID or Logical Switch ID.

        Args:
           id_str: port id or switch id to validate

        Return:
           True, if specified id_str is valid; otherwise, False.
        """
        try:
            uuid_val = uuid.UUID(id_str)
        except ValueError:
            return False
        return True

class PciAddressArgument(object):
    @staticmethod
    def verify(addr, req_dict):
        """Verify PCI physical ports"""
        t = addr.split(",")
        for pci_addr in t:
            if pci_addr.count('.') != 1:
                return False
            parts = pci_addr.split(":")
            if len(parts) != 3:
                return False
            for p in parts:
                if len(p) == 0:
                    return False
                if p.count('.') == 0:
                    try:
                        n = int(p, 16)
                        if n < 0 or n > 0xffff:
                            return False
                    except ValueError:
                        return False
        return True


class ScpFileUrlArgument(object):
    _ALLOWED_SCHEMES = ["scp"]

    """Verify argument to validate a file url."""
    @classmethod
    def verify(cls, specified_url, req_dict):
        """Validate file url.

        Args:
            specified_url: File url to validate.
        Return:
            True, if specified file url is valid; otherwise, False.
        Notes:
            We currently only support scp urls in the form of:
            - scp://username@host[:port]/path/filename
            - http://[username:password@]host[:port]/path/filename
            - passwords in scp urls are not supported, the scp command
              prompts the user for the password.
        """
        scheme_matched = False
        for scheme in cls._ALLOWED_SCHEMES:
            if specified_url.lower().startswith("%s://" % scheme):
                scheme_matched = True
                break
        if not scheme_matched:
            return False
        url = compat23_urlparse(specified_url)
        # Verify port is None or numeric
        try:
            # If specified_url = "scp://admin@localhost:/path", assigning
            # port = url.port causes Ubuntu 12.04 to raise ValueError;
            # however, Ubuntu 14.04 sets port = None. Ubuntu 14.04's
            # behavior seems more correct.
            parts = url.netloc.split("@")[-1].split(":")
            num_parts = len(parts)
            if num_parts > 2:
                return False
            if (num_parts == 1 or parts[1] == ""):
                port = None
            else:
                port = int(parts[1])
                if port is not None and (port <= 0 or port > 65535):
                    return False
        except ValueError:
            return False
        if url.scheme not in cls._ALLOWED_SCHEMES:
            return False
        elif url.scheme == "scp" and (not url.username or url.password):
            # scp prompts for password
            return False
        elif url.password and not url.username:
            return False
        return (
            isinstance(url.hostname, compat23_string_types) and
            isinstance(url.path, compat23_string_types) and
            len(url.hostname) > 0 and len(url.path) > 0)


class CaptureSessionIdArgument(object):
    """Query argument for span session id"""
    def __init__(self, descriptor):
        self._descriptor = descriptor

    def __call__(self, req_dict=None):
        cmd = ["span/show"]
        num_sessions = 0
        result = ["1", "2", "3"]
        if g_node_type not in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE,
            NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
            return simplejson.dumps(result)
        rc, out, err = invoke_edge_appctl_command(cmd, "nsxa")
        if rc == 0:
            result.clear()
            out_json = simplejson.loads(out)
            if out_json:
                num_sessions = len(out_json)
            for i in range(num_sessions):
                result.append(str(i))
        return simplejson.dumps(result)
