# ***************************************************************************
# Copyright 2015-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
try:
    import simplejson as json
except ImportError:
    import json
import sys

from .command_utils import invoke_command
from cli.utils.error_code import ErrorCode as ec
from cli.utils.location_utils import FilePaths
from cli.utils.node_utils import is_windows

import nestdb_cli.nestdb_cli_client as ndbcli

lg = nsx_logging.getLogger("cli.utils.nestdb_utils")

if is_windows():
    _NESTDB_CLI = FilePaths.NSX_NESTDB_CLI
else:
    _NESTDB_CLI = "/opt/vmware/nsx-nestdb/bin/nestdb-cli"

L2VPN_SESSION_DB = "vmware.nsx.nestdb.L2VpnSessionMsg"

def invoke_nestdb_command(cmd_list):
    """Invoke an nestdb CLI command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["get", "<type>"]).
    Return:
        (returncode, output, error) tuple.
    """
    cmd = [_NESTDB_CLI, "--json", "--cmd"]
    cmd.extend(cmd_list)
    rc, out, err = invoke_command(cmd)
    if rc != 0:
        out = out + "Error message: " + err
        # CLI shows err (doesn't show out) when rc is not 0
        err = "Cannot access NestDB"
    return rc, out, err

def get_type(msg_type, key=None, filter_dict=None, filter_func=None, vdb_id=None):
    """Get type from nestdb by key string

    Args:
        msg_type: message type name
        key: key string to match the message
             None to get all messages
        filter_dict: key value pair to filter message
        filter_func: function to filter output

    Return:
        (returncode, result_json, err_msg) tuple.
    """
    out = []
    err = ""
    cmds = ["get", msg_type]
    if key:
        cmds.append(key)
    if vdb_id is not None:
        cmds.append(vdb_id)
    try:
        rc, rv, err = invoke_nestdb_command(cmds)
        if rc == 0:
            lg.debug("Successful to Nestdb-Get type: %s, key: %s, rc: %d\n%s",
                    msg_type, key, rc, rv)
            resp = json.loads(rv)
            if "result" in resp:
                out = [v["value"] for v in resp["result"]]
            if msg_type == L2VPN_SESSION_DB:
                for l2vpn_session in out:
                    if "transport_tunnel" in l2vpn_session:
                        l2vpn_session["transport_tunnels"] = l2vpn_session["transport_tunnel"]
                        del l2vpn_session["transport_tunnel"]
        else:
            lg.error("Failed to Nestdb-Get type: %s, key: %s, rc: %d\n%s\nerr: %s",
                     msg_type, key, rc, rv, err, error_code=ec.OPERATION_FAILED)
        if out and isinstance(filter_dict, dict):
            for k, v in list(filter_dict.items()):
                out = [item for item in out
                       if isinstance(item, dict) and v in item[k]]
        if filter_func is not None:
            out = [item for item in out if filter_func(item)]
    except Exception as e:
        err = "Nestdb may not be responding: %s" % e
        lg.error(err, error_code=ec.INTERNAL_ERROR)
        rc = -1
    return rc, json.dumps(out), err

def query_type(msg_type, value_list, filter_dict=None, filter_func=None, vdb_id=None):
    """Get type from nestdb by field equal value

    Args:
        msg_type: message type name
        value_list: list of (field, op, value)
        field: field name
        value: value for field
        euqual: operator
        filter_dict: key value pair to filter message
        filter_func: function to filter output

    Return:
        (returncode, result_json, err_msg) tuple.
    """
    out = []
    err = ""
    opmsg = {}
    vmsg = {}
    for entry in value_list:
        key, op, value = entry
        opmsg[key] = {"op": op}
        vmsg[key] = value
    cmds = ["query", msg_type]
    try:
        cmds.append(json.dumps(opmsg))
        cmds.append(json.dumps(vmsg))
        if vdb_id is not None:
            cmds.append(vdb_id)
        rc, rv, err = invoke_nestdb_command(cmds)
        if rc == 0:
            lg.debug("Successful to Nestdb-Quert type: %s where %s, rc: %d\n%s",
                    msg_type, value_list, rc, rv)
            resp = json.loads(rv)
            if "result" in resp:
                out = [v["value"] for v in resp["result"]]
        else:
            lg.error("Failed to Nestdb-Query type: %s where %s, rc: %d\n%s\nerr: %s",
                     msg_type, value_list, rc, rv, err,
                     error_code=ec.OPERATION_FAILED)
        if out and isinstance(filter_dict, dict):
            for k, v in list(filter_dict.items()):
                out = [item for item in out
                       if isinstance(item, dict) and v in item[k]]
        if filter_func is not None:
            out = [item for item in out if filter_func(item)]
    except Exception as e:
        err = "Nestdb may not be responding: %s" % e
        lg.error(err, error_code=ec.INTERNAL_ERROR)
        rc = -1
    return rc, json.dumps(out), err

def get_msg(msg, uuid=None):
    rc = -1
    out = []
    err = ""
    ndbcli.set_schema_prefix("vmware.nsx.nestdb")
    ndbcli.set_client_id("nestdb-cli")
    if uuid is None:
        cmd = [msg]
    else:
        cmd = [msg, uuid]
    with open('/dev/null', 'w') as f:
        sys.stdout = f
        sys.stderr = f
        try:
            (res, res_json) = ndbcli.run_get_command(cmd, True)
        finally:
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__
    try:
        res = json.loads(res_json)
        for record in res['result']:
            if 'value' in record:
                out.append(record['value'])
        if len(out) > 0:
            rc = 0
    except Exception as e:
        rc = -1
        out = []
        err = "Fail to get: %s\n%s" % (msg, e)
    return rc, json.dumps(out), err

def put_msg(msg_type, msg):
    rc = -1
    out = ""
    err = ""
    ndbcli.set_schema_prefix("vmware.nsx.nestdb")
    ndbcli.set_client_id("nestdb-cli")
    cmd = [msg_type, msg]
    with open('/dev/null', 'w') as f:
        sys.stdout = f
        (res, res_json) = ndbcli.run_put_command(cmd, True)
    sys.stdout = sys.__stdout__
    if res:
        rc = 0
    else:
        out = "Failed"
        err = "Fail to put: %s\n%s" % (msg, res_json)
    return rc, json.dumps(out), err
