# Copyright (C) 2016-2019 VMware, Inc. All rights reserved.
import argparse
import base64
import cmd
import datetime
import os
import sys

# Workaround for rhel environment, where the default
# python-protobuf is 2.5, which is incompatible with
# the nestdb-built python proto bindings.
# This forces python to load the python proto bindings
# from nsx-python-protobuf.
if os.path.isfile('/etc/redhat-release'):
    if ('google' in sys.modules and
        sys.modules['google'].__path__ !=
            '/opt/vmware/nsx-common/lib/python/google'):
        del sys.modules['google']
    import google

from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.descriptor_pb2 import FileDescriptorSet
from google.protobuf.descriptor_pool import DescriptorPool
from google.protobuf.internal import decoder
from google.protobuf.message import Message
from google.protobuf.message_factory import MessageFactory
# Optionally try to import ipaddress or ipaddr
try:
    import ipaddress
except ImportError:
    try:
        import ipaddr as ipaddress
    except:
        pass
import logging
import re
try:
    import simplejson as json
except ImportError:
    import json
import socket
import struct
import uuid

from vmware.nsx.rpc import NsxRpcClient
from vmware.nsx.rpc import NsxRpcCallError
from vmware.nsx.nestdb import nestdb_pb2
from vmware.nsx.nestdb import nestdb_common_pb2

logger = logging.getLogger(__name__)

DEFAULT_SCHEMA = '/opt/vmware/nsx-nestdb/schema/nestdb.schema'
DEFAULT_SCHEMA_PREFIX = 'vmware.nsx.nestdb'
DEFAULT_CLIENT_ID = 'nestdb-cli'
DEFAULT_TIMEOUT = 10
DEFAULT_ENDPOINT = 'unix:///var/run/vmware/nestdb/nestdb-server.sock'
descriptor_pool = None
message_factory = MessageFactory()

client_id = ''
schema_prefix = ''

# Following is the list of NestDbCLI commands, their usage and descriptions.
nestdbcli_commands = {
    'EOF': {
        'usage': ["quit", "<Ctrl>+D"],
        'description': "To quit the NestDb shell do 'quit' or <Ctrl>+D"},
    'quit': {
        'usage': ["quit", "<Ctrl>+D"],
        'description': "To quit the NestDb shell do 'quit' or <Ctrl>+D"},
    'server': {
        'usage': ["server", "server <endpoint>"],
        'description': "Set this CLI's server endpoint to connect to "
                       "<endpoint>. To get the current endpoint in use, "
                       "issue the command without arguments."},
    'output_format': {
        'usage': ["output_format", "output_format json",
                  "output_format plain"],
        'description': "Set the formatting for commands' output. To query for "
                       "current output_format value, issue the command without"
                       " arguments."},
    'cli_schema': {
        'usage': ["cli_schema", "cli_schema <schema file>"],
        'description': "Set this CLI's local schema file to <schema file>. To "
                       "get the current schema file in use, issue the command "
                       "without arguments."},
    'schema': {
        'usage': ["schema load <schema file>", "schema show",
                  "schema version"],
        'description': "The schema command can be used to load a schema file, "
                       "show the currently loaded schema file, or to show the "
                       "version of the currently loaded schema file."},
    'get': {
        'usage': ["get <object type> [embedded id (json format)] [vdb_id]"],
        'description': "Gets all objects from NestDb that match the object "
                       "type. If embedded id is given, then only get the "
                       "object that matches the type and id. If vdb_id "
                       "is given, then the get operation is applied on the "
                       "specified vdb_id. Otherwise, the main virtual "
                       "database is used."},
    'put': {
        'usage': ["put <object type> <object in json> "
                  "[vdb_list [vdb_id] ... [vdb_id]]"],
        'description': "Adds the object to the NestDb. If an object with the "
                       "same type and id is already present in the NestDb, "
                       "it will be overwritten to the supplied object. "
                       "If vdb_list is given, then the operation is applied "
                       "on the specified vdb_ids. Otherwise, the main virtual "
                       "database is used."
                       "Note that sync option is not set for this request."},
    'delete': {
        'usage': ["delete <object type> <embedded id (json format)> "
                  "[vdb_list [vdb_id] ... [vdb_id]]"],
        'description': "Deletes the object with the supplied object id from "
                       "NestDb. "
                       "If vdb_list is given, then the operation is applied "
                       "on the specified vdb_ids. Otherwise, the main virtual "
                       "database is used."
                       "Note that sync option is not set for this request."},
    'flush': {
        'usage': ["flush <object type> [vdb_list [vdb_id] ... [vdb_id]]"],
        'description': "Deletes all objects of the given type from NestDb. "
                       "If vdb_list is given, then the operation is applied "
                       "on the specified vdb_ids. Otherwise, the main virtual "
                       "database is used."
                       "Note that sync option is not set for this request."},
    'copy': {
        'usage': ["copy <object type> <embedded id (json format)> "
                  "[src_vdb_id <vdb_id>] "
                  "[dst_vdb_list [vdb_id] ... [vdb_id]]"],
        'description': "Copies the object with the supplied object id from "
                       "the src_vdb_id to the list of destination vdb_ids."
                       "If src_vdb_id or dst_vdb_list is not given, then "
                       "the main virtual database is used."
                       "Note that sync option is not set for this request."},
    'query': {
        'usage': ["query <object type> <op in json> <val in json> "
                  "[sort_by in json] [vdb_id]"],
        'description': "Given a condition, specified by the operation 'op' "
                       "and value 'val' in json format, retrieve matching "
                       "objects in NestDb. To retreive objects of a "
                       "particular type (for ex., RuleMsg), the 'op' in the "
                       "condition will be a json representation of a "
                       "RuleOpMsg object and the 'val' in the condition will "
                       "be a json representation of a RuleMsg object. "
                       "'sort_by' in json format can optionally be given. "
                       "'sort_by' is the json representation of a RuleOpMsg "
                       "and the result will be sorted according to the "
                       "populated field. "
                       "If vdb_id is given, then the query operation is "
                       "applied on the specified vdb_id. Otherwise, the "
                       "main virtual database is used."},
    'transaction_log': {
        'usage': ["transaction_log [limit_size <in byte>] "
                  "[[client <client id>] [object_type <object_type>] "
                  "[object_id <object_id>] [vdb_id <vdb_id>]]"],
        'description': "'transaction_log' prints transactions processed by "
                       "NestDb in chronological order. Specify in bytes the "
                       "size limit on the output. Specify any filtering "
                       "conditions on 'client' (client name), 'object_type', "
                       "'object_id', and 'vdb_id' as optional arguments. "
                       "Include the argument name each time if you are "
                       "specifying multiple values for the same "
                       "filtering attribute."},
    'replay': {
        'usage': ["replay <transaction_log file (json format)> "
                  "[[client <client id>] [object_type <object_type>] "
                  "[vdb_id <vdb_id>]] [min_txn_id <txn_id>] "
                  "[max_txn_id <txn_id>]"],
        'description': "'replay' reads and parses the given transaction log "
                       "file and makes the same requests to NestDB server. "
                       "Specify and filtering conditions on 'client' "
                       "(client name), 'object_type', 'vdb_id', 'min_txn_id', "
                       "and 'max_txn_id' as optional arguments. 'client', "
                       "'object_type', 'vdb_id' can be specified multiple "
                       "times."},
    'metrics': {
        'usage': ["metrics [<prefix>]"],
        'description': "Print all metrics' key value pairs whose keys share "
                       "the given prefix."},
    'object_type': {
        'usage': ["object_type list", "object_type show <object type>"],
        'description': "The 'list' sub command lists all valid object types. "
                       "The 'show' sub command prints the definition of a "
                       "supplied object type."},
    'attribute': {
        'usage': ["attribute query <object type> <path to field> "
                  "<field value> [vdb_id]",
                  "attribute update <object type> <embedded id (json format)> "
                  "<path to field> <field value> "
                  "[vdb_list [vdb_id] ... [vdb_id]]"],
        'description': "The first command can be used to retrieve objects of "
                       "a supplied type that have the given field value for "
                       "the specified field. The query field can be a nested "
                       "field and '.' can be used to indicate the nesting. "
                       "The second command can be used to update a specific "
                       "attribute  of an object. Note that sync option is not "
                       "set for this request. "
                       "If vdb_list for update or vdb_id for query is given, "
                       "then the operation is applied on the specified "
                       "vdb_id(s). "
                       "Otherwise, the main virtual database is used."},
    'vdb': {
        'usage': ["vdb create <vdb_name>",
                  "vdb delete <vdb_name>",
                  "vdb get [<vdb_name>]"],
        'description': "The first command can be used to create a new "
                       "virtual database. The second command can be used "
                       "to delete an existing virtual database. The last "
                       "command is used to get the list of virtual database "
                       "name to vdb_id mapping."},
    'certificate': {
        'usage': ["certificate add <client_id> <cert file>",
                  "certificate delete <client_id>",
                  "certificate get <client_id>"],
        'description': "The first command can be used to add a new "
                       "certificate for the given client id. "
                       "Note that it overwrites existing certificate for "
                       "the given client id. The second command can be used "
                       "to delete an existing certificate."},

    'merkle_tree': {
        'usage': ["merkle_tree <client_id> [vdb_id]"],
        'description': "Prints out the merkle tree representation of "
                       "objects owned by the given client id in the "
                       "given vdb_id."}}

#
# Utility functions needed for processing CLIs.
#

# Extend the json.JSONEncoder class to deal with byte instances.
class JSONSafeEncoder(json.JSONEncoder):

    def default(self, obj):
        # Encode all byte fields as Base64.
        if isinstance(obj, (bytes, bytearray)):
            return str(base64.b64encode(obj))

        # Otherwise, use the default implementation.
        return json.JSONEncoder.default(self, obj)

def dump_json(data):
    if sys.version_info.major > 2:
        return json.dumps(data, indent=4,
                cls=JSONSafeEncoder, separators=(', ', ' : '))
    else:
        return json.dumps(data, indent=4,
                encoding='latin-1', separators=(', ', ' : '))

def set_schema_prefix(prefix):
    global schema_prefix
    schema_prefix = prefix


def set_client_id(cid):
    global client_id
    client_id = cid


def initialize_descriptor_pool(schema_file):
    global descriptor_pool
    if descriptor_pool is not None:
        return
    with open(schema_file, mode='rb') as file:
        schema = file.read()
    file_descriptor_set = FileDescriptorSet()
    file_descriptor_set.ParseFromString(schema)
    descriptor_pool = DescriptorPool()
    for meta_file in file_descriptor_set.file:
        descriptor_pool.Add(meta_file)


def convert_ipv6_string_to_bytes(ipv6):
    """
    Utility method that converts an ipv6 string notation form
    into bytes.
    If ipaddress or ipaddr module is not available, then this method
    simply encodes the input.
    """
    if "ipaddress" not in sys.modules and "ipaddr" not in sys.modules:
        try:
            if (type(e_v) is unicode):
                return str(ipv6)
        except:
            return ipv6.encode()

    try:
        ipv6 = unicode(ipv6)
    except:
        pass
    try:
        return ipaddress.IPv6Address(ipv6).packed
    except:
        try:
            if (type(e_v) is unicode):
                return str(ipv6)
        except:
            return ipv6.encode()


def convert_ipv4_to_string(ipv4):
    return socket.inet_ntoa(struct.pack('!L', ipv4))


def convert_ipv6_bytes_to_string(ipv6):
    """
    Utility method that converts an ipv6 bytes to
    string notation form.
    If ipaddress or ipaddr module is not available, then this method
    simply decodes the input into a string.
    """
    if "ipaddress" not in sys.modules and "ipaddr" not in sys.modules:
        return ipv6.decode()

    try:
        ipv6_str = str(ipaddress.IPv6Address(ipv6))
    except:
        tmp = struct.unpack('!QQ', ipv6)
        tmp_ipv6 = ipaddress.IPv6Address('::')
        tmp_ipv6._ip = (tmp[0] << 64) | tmp[1]
        ipv6_str = str(tmp_ipv6)
    return ipv6_str


def create_json_from_field_val(field_path, value):
    """
    Utility method that creates a json structure from the given
    field path (i.e., priority)
    and value.
    """
    fields = field_path.split(".")
    if len(fields) == 0:
        return {}
    try:
        new_value = int(value)
    except ValueError:
        try:
            new_value = float(value)
        except ValueError:
            if value.lower() == 'false':
                new_value = False
            elif value.lower() == 'true':
                new_value = True
            else:
                new_value = value
    curr_dict = {fields[-1]: new_value}
    for index in reversed(list(range(len(fields) - 1))):
        new_dict = {fields[index]: curr_dict}
        curr_dict = new_dict
    return curr_dict


def extract_object_type_enum_name(msg_type):
    """
    Utility function that converts a protobuf fully qualified
    name into ObjectType enum string name.
    """
    stripped_msg_type = msg_type.split('.')[-1].replace('Msg', '')
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', stripped_msg_type)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).upper()


def get_msg_type_from_object_type(value):
    """
    Utility function that converts an ObjectType enum string name
    into a protobuf fully qualified name.
    """
    def camelcase():
        while True:
            yield str.capitalize

    c = camelcase()
    msg_type = schema_prefix + '.'
    msg_type += "".join(next(c)(x) if x else '_' for x in value.split("_"))
    msg_type += "Msg"
    return msg_type


def get_msg_types(schema_file):
    """
    Utility function that returns the list of currently supported
    message types in nestdb.
    """
    msg_types = []
    for object_type_val in get_object_type_vals(schema_file):
        msg_types.append(get_msg_type_from_object_type(object_type_val))
    return msg_types


def find_proto_msg_type(schema_file, msg_type):
    """
    Utility function that finds the msg_type
    in the descriptor pool. If an exact
    match does not exist, then it will
    attempt to find a partial match,
    only if a single match exist. Otherwise,
    this will raise an error.
    """
    matched_msg_types = [
        str(x) for x in get_msg_types(schema_file) if msg_type in str(x)]
    if len(matched_msg_types) != 1:
        raise
    return matched_msg_types[0]


def get_proto_object(schema_file, msg_type):
    """
    Utility function that creates a protobuf object
    of msg_type using protobuf descriptors from
    the given schema file.
    """
    initialize_descriptor_pool(schema_file)
    descriptor = descriptor_pool.FindMessageTypeByName(msg_type)
    return message_factory.GetPrototype(descriptor)()


def field_type_to_string(field_type):
    """
    Utility function that converts the field
    descriptor type into a string.
    """
    if field_type == FieldDescriptor.TYPE_BOOL:
        return "BOOL"
    elif field_type == FieldDescriptor.TYPE_BYTES:
        return "BYTES"
    elif field_type == FieldDescriptor.TYPE_DOUBLE:
        return "DOUBLE"
    elif field_type == FieldDescriptor.TYPE_ENUM:
        return "ENUM"
    elif field_type == FieldDescriptor.TYPE_FIXED32:
        return "FIXED32"
    elif field_type == FieldDescriptor.TYPE_FIXED64:
        return "FIXED64"
    elif field_type == FieldDescriptor.TYPE_FLOAT:
        return "FLOAT"
    elif field_type == FieldDescriptor.TYPE_INT32:
        return "INT32"
    elif field_type == FieldDescriptor.TYPE_INT64:
        return "INT64"
    elif field_type == FieldDescriptor.TYPE_SFIXED32:
        return "SFIXED32"
    elif field_type == FieldDescriptor.TYPE_SFIXED64:
        return "SFIXED64"
    elif field_type == FieldDescriptor.TYPE_SINT32:
        return "SINT32"
    elif field_type == FieldDescriptor.TYPE_SINT64:
        return "SINT64"
    elif field_type == FieldDescriptor.TYPE_SINT64:
        return "SINT64"
    elif field_type == FieldDescriptor.TYPE_STRING:
        return "SINT64"
    elif field_type == FieldDescriptor.TYPE_UINT32:
        return "UINT32"
    elif field_type == FieldDescriptor.TYPE_UINT64:
        return "UINT64"
    return str(field_type)


def get_object_type_schema_internal(descriptor):
    schema_dict = {}
    for field in descriptor.fields:
        field_name = str(field.name)
        if field.label == FieldDescriptor.LABEL_REPEATED:
            object_list = []
            schema_dict[field_name] = object_list
            if (field.cpp_type == FieldDescriptor.TYPE_MESSAGE or
                    field.cpp_type == FieldDescriptor.TYPE_GROUP):
                object_list.append(
                    get_object_type_schema_internal(field.message_type))
            else:
                object_list.append(field_type_to_string(field.type))
        else:
            if (field.cpp_type == FieldDescriptor.TYPE_MESSAGE or
                    field.cpp_type == FieldDescriptor.TYPE_GROUP):
                schema_dict[field_name] = get_object_type_schema_internal(
                    field.message_type)
            else:
                schema_dict[field_name] = field_type_to_string(field.type)
    return schema_dict


def get_object_type_schema(schema_file, object_type):
    """
    Utility function that returns the schema in json
    format of an object type.
    """
    initialize_descriptor_pool(schema_file)
    return get_object_type_schema_internal(
        descriptor_pool.FindMessageTypeByName(object_type))


def get_object_type(schema_file, enum_type):
    """
    Utility function that returns the object_type
    number from enum type name using protobuf descriptors
    from the given schema file.
    """
    initialize_descriptor_pool(schema_file)
    descriptor = descriptor_pool.FindEnumTypeByName(
        schema_prefix + ".ObjectType")
    return descriptor.values_by_name[enum_type].number


def get_object_type_val(schema_file, object_type):
    """
    Utility function that returns the ObjectType
    enum name from a given object_type using
    protobuf descriptors from the given schema file.
    """
    initialize_descriptor_pool(schema_file)
    descriptor = descriptor_pool.FindEnumTypeByName(
        schema_prefix + ".ObjectType")
    return str(descriptor.values_by_number[object_type].name)


def get_object_type_vals(schema_file):
    """
    Utility function that returns the list of ObjectType
    enum names using protobuf descriptors from the given schema file.
    """
    initialize_descriptor_pool(schema_file)
    descriptor = descriptor_pool.FindEnumTypeByName(
        schema_prefix + ".ObjectType")
    return (str(x) for x in descriptor.values_by_name)


def get_key_type_internal(descriptor):
    """
    Utility function that returns the encoding type of a key field.
    It returns either STRING, UUID, SIGNED_VARINT, or UNSIGNED_VARINT.
    """
    # Returns String, signed_varint, unsigned_varint, uuid int
    for field in descriptor.fields:
        if field.has_options:
            if field.GetOptions().HasExtension(nestdb_common_pb2.key):
                if (field.cpp_type == FieldDescriptor.TYPE_MESSAGE or
                        field.cpp_type == FieldDescriptor.TYPE_GROUP):
                    if "UuidMsg" in field.message_type.full_name:
                        return "UUID"
                    else:
                        return "STRING"
                else:
                    if (field.type == FieldDescriptor.TYPE_DOUBLE or
                            field.type == FieldDescriptor.TYPE_FIXED32 or
                            field.type == FieldDescriptor.TYPE_FIXED64 or
                            field.type == FieldDescriptor.TYPE_FLOAT or
                            field.type == FieldDescriptor.TYPE_INT32 or
                            field.type == FieldDescriptor.TYPE_INT64 or
                            field.type == FieldDescriptor.TYPE_SFIXED32 or
                            field.type == FieldDescriptor.TYPE_SFIXED64 or
                            field.type == FieldDescriptor.TYPE_SINT32 or
                            field.type == FieldDescriptor.TYPE_SINT64 or
                            field.type == FieldDescriptor.TYPE_SINT64):
                        return "SIGNED_VARINT"
                    elif (field.type == FieldDescriptor.TYPE_UINT32 or
                            field.type == FieldDescriptor.TYPE_UINT64):
                        return "UNSIGNED_VARINT"
                    else:
                        return "STRING"


def get_object_type_key_type(schema_file, object_type):
    """
    Utility function that returns the field type of the
    field annotated with key for the specified object type.
    """
    initialize_descriptor_pool(schema_file)
    return get_key_type_internal(
        descriptor_pool.FindMessageTypeByName(object_type))


def get_human_readable_id(key_type, encoded_id):
    """
    Utility function that returns the human readable form
    of the encoded id.
    """
    if key_type == "UUID":
        left = struct.unpack('Q', encoded_id[:8])[0]
        right = struct.unpack('Q', encoded_id[8:])[0]
        return str(uuid.UUID(int=(left << 64 | right)))
    if key_type == "SIGNED_VARINT":
        pos = 0
        return str(decoder._DecodeSignedVarint(encoded_id, pos)[0])
    if key_type == "UNSIGNED_VARINT":
        pos = 0
        return str(decoder._DecodeVarint(encoded_id, pos)[0])
    try:
        encoded_id = encoded_id.decode()
    except:
        pass
    return encoded_id


def protobuf_uuid_msg_builder(msg):
    val = int(msg.left) << 64 | int(msg.right)

    return str(uuid.UUID(int=val))


def protobuf_ip_addr_msg_builder(msg):
    if (msg.HasField("ipv4")):
        return convert_ipv4_to_string(msg.ipv4)
    elif (msg.HasField("ipv6")):
        return convert_ipv6_bytes_to_string(msg.ipv6)


def protobuf_ip_prefix_msg_builder(msg):
    addr = ""
    prefixlen = 0

    if (msg.HasField("ipv4")):
        addr = convert_ipv4_to_string(msg.ipv4)
    elif (msg.HasField("ipv6")):
        addr = convert_ipv6_bytes_to_string(msg.ipv6)

    if (msg.HasField("prefix_length")):
        prefixlen = msg.prefix_length

    return "%s/%d" % (addr, prefixlen)


def protobuf_ip_subnet_addr_msg_builder(msg):
    prefixlen = 0
    if (msg.HasField("prefix_length")):
        prefixlen = msg.prefix_length

    addrs = []
    for ipv4 in msg.ipv4:
        addrs.append("%s/%d" % (
            convert_ipv4_to_string(ipv4), prefixlen))

    for ipv6 in msg.ipv6:
        addrs.append("%s/%d" % (
            convert_ipv6_bytes_to_string(ipv6), prefixlen))

    return addrs


def protobuf_mac_addr_msg_builder(msg):
    if not msg.mac:
        return str(msg.mac)

    return "{:02x}:{:02x}:{:02x}:{:02x}:{:02x}:{:02x}".format(
               (msg.mac >> 40) & 0xFF,
               (msg.mac >> 32) & 0xFF,
               (msg.mac >> 24) & 0xFF,
               (msg.mac >> 16) & 0xFF,
               (msg.mac >> 8) & 0xFF,
               (msg.mac) & 0xFF)


def protobuf_value_builder(object_instance, field, val, beautify):
    if (field.cpp_type == FieldDescriptor.TYPE_MESSAGE or
            field.cpp_type == FieldDescriptor.TYPE_GROUP):
        if (beautify):
            if (val.DESCRIPTOR.name == "UuidMsg"):
                return protobuf_uuid_msg_builder(val)
            if (val.DESCRIPTOR.name == "IpAddressMsg"):
                return protobuf_ip_addr_msg_builder(val)
            if (val.DESCRIPTOR.name == "IpPrefixMsg"):
                return protobuf_ip_prefix_msg_builder(val)
            if (val.DESCRIPTOR.name == "IpAddressWithPrefixMsg"):
                return protobuf_ip_prefix_msg_builder(val)
            if (val.DESCRIPTOR.name == "IpSubnetAddressMsg"):
                return protobuf_ip_subnet_addr_msg_builder(val)
            if (val.DESCRIPTOR.name == "MacAddressMsg"):
                return protobuf_mac_addr_msg_builder(val)
        return protobuf_dict_builder(val, beautify)
    else:
        if field.type == FieldDescriptor.TYPE_ENUM:
            return str(field.enum_type.values_by_number[val].name)
        elif field.type == FieldDescriptor.TYPE_BYTES:
            try:
                if field.name == "ipv6":
                    val = convert_ipv6_bytes_to_string(val)
                elif ((field.name == "client_id" and
                       "NsxAgentIpDiscoveryVifBindingMsg" in
                       object_instance.DESCRIPTOR.full_name) or
                      (field.name == "headers" and
                       "AmqpHeaderMsg" in
                       object_instance.DESCRIPTOR.full_name) or
                      ("file" in field.name and
                       "IdsSignaturesMsg" in
                       object_instance.DESCRIPTOR.full_name)):
                    val = base64.b64encode(val).decode()
                else:
                    val = val.decode()
            except:
                pass
        return val


def protobuf_dict_builder(object_instance, beautify=False):
    """
    Utility function that traverses the populated fields
    of the protobuf object and returns a dict representation.
    The key of the dict is the protobuf field name and the
    value is the protobuf field value.
    """
    object_dict = {}
    fields = object_instance.ListFields()
    for (field, value) in fields:
        if field.label == FieldDescriptor.LABEL_REPEATED:
            object_list = []
            object_dict[field.name] = object_list
            for val in value:
                object_list.append(protobuf_value_builder(object_instance,
                                                          field, val,
                                                          beautify))
        else:
            if ((field.name != "file_content")):
                object_dict[field.name] = protobuf_value_builder(
                    object_instance, field, value, beautify)

    return object_dict


def populate_proto_object(json_object, proto_object):
    """
    Utility function that traverses the json object and
    populates the protobuf object. The key in the json
    object corresponds to the field name in the protobuf
    object.
    """
    if not isinstance(json_object, dict):
        return False
    for k, v, in list(json_object.items()):
        try:
            if type(v) is dict:
                if not populate_proto_object(v, getattr(proto_object, k)):
                    return False
            elif type(v) is list:
                index = 0
                for e_v in v:
                    if type(e_v) is dict:
                        if not populate_proto_object(
                                e_v, getattr(proto_object, k).add()):
                            return False
                    else:
                        proto_attr = getattr(proto_object, k)
                        if (proto_attr is not None and
                                isinstance(proto_attr, Message) and
                                proto_attr._message_descriptor.name ==
                                "UuidMsg"):
                            u = uuid.UUID(e_v)
                            e_v = {"left": u.int >> 64,
                                   "right": u.int & ((1 << 64) - 1)}
                            if not populate_proto_object(e_v,
                                                         proto_attr.add()):
                                return False
                            continue

                        if k == "ipv6":
                            try:
                                e_v = convert_ipv6_string_to_bytes(e_v)
                            except:
                                pass
                        # JSON input does not support bytes and PB don't
                        # support string. Making sure if the PB schema says it
                        # is bytes and we have received string we make the
                        # conversion for rpcclient to accept the input.
                        elif (proto_object.DESCRIPTOR.fields_by_name[k].type ==
                                FieldDescriptor.TYPE_BYTES):
                            try:
                                if (type(e_v) is unicode):
                                    e_v = str(e_v)
                            except:
                                e_v = e_v.encode()
                        try:
                            getattr(proto_object, k).append(e_v)
                        except TypeError as te:
                            if (type(e_v) is not str and
                                    type(e_v) is not unicode):
                                raise AttributeError(te)
                            # If the user gave a string value, attempt to find
                            # the corresponding enum value (if the key is an
                            # enum type).
                            fields = proto_object.DESCRIPTOR.fields_by_name
                            if k not in fields:
                                raise AttributeError(te)
                            enum_type = fields[k].enum_type
                            if not enum_type:
                                raise AttributeError(te)
                            if e_v not in enum_type.values_by_name:
                                raise AttributeError(te)
                            getattr(proto_object, k).append(
                                    enum_type.values_by_name[e_v].number)
                    index = index + 1
            else:
                try:
                    proto_attr = getattr(proto_object, k)
                    if (proto_attr is not None and
                            isinstance(proto_attr, Message) and
                            proto_attr.DESCRIPTOR.name == "UuidMsg"):
                        u = uuid.UUID(v)
                        v = {"left": u.int >> 64,
                             "right": u.int & ((1 << 64) - 1)}
                        if not populate_proto_object(v, proto_attr):
                            return False
                        continue

                    if k == "ipv6":
                        try:
                            v = convert_ipv6_string_to_bytes(v)
                        except:
                            pass
                    elif ((k == "client_id" and
                           "NsxAgentIpDiscoveryVifBindingMsg" in
                           proto_object.DESCRIPTOR.full_name) or
                          (k == "headers" and
                           "AmqpHeaderMsg" in
                           proto_object.DESCRIPTOR.full_name) or
                          ("file" in k and
                           "IdsSignaturesMsg" in
                           proto_object.DESCRIPTOR.full_name)):
                        try:
                            v = base64.b64decode(v)
                        except:
                            pass
                    elif (proto_object.DESCRIPTOR.fields_by_name[k].type ==
                            FieldDescriptor.TYPE_BYTES):
                        try:
                            if (type(v) is unicode):
                                v = str(v)
                        except:
                            v = v.encode()
                    setattr(proto_object, k, v)
                except TypeError as te:
                    if (type(v) is not str and
                            type(v) is not unicode):
                        raise AttributeError(te)
                    # If the user gave a string value, attempt to find the
                    # corresponding enum value (if the key is an enum type).
                    fields = proto_object.DESCRIPTOR.fields_by_name
                    if k not in fields:
                        raise AttributeError(te)
                    enum_type = fields[k].enum_type
                    if not enum_type:
                        raise AttributeError(te)
                    if v not in enum_type.values_by_name:
                        raise AttributeError(te)
                    setattr(proto_object, k,
                            enum_type.values_by_name[v].number)
        except AttributeError as e:
            print(e)
            return False
    return True


def load_jsons(json_strs):
    """
    Utility function that parses multiple top-level json
    objects in the same input string.
    """
    decoder = json.JSONDecoder()
    idx = json.decoder.WHITESPACE.match(json_strs, 0).end()
    end = len(json_strs)
    try:
        while idx != end:
            (val, idx) = decoder.raw_decode(json_strs, idx=idx)
            yield val
            idx = json.decoder.WHITESPACE.match(json_strs, idx).end()
    except ValueError as exc:
        raise ValueError('%s (%r at position %d).' % (exc, json_strs[idx:],
                                                      idx))


#
#  Functions for processing CLI commands.
#


def run_schema_command_impl(cmd, nsxRpcClient, json_output):
    rv = '""'

    if len(cmd) < 1:
        return False, rv
    cli_output = {}
    if 'load' in cmd[0] and len(cmd) >= 2:
        msg = nestdb_pb2.LoadSchemaRequestMsg()
        msg.schema_file = cmd[1]

        try:
            rsp = nsxRpcClient.LoadSchema(msg, DEFAULT_TIMEOUT)
        except NsxRpcCallError as e:
            sys.stderr.write("NsxRpcClient encountered an error: "
                             "%s, code: %s, message: %s.\n" %
                             (e.name, e.code, e.message))
            return False, rv

        if rsp.success:
            if json_output is False:
                print(("Successfully loaded %s" % cmd[1]))
            else:
                cli_output["status"] = "Successfully loaded %s" % cmd[1]
        else:
            if json_output is False:
                print("Schema not loaded")
            else:
                cli_output["status"] = "Schema not loaded"
    elif 'show' in cmd[0]:
        msg = nestdb_pb2.GetSchemaRequestMsg()

        try:
            rsp = nsxRpcClient.GetSchema(msg, DEFAULT_TIMEOUT)
        except NsxRpcCallError as e:
            sys.stderr.write("NsxRpcClient encountered an error: "
                             "%s, code: %s, message: %s.\n" %
                             (e.name, e.code, e.message))
            return False, rv

        if json_output is False:
            print(("Schema file: %s" % rsp.schema_file))
            print(("Schema: %s" % rsp.schema))
        else:
            cli_output["file"] = rsp.schema_file
            cli_output["schema"] = rsp.schema
    elif 'version' in cmd[0]:
        msg = nestdb_pb2.GetSchemaVersionRequestMsg()

        try:
            rsp = nsxRpcClient.GetSchemaVersion(msg, DEFAULT_TIMEOUT)
        except NsxRpcCallError as e:
            sys.stderr.write("NsxRpcClient encountered an error: "
                             "%s, code: %s, message: %s.\n" %
                             (e.name, e.code, e.message))
            return False, rv

        if json_output is False:
            print(("Schema version: %s" % rsp.version))
        else:
            cli_output["version"] = rsp.version
    else:
        return False, rv
    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return True, rv


# json_output has three states
#   None (default): only return json without print (used in CLI function call)
#   True: print and return json
#   False: print result in text
def run_schema_command(cmd, json_output=None, endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDbOps_Stub, endpoint) as nsxRpcClient:
            return run_schema_command_impl(cmd, nsxRpcClient, json_output)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_get_command_impl(cmd, nsxRpcClient, json_output, beautify,
                         schema_file):
    rv = '""'

    if len(cmd) < 1:
        return False, rv
    msg = nestdb_pb2.GetRequestMsg()
    requested_object = msg.object.add()
    try:
        msg_type = find_proto_msg_type(schema_file, cmd[0])
        object_type = get_object_type(schema_file,
                                      extract_object_type_enum_name(msg_type))
    except:
        sys.stderr.write("Invalid object_type: %s\n" % cmd[0])
        return False, rv

    if len(cmd) >= 3:
        try:
            msg.vdb_id = int(cmd[2])
        except ValueError:
            sys.stderr.write("Invalid vdb_id: %s\n" % cmd[2])
            return False, rv

    requested_object.object_type = object_type
    if len(cmd) >= 2:
        parsed_vdb_id = False
        # If exactly 2 parameters were given, check if the second parameter
        # is a vdb_id.
        if len(cmd) == 2:
            try:
                vdb_id = int(cmd[1])
                msg.vdb_id = vdb_id
                parsed_vdb_id = True
            except ValueError:
                pass

        if not parsed_vdb_id:
            try:
                json_object = json.loads(cmd[1])
            except json.scanner.JSONDecodeError as e:
                sys.stderr.write(
                    "Unable to decode json object from input. %s\n" % e.msg)
                return False, rv
            object_instance = get_proto_object(schema_file, msg_type)

            if not populate_proto_object(json_object, object_instance):
                return False, rv

            requested_object.embedded_id = object_instance.SerializeToString()

    client_context = nsxRpcClient.CreateClientContext()
    client_context.SetClientId(client_id)

    try:
        call = nsxRpcClient.Get(msg, DEFAULT_TIMEOUT, client_context)
    except NsxRpcCallError as e:
        sys.stderr.write("NsxRpcClient encountered an error: "
                         "%s, code: %s, message: %s.\n" %
                         (e.name, e.code, e.message))
        return False, rv

    cli_output = {}
    cli_output["result"] = []
    for rsp in call.responseList:
        for typed_object in rsp.object:
            if (json_output is False and
                    not typed_object.HasField('object_type')):
                print("Result is missing object_type")
                continue

            msg_type = get_msg_type_from_object_type(
                get_object_type_val(schema_file, typed_object.object_type))

            if (json_output is False and
                    not typed_object.HasField('object_body')):
                print(("Did not find any %s objects" % msg_type))
                continue

            object_instance = get_proto_object(schema_file, msg_type)
            object_instance.ParseFromString(typed_object.object_body)
            object_instance = protobuf_dict_builder(object_instance, beautify)
            if json_output is False:
                print(("Object type: %s" % msg_type))
                obj_str = str(object_instance)
                print(("\t%s" % obj_str))
            else:
                object_dict = {}
                object_dict["object_type"] = msg_type
                object_dict["value"] = object_instance
                cli_output["result"].append(object_dict)
    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return True, rv


def run_get_command(cmd, json_output=None, beautify=False,
                    schema_file=DEFAULT_SCHEMA,
                    endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_get_command_impl(
                cmd, nsxRpcClient, json_output, beautify, schema_file)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_put_command_impl(cmd, nsxRpcClient, json_output, schema_file):
    rv = '""'

    if len(cmd) < 2:
        return False, rv
    msg = nestdb_pb2.PutRequestMsg()
    put_msg = msg.put.add()

    try:
        json_object = json.loads(cmd[1])
    except json.scanner.JSONDecodeError as e:
        sys.stderr.write("Unable to decode json object from input. %s\n" %
                         e.msg)
        return False, rv

    try:
        msg_type = find_proto_msg_type(schema_file, cmd[0])
        object_type = get_object_type(schema_file,
                                      extract_object_type_enum_name(msg_type))
    except:
        sys.stderr.write("Invalid object_type: %s\n" % cmd[0])
        return False, rv

    object_instance = get_proto_object(schema_file, msg_type)

    if not populate_proto_object(json_object, object_instance):
        return False, rv

    put_msg.object.object_type = object_type
    put_msg.object.object_body = object_instance.SerializeToString()

    put_msg.vdb_id.extend(cmd[2:])

    client_context = nsxRpcClient.CreateClientContext()
    client_context.SetClientId(client_id)

    cli_output = {}
    has_error = False

    try:
        writer, reader = nsxRpcClient.Put(None, DEFAULT_TIMEOUT,
                                          client_context)
        writer.next(msg)
        writer.done()
        rsp = next(reader.responseList)
    except NsxRpcCallError as e:
        error_string = ("Unable to put the object into NestDB "
                        "(error: %s, code: %s, message: %s)" %
                        (e.name, e.code, e.message))
        has_error = True
        if json_output is False:
            print(error_string)
        else:
            cli_output["status"] = error_string

    if not has_error:
        if json_output is False:
            print("Successfully put the object into NestDB")
        else:
            cli_output["status"] = "Successfully put the object into NestDB"

    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return not has_error, rv


def run_put_command(cmd, json_output=None, schema_file=DEFAULT_SCHEMA,
                    endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_put_command_impl(
                cmd, nsxRpcClient, json_output, schema_file)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_delete_command_impl(cmd, nsxRpcClient, json_output, schema_file):
    rv = '""'

    if len(cmd) < 2:
        return False, rv
    msg = nestdb_pb2.DeleteRequestMsg()
    delete_msg = msg.delete.add()

    try:
        msg_type = find_proto_msg_type(schema_file, cmd[0])
        object_type = get_object_type(schema_file,
                                      extract_object_type_enum_name(msg_type))
    except:
        sys.stderr.write("Invalid object_type: %s\n" % cmd[0])
        return False, rv

    try:
        json_object = json.loads(cmd[1])
    except json.scanner.JSONDecodeError as e:
        sys.stderr.write("Unable to decode json object from input. %s\n" %
                         e.msg)
        return False, rv

    id_instance = get_proto_object(schema_file, msg_type)

    if not populate_proto_object(json_object, id_instance):
        return False, rv

    delete_msg.object.object_type = object_type
    delete_msg.object.embedded_id = id_instance.SerializeToString()

    delete_msg.vdb_id.extend(cmd[2:])

    client_context = nsxRpcClient.CreateClientContext()
    client_context.SetClientId(client_id)

    cli_output = {}
    has_error = False

    try:
        writer, reader = nsxRpcClient.Delete(None, DEFAULT_TIMEOUT,
                                             client_context)
        writer.next(msg)
        writer.done()
        rsp = next(reader.responseList)
    except NsxRpcCallError as e:
        error_string = ("Unable to delete the object from NestDB "
                        "(error: %s, code: %s, message: %s)" %
                        (e.name, e.code, e.message))
        has_error = True
        if json_output is False:
            print(error_string)
        else:
            cli_output["status"] = error_string

    if not has_error:
        if json_output is False:
            print("Successfully delete the object from NestDB")
        else:
            cli_output["status"] = "Successfully delete the object from NestDB"

    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return not has_error, rv


def run_delete_command(cmd, json_output=None, schema_file=DEFAULT_SCHEMA,
                       endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_delete_command_impl(cmd, nsxRpcClient, json_output,
                                           schema_file)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_flush_command_impl(cmd, nsxRpcClient, json_output, schema_file):
    rv = '""'

    if len(cmd) < 1:
        return False, rv
    msg = nestdb_pb2.FlushRequestMsg()
    flush_msg = msg.flush.add()

    try:
        msg_type = find_proto_msg_type(schema_file, cmd[0])
        object_type = get_object_type(schema_file,
                                      extract_object_type_enum_name(msg_type))
    except:
        sys.stderr.write("Invalid object_type: %s\n" % cmd[0])
        return False, rv

    flush_msg.object_type = object_type

    flush_msg.vdb_id.extend(cmd[1:])

    client_context = nsxRpcClient.CreateClientContext()
    client_context.SetClientId(client_id)

    cli_output = {}
    has_error = False

    try:
        writer, reader = nsxRpcClient.Flush(None, DEFAULT_TIMEOUT,
                                            client_context)
        writer.next(msg)
        writer.done()
        rsp = next(reader.responseList)
    except NsxRpcCallError as e:
        error_string = ("Unable to flush the %s from NestDB "
                        "(error: %s, code: %s, message: %s)" %
                        (msg_type, e.name, e.code, e.message))
        has_error = True
        if json_output is False:
            print(error_string)
        else:
            cli_output["status"] = error_string

    if not has_error:
        if json_output is False:
            print(("Successfully flushed %s from NestDB" % msg_type))
        else:
            cli_output["status"] = (
                "Successfully flushed %s from NestDB" % msg_type)

    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return not has_error, rv


def run_flush_command(cmd, json_output=None, schema_file=DEFAULT_SCHEMA,
                      endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_flush_command_impl(cmd, nsxRpcClient, json_output,
                                          schema_file)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_copy_command_impl(cmd, nsxRpcClient, json_output, schema_file):
    rv = '""'
    if len(cmd) < 4:
        return False, rv
    msg = nestdb_pb2.CopyRequestMsg()
    copy_msg = msg.copy.add()

    try:
        msg_type = find_proto_msg_type(schema_file, cmd[0])
        object_type = get_object_type(schema_file,
                                      extract_object_type_enum_name(msg_type))
    except:
        sys.stderr.write("Invalid object_type: %s\n" % cmd[0])
        return False, rv

    try:
        json_object = json.loads(cmd[1])
    except json.scanner.JSONDecodeError as e:
        sys.stderr.write("Unable to decode json object from input. %s\n" %
                         e.msg)
        return False, rv

    id_instance = get_proto_object(schema_file, msg_type)

    if not populate_proto_object(json_object, id_instance):
        return False, rv

    copy_msg.object.object_type = object_type
    copy_msg.object.embedded_id = id_instance.SerializeToString()

    if cmd[2] is not None:
        copy_msg.src_vdb_id = cmd[2]

    copy_msg.dst_vdb_id.extend(cmd[3])

    client_context = nsxRpcClient.CreateClientContext()
    client_context.SetClientId(client_id)

    cli_output = {}
    has_error = False

    try:
        writer, reader = nsxRpcClient.Copy(None, DEFAULT_TIMEOUT,
                                           client_context)
        writer.next(msg)
        writer.done()
        rsp = next(reader.responseList)
    except NsxRpcCallError as e:
        error_string = ("Unable to delete the object from NestDB "
                        "(error: %s, code: %s, message: %s)" %
                        (e.name, e.code, e.message))
        has_error = True
        if json_output is False:
            print(error_string)
        else:
            cli_output["status"] = error_string

    if not has_error:
        if json_output is False:
            print("Successfully copied the object in NestDB")
        else:
            cli_output["status"] = "Successfully copied the object in NestDB"

    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return not has_error, rv


def run_copy_command(cmd, json_output=None, schema_file=DEFAULT_SCHEMA,
                     endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_copy_command_impl(cmd, nsxRpcClient, json_output,
                                         schema_file)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_query_command_impl(cmd, nsxRpcClient, json_output, beautify,
                           schema_file):
    rv = '""'

    if len(cmd) < 4:
        return False, rv
    msg = nestdb_pb2.QueryRequestMsg()

    try:
        json_op_object = json.loads(cmd[1].replace('\'', '"'))
        json_val_object = json.loads(cmd[2].replace('\'', '"'))
    except json.scanner.JSONDecodeError as e:
        sys.stderr.write("Unable to decode json object from input. %s\n" %
                         e.msg)
        return False, rv

    try:
        msg_type = find_proto_msg_type(schema_file, cmd[0])
        object_type = get_object_type(schema_file,
                                      extract_object_type_enum_name(msg_type))
    except:
        sys.stderr.write("Invalid object_type: %s\n" % cmd[0])
        return False, rv

    op_object_instance = get_proto_object(schema_file,
                                          msg_type.replace('Msg', 'OpMsg'))
    val_object_instance = get_proto_object(schema_file, msg_type)

    if not populate_proto_object(json_val_object, val_object_instance):
        return False, rv

    if not populate_proto_object(json_op_object, op_object_instance):
        return False, rv

    if len(cmd) > 4:
        msg.vdb_id = cmd[4]

    msg.object.object_type = object_type
    where = msg.object.where.add()
    where.op_msg = op_object_instance.SerializeToString()
    where.value = val_object_instance.SerializeToString()

    if cmd[3] is not None:
        try:
            json_sort_object = json.loads(cmd[3].replace('\'', '"'))
        except json.scanner.JSONDecodeError as e:
            sys.stderr.write("Unable to decode json object from input. %s\n" %
                             e.msg)
            return False, rv

        sort_op_object_instance = get_proto_object(
            schema_file, msg_type.replace('Msg', 'OpMsg'))
        if not populate_proto_object(json_sort_object,
                                     sort_op_object_instance):
            return False, rv
        msg.object.sort_by.op_msg = sort_op_object_instance.SerializeToString()

    client_context = nsxRpcClient.CreateClientContext()
    client_context.SetClientId(client_id)

    try:
        call = nsxRpcClient.Query(msg, DEFAULT_TIMEOUT, client_context)
    except NsxRpcCallError as e:
        sys.stderr.write("NsxRpcClient encountered an error: "
                         "%s, code: %s, message: %s.\n" %
                         (e.name, e.code, e.message))
        return False, rv

    cli_output = {}
    cli_output["result"] = []
    for rsp in call.responseList:
        for typed_object in rsp.object:
            if (json_output is False and
                    not typed_object.HasField('object_type')):
                print("Result is missing object_type")
                continue

            msg_type = get_msg_type_from_object_type(
                get_object_type_val(schema_file, typed_object.object_type))

            if (json_output is False and
                    not typed_object.HasField('object_body')):
                print(("Did not find any %s objects" % msg_type))
                continue

            object_instance = get_proto_object(schema_file, msg_type)
            object_instance.ParseFromString(typed_object.object_body)
            if json_output is False:
                print(("Object type: %s" % msg_type))
                obj_str = str(object_instance).replace('\n', '\n\t')
                print(("\t%s" % obj_str))
            else:
                object_dict = {}
                object_dict["object_type"] = msg_type
                object_dict["value"] = protobuf_dict_builder(object_instance,
                                                             beautify)
                cli_output["result"].append(object_dict)
    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return True, rv


def run_query_command(cmd, json_output=None, beautify=False,
                      schema_file=DEFAULT_SCHEMA,
                      endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_query_command_impl(cmd, nsxRpcClient, json_output,
                                          beautify, schema_file)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_attribute_command_impl(cmd, nsxRpcClient, json_output, beautify,
                               schema_file):
    rv = '""'

    if len(cmd) < 4:
        return False, rv

    try:
        msg_type = find_proto_msg_type(schema_file, cmd[1])
        object_type = get_object_type(schema_file,
                                      extract_object_type_enum_name(msg_type))
    except:
        sys.stderr.write("Invalid object_type: %s\n" % cmd[1])
        return False, rv

    op_object_instance = get_proto_object(schema_file,
                                          msg_type.replace('Msg', 'OpMsg'))
    val_object_instance = get_proto_object(schema_file, msg_type)

    cli_output = {}
    has_error = False

    if cmd[0] == "query":
        if not populate_proto_object(
                create_json_from_field_val(cmd[2], cmd[3]),
                val_object_instance):
            return False, rv
        if not populate_proto_object(
                create_json_from_field_val(cmd[2] + ".op", 'EQUALS'),
                op_object_instance):
            return False, rv

        msg = nestdb_pb2.QueryRequestMsg()

        if len(cmd) > 4:
            try:
                msg.vdb_id = int(cmd[4])
            except ValueError:
                sys.stderr.write("Invalid vdb_id: %s\n" % cmd[4])
                return False, rv

        msg.object.object_type = object_type
        where = msg.object.where.add()
        where.op_msg = op_object_instance.SerializeToString()
        where.value = val_object_instance.SerializeToString()

        client_context = nsxRpcClient.CreateClientContext()
        client_context.SetClientId(client_id)

        try:
            call = nsxRpcClient.Query(msg, DEFAULT_TIMEOUT, client_context)
        except NsxRpcCallError as e:
            sys.stderr.write("NsxRpcClient encountered an error: "
                             "%s, code: %s, message: %s.\n" %
                             (e.name, e.code, e.message))
            return False, rv

        cli_output["result"] = []
        for rsp in call.responseList:
            for typed_object in rsp.object:
                if (json_output is False and
                        not typed_object.HasField('object_type')):
                    print("Result is missing object_type")
                    continue

                msg_type = get_msg_type_from_object_type(
                    get_object_type_val(schema_file, typed_object.object_type))

                if (json_output is False and
                        not typed_object.HasField('object_body')):
                    print(("Did not find any %s objects" % msg_type))
                    continue

                object_instance = get_proto_object(schema_file, msg_type)
                object_instance.ParseFromString(typed_object.object_body)
                if json_output is False:
                    print(("Object type: %s" % msg_type))
                    obj_str = str(object_instance).replace('\n', '\n\t')
                    print(("\t%s" % obj_str))
                else:
                    object_dict = {}
                    object_dict["object_type"] = msg_type
                    object_dict["value"] = protobuf_dict_builder(
                        object_instance, beautify)
                    cli_output["result"].append(object_dict)
    elif cmd[0] == "update":
        if len(cmd) < 5:
            return False, rv

        try:
            id_json_object = json.loads(cmd[2])
        except json.scanner.JSONDecodeError as e:
            sys.stderr.write(
                "Unable to decode json object from input. %s\n" % e.msg)
            return False, rv

        val_json_object = create_json_from_field_val(cmd[3], cmd[4])
        val_json_object.update(id_json_object)

        if not populate_proto_object(
                val_json_object,
                val_object_instance):
            return False, rv
        if not populate_proto_object(
                create_json_from_field_val(cmd[3] + ".op", 'SET'),
                op_object_instance):
            return False, rv

        msg = nestdb_pb2.PatchRequestMsg()
        patch = msg.patch.add()
        patch.object_type = object_type
        patch.op_msg = op_object_instance.SerializeToString()
        patch.value = val_object_instance.SerializeToString()

        try:
            patch.vdb_id.extend([int(x) for x in cmd[5:]])
        except ValueError:
            sys.stderr.write("Invalid vdb_id in vdb_list: %s\n" % cmd[5:])
            return False, rv

        client_context = nsxRpcClient.CreateClientContext()
        client_context.SetClientId(client_id)

        try:
            writer, reader = nsxRpcClient.Patch(None, DEFAULT_TIMEOUT,
                                                client_context)
            writer.next(msg)
            writer.done()
            rsp = next(reader.responseList)
        except NsxRpcCallError as e:
            error_string = ("Unable to update the object in NestDB "
                            "(error: %s, code: %s, message: %s)" %
                            (e.name, e.code, e.message))
            has_error = True
            if json_output is False:
                print(error_string)
            else:
                cli_output["status"] = error_string

        if not has_error:
            if json_output is False:
                print("Successfully updated object in NestDB")
            else:
                cli_output["status"] = "Successfully updated object in NestDB"
    else:
        return False, rv

    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return not has_error, rv


def run_attribute_command(cmd, json_output=None, beautify=False,
                          schema_file=DEFAULT_SCHEMA,
                          endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_attribute_command_impl(cmd, nsxRpcClient, json_output,
                                              beautify, schema_file)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_vdb_command_impl(cmd, nsxRpcClient, json_output):
    rv = '""'

    cli_output = {}
    has_error = False

    if cmd[0] == "create":
        if len(cmd) < 2:
            return False, rv

        msg = nestdb_pb2.CreateVDbRequestMsg()
        msg.vdb_name = cmd[1]

        client_context = nsxRpcClient.CreateClientContext()
        client_context.SetClientId(client_id)
        cli_output["result"] = {}
        try:
            call = nsxRpcClient.CreateVDb(msg, DEFAULT_TIMEOUT, client_context)
            if json_output is False:
                print(("Generated vdb_id: %s" % call.vdb_id))
            else:
                cli_output["result"] = {cmd[1]: call.vdb_id}
        except NsxRpcCallError as e:
            error_string = ("Unable to create virtual database in NestDB "
                            "(error: %s, code: %s, message: %s)" %
                            (e.name, e.code, e.message))
            has_error = True
            if json_output is False:
                print(error_string)
            else:
                cli_output["status"] = error_string

        if not has_error:
            if json_output is not False:
                cli_output["status"] = (
                    "Successfully created virtual database in NestDB")
    elif cmd[0] == "delete":
        if len(cmd) < 2:
            return False, rv

        msg = nestdb_pb2.DeleteVDbRequestMsg()
        msg.vdb_name = cmd[1]

        client_context = nsxRpcClient.CreateClientContext()
        client_context.SetClientId(client_id)
        try:
            call = nsxRpcClient.DeleteVDb(msg, DEFAULT_TIMEOUT, client_context)
        except NsxRpcCallError as e:
            error_string = ("Unable to delete virtual database from NestDB "
                            "(error: %s, code: %s, message: %s)" %
                            (e.name, e.code, e.message))
            has_error = True
            if json_output is False:
                print(error_string)
            else:
                cli_output["status"] = error_string

        if not has_error:
            if json_output is False:
                print("Successfully deleted virtual database from NestDB")
            else:
                cli_output["status"] = (
                    "Successfully deleted virtual database from NestDB")
    elif cmd[0] == "get":
        msg = nestdb_pb2.GetVDbIdRequestMsg()
        if len(cmd) > 1:
            msg.vdb_name.append(cmd[1])

        client_context = nsxRpcClient.CreateClientContext()
        client_context.SetClientId(client_id)
        cli_output["result"] = {}
        try:
            call = nsxRpcClient.GetVDbId(msg, DEFAULT_TIMEOUT, client_context)
            for vdb_msg in call.vdb:
                if json_output is False:
                    print((vdb_msg.vdb_name + ": " + str(vdb_msg.vdb_id)))
                else:
                    cli_output["result"][vdb_msg.vdb_name] = vdb_msg.vdb_id
        except NsxRpcCallError as e:
            error_string = ("Unable to get virtual database from NestDB "
                            "(error: %s, code: %s, message: %s)" %
                            (e.name, e.code, e.message))
            has_error = True
            if json_output is False:
                print(error_string)
            else:
                cli_output["status"] = error_string

        if not has_error:
            if json_output is not False:
                cli_output["status"] = "Success"
    else:
        return False, rv

    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return not has_error, rv


def run_vdb_command(cmd, json_output=None, endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_vdb_command_impl(cmd, nsxRpcClient, json_output)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_certificate_command_impl(cmd, nsxRpcClient, json_output):
    rv = '""'

    cli_output = {}
    has_error = False

    if cmd[0] == "add":
        msg = nestdb_pb2.AddClientCertificateRequestMsg()
        msg.client_id = cmd[1]
        with open(cmd[2], "r") as cert_file:
            msg.certificate = cert_file.read()

        client_context = nsxRpcClient.CreateClientContext()
        client_context.SetClientId(client_id)
        cli_output["result"] = {}
        try:
            call = nsxRpcClient.AddClientCertificate(msg, DEFAULT_TIMEOUT,
                                                     client_context)
        except NsxRpcCallError as e:
            error_string = ("Unable to add certificate in NestDB "
                            "(error: %s, code: %s, message: %s)" %
                            (e.name, e.code, e.message))
            has_error = True
            if json_output is False:
                print(error_string)
            else:
                cli_output["status"] = error_string

        if not has_error:
            if json_output is False:
                print("Successfully added certificate in NestDB")
            else:
                cli_output["status"] = (
                    "Successfully added certificate in NestDB")
    elif cmd[0] == "delete":
        msg = nestdb_pb2.DeleteClientCertificateRequestMsg()
        msg.client_id = cmd[1]

        client_context = nsxRpcClient.CreateClientContext()
        client_context.SetClientId(client_id)
        try:
            call = nsxRpcClient.DeleteClientCertificate(msg, DEFAULT_TIMEOUT,
                                                        client_context)
        except NsxRpcCallError as e:
            error_string = ("Unable to delete certificate from NestDB "
                            "(error: %s, code: %s, message: %s)" %
                            (e.name, e.code, e.message))
            has_error = True
            if json_output is False:
                print(error_string)
            else:
                cli_output["status"] = error_string

        if not has_error:
            if json_output is False:
                print("Successfully deleted certificate from NestDB")
            else:
                cli_output["status"] = (
                    "Successfully deleted certificate from NestDB")
    elif cmd[0] == "get":
        msg = nestdb_pb2.GetClientCertificateRequestMsg()
        msg.client_id = cmd[1]

        client_context = nsxRpcClient.CreateClientContext()
        client_context.SetClientId(client_id)
        try:
            call = nsxRpcClient.GetClientCertificate(msg, DEFAULT_TIMEOUT,
                                                     client_context)
        except NsxRpcCallError as e:
            error_string = ("Unable to get certificate from NestDB "
                            "(error: %s, code: %s, message: %s)" %
                            (e.name, e.code, e.message))
            has_error = True
            if json_output is False:
                print(error_string)
            else:
                cli_output["status"] = error_string

        if not has_error:
            if json_output is False:
                print(("Digest is %s" % call.digest))
            else:
                cli_output["digest"] = call.digest
                cli_output["status"] = "Success"
    else:
        return False, rv

    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return not has_error, rv


def run_certificate_command(cmd, json_output=None, endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDbOps_Stub, endpoint) as nsxRpcClient:
            return run_certificate_command_impl(cmd, nsxRpcClient, json_output)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_merkle_tree_command_impl(cmd, nsxRpcClient, json_output):
    rv = '""'

    if len(cmd) < 1:
        return False, rv

    msg = nestdb_pb2.GetMerkleTreeRequestMsg()
    if len(cmd) >= 2:
        msg.vdb_id = int(cmd[1])

    client_context = nsxRpcClient.CreateClientContext()
    client_context.SetClientId(cmd[0])

    cli_output = {}
    has_error = False

    try:
        call = nsxRpcClient.GetMerkleTree(msg, DEFAULT_TIMEOUT,
                                          client_context)
    except NsxRpcCallError as e:
        error_string = ("Unable to get merkle tree from NestDB "
                        "(error: %s, code: %s, message: %s)" %
                        (e.name, e.code, e.message))
        if json_output is False:
            print(error_string)
        else:
            cli_output["status"] = error_string
        return False, rv

    if json_output is False:
        print(("Merkle Tree is %s" % call.root))
    else:
        cli_output["merkle_tree"] = str(call.root)
        cli_output["status"] = "Success"

    rv = dump_json(cli_output)
    if json_output:
        print(rv)

    return True, rv


def run_merkle_tree_command(cmd, json_output=None, endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_merkle_tree_command_impl(cmd, nsxRpcClient, json_output)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_object_type_command(cmd, json_output=None, schema_file=DEFAULT_SCHEMA):
    rv = '""'
    if len(cmd) < 1:
        return False, rv
    cli_output = {}
    object_types = get_msg_types(schema_file)

    if cmd[0] == "list":
        cli_output["object_types"] = []
        if json_output is False:
            print(('\n'.join(object_types)))
        else:
            cli_output["object_types"] = object_types
    elif cmd[0] == "show":
        if len(cmd) < 2:
            return False, rv
        object_type = cmd[1]
        matched_msg_types = [
            str(x) for x in object_types if object_type in str(x)]

        if len(matched_msg_types) != 1:
            sys.stderr.write("Invalid object type %s.\n" % object_type)
            return False, rv

        try:
            object_type_schema = get_object_type_schema(
                schema_file, matched_msg_types[0])
        except KeyError as e:
            sys.stderr.write(
                "Unable to get schema for the object type %s.\n" % e.message)
            return False, rv
        if json_output is False:
            print(("Object Type: %s" % matched_msg_types[0]))
            print(object_type_schema)
        else:
            cli_output[matched_msg_types[0]] = object_type_schema
    else:
        return False, rv

    if json_output is not False:
        rv = json.dumps(cli_output, indent=2, sort_keys=True)
        if json_output:
            print(rv)
    return True, rv


def unix_time_ms_to_datetime(unix_time):
    return datetime.datetime.utcfromtimestamp(
        float(unix_time)/1000000.0).strftime('%Y-%m-%dT%H:%M:%SZ.%f')


def deserialize_put_msg(schema_file, put_msg, beautify):
    if not put_msg.HasField('object'):
        print("Error: PutMsg has no object")
        return {}
    if not put_msg.object.HasField('object_type'):
        print("Error: PutMsg's object has no object_type")
        return {}
    if not put_msg.object.HasField('object_body'):
        print("Error: PutMsg's object has no object_body")
        return {}
    msg_type = get_msg_type_from_object_type(
        get_object_type_val(schema_file,
                            put_msg.object.object_type))
    object_instance = get_proto_object(schema_file, msg_type)
    object_instance.ParseFromString(put_msg.object.object_body)
    vdb_id = []
    vdb_id.extend(put_msg.vdb_id)
    rv = {'object_type': msg_type,
          'object': protobuf_dict_builder(object_instance, beautify),
          'vdb_id': vdb_id}
    return rv


def deserialize_put_request(schema_file, put_request, beautify):
    sync = False
    if put_request.HasField('sync'):
        sync = put_request.sync
    rv = {'sync': sync, 'objects': []}
    for put_msg in put_request.put:
        put_msg_object = deserialize_put_msg(schema_file, put_msg, beautify)
        if put_msg_object:
            rv['objects'].append(put_msg_object)
    return rv


def deserialize_delete_msg(schema_file, delete_msg, beautify):
    if not delete_msg.HasField('object'):
        print("Error: DeleteMsg has no object")
        return {}
    if not delete_msg.object.HasField('object_type'):
        print("Error: DeleteMsg's object has no object_type")
        return {}
    if not delete_msg.object.HasField('embedded_id'):
        print("Error: DeleteMsg's object has no embedded_id")
        return {}
    delete_dependents = False
    if delete_msg.HasField('delete_dependents'):
        if delete_msg.delete_dependents:
            delete_dependents = True
    msg_type = get_msg_type_from_object_type(
        get_object_type_val(schema_file,
                            delete_msg.object.object_type))
    object_instance = get_proto_object(schema_file, msg_type)
    object_instance.ParseFromString(delete_msg.object.embedded_id)
    vdb_id = []
    vdb_id.extend(delete_msg.vdb_id)
    rv = {'object_type': msg_type,
          'delete_dependents': delete_dependents,
          'embedded_id': protobuf_dict_builder(object_instance, beautify),
          'vdb_id': vdb_id}
    return rv


def deserialize_delete_request(schema_file, delete_request, beautify):
    sync = False
    if delete_request.HasField('sync'):
        sync = delete_request.sync
    rv = {'sync': sync, 'objects': []}
    for delete_msg in delete_request.delete:
        delete_msg_object = deserialize_delete_msg(schema_file, delete_msg,
                                                   beautify)
        if delete_msg_object:
            rv['objects'].append(delete_msg_object)
    return rv


def deserialize_flush_msg(schema_file, flush_msg):
    if not flush_msg.HasField('object_type'):
        print("Error: Missing object_type in flush msg")
        return {}
    msg_type = get_msg_type_from_object_type(
        get_object_type_val(schema_file, flush_msg.object_type))
    vdb_id = []
    vdb_id.extend(flush_msg.vdb_id)
    rv = {'object_type': msg_type,
          'vdb_id': vdb_id}
    return rv


def deserialize_flush_request(schema_file, flush_request):
    sync = False
    if flush_request.HasField('sync'):
        sync = flush_request.sync
    rv = {'sync': sync, 'objects': []}
    for flush_msg in flush_request.flush:
        flush_msg_object = deserialize_flush_msg(schema_file, flush_msg)
        if flush_msg_object:
            rv['objects'].append(flush_msg_object)
    return rv


def deserialize_patch_msg(schema_file, patch_msg, beautify):
    if not patch_msg.HasField('object_type'):
        print("Error: Missing object_type in patch msg")
        return {}
    msg_type = get_msg_type_from_object_type(
        get_object_type_val(schema_file, patch_msg.object_type))
    object_instance = get_proto_object(schema_file, msg_type)
    if not patch_msg.HasField('op_msg'):
        print("Error: Missing op_msg in patch msg")
        return {}
    if not patch_msg.HasField('value'):
        print("Error: Missing value in patch msg")
        return {}
    object_instance.ParseFromString(patch_msg.value)
    op_object_instance = get_proto_object(
        schema_file, msg_type.replace('Msg', 'OpMsg'))
    op_object_instance.ParseFromString(patch_msg.op_msg)
    vdb_id = []
    vdb_id.extend(patch_msg.vdb_id)
    return {'object_type': msg_type,
            'op_msg': protobuf_dict_builder(op_object_instance, beautify),
            'value': protobuf_dict_builder(object_instance, beautify),
            'vdb_id': vdb_id}


def deserialize_patch_request(schema_file, patch_request, beautify):
    sync = False
    if patch_request.HasField('sync'):
        sync = patch_request.sync
    rv = {'sync': sync, 'patches': []}
    for patch_msg in patch_request.patch:
        patch_msg_object = deserialize_patch_msg(
            schema_file, patch_msg, beautify)
        if patch_msg_object:
            rv['patches'].append(patch_msg_object)
    return rv


def deserialize_copy_msg(schema_file, copy_msg, beautify):
    if not copy_msg.HasField('object'):
        print("Error: CopyMsg has no object")
        return {}
    if not copy_msg.object.HasField('object_type'):
        print("Error: CopyMsg's object has no object_type")
        return {}
    if not copy_msg.object.HasField('embedded_id'):
        print("Error: CopyMsg's object has no embedded_id")
        return {}
    msg_type = get_msg_type_from_object_type(
        get_object_type_val(schema_file,
                            copy_msg.object.object_type))
    object_instance = get_proto_object(schema_file, msg_type)
    object_instance.ParseFromString(copy_msg.object.embedded_id)
    dst_vdb_id = []
    dst_vdb_id.extend(copy_msg.dst_vdb_id)
    rv = {'object_type': msg_type,
          'embedded_id': protobuf_dict_builder(object_instance,
                                               beautify),
          'dst_vdb_id': dst_vdb_id}
    if copy_msg.HasField('src_vdb_id'):
        rv['src_vdb_id'] = copy_msg.src_vdb_id
    return rv


def deserialize_copy_request(schema_file, copy_request, beautify):
    sync = False
    if copy_request.HasField('sync'):
        sync = copy_request.sync
    rv = {'sync': sync, 'objects': []}
    for copy_msg in copy_request.copy:
        copy_msg_object = deserialize_copy_msg(schema_file, copy_msg,
                                               beautify)
        if copy_msg_object:
            rv['objects'].append(copy_msg_object)
    return rv


def deserialize_flush_by_block_id_msg(schema_file, flush_msg):
    if not flush_msg.HasField('block_id'):
        print("Error: Missing block_id in flush_by_block_id msg")
        return {}
    vdb_id = []
    vdb_id.extend(flush_msg.vdb_id)
    rv = {'block_id': flush_msg.block_id,
          'vdb_id': vdb_id}
    return rv


def deserialize_flush_by_block_id_request(schema_file, flush_request):
    sync = False
    if flush_request.HasField('sync'):
        sync = flush_request.sync
    rv = {'sync': sync, 'objects': []}
    for flush_msg in flush_request.flush:
        flush_msg_object = deserialize_flush_by_block_id_msg(
            schema_file, flush_msg)
        if flush_msg_object:
            rv['objects'].append(flush_msg_object)
    return rv


def deserialize_modify_request(schema_file, modify_request, beautify):
    sync = False
    if modify_request.HasField('sync'):
        sync = modify_request.sync
    rv = {'sync': sync, 'modify': []}

    transaction = None
    if modify_request.HasField('transaction'):
        transaction = {}
        if modify_request.transaction.HasField('transaction_id'):
            transaction['transaction_id'] = (
                modify_request.transaction.transaction_id)
        if modify_request.transaction.HasField('transaction_end'):
            transaction['transaction_end'] = (
                modify_request.transaction.transaction_end)
        rv['transaction'] = transaction

    for modify_msg in modify_request.modify:
        if modify_msg.HasField('put'):
            rv['modify'].append(
                {'put': deserialize_put_msg(schema_file,
                                            modify_msg.put, beautify)})
        elif modify_msg.HasField('delete'):
            rv['modify'].append(
                {'delete': deserialize_delete_msg(schema_file,
                                                  modify_msg.delete,
                                                  beautify)})
        elif modify_msg.HasField('flush'):
            rv['modify'].append(
                {'flush': deserialize_flush_msg(schema_file,
                                                modify_msg.flush)})
        elif modify_msg.HasField('patch'):
            rv['modify'].append(
                {'patch': deserialize_patch_msg(schema_file,
                                                modify_msg.patch, beautify)})
        elif modify_msg.HasField('copy'):
            rv['modify'].append(
                {'copy': deserialize_copy_msg(schema_file, modify_msg.copy,
                                              beautify)})
        elif modify_msg.HasField('flush_by_block_id'):
            rv['modify'].append(
                {'flush_by_block_id': deserialize_flush_by_block_id_msg(
                     schema_file, modify_msg.flush_by_block_id)})
    return rv


def deserialize_create_vdb_request(schema_file, create_vdb_request):
    if not create_vdb_request.HasField('vdb_name'):
        print("Error: CreateVDbRequest has no vdb_name")
        return {}
    rv = {'vdb_name': create_vdb_request.vdb_name}
    return rv


def deserialize_delete_vdb_request(schema_file, delete_vdb_request):
    if not delete_vdb_request.HasField('vdb_name'):
        print("Error: DeleteVDbRequest has no vdb_name")
        return {}
    rv = {'vdb_name': delete_vdb_request.vdb_name}
    return rv


def get_transaction_logs(nsxRpcClient):
    rv = '""'

    msg = nestdb_pb2.GetTransactionLogsRequestMsg()
    client_context = nsxRpcClient.CreateClientContext()
    client_context.SetClientId(client_id)
    call = nsxRpcClient.GetTransactionLogs(msg, DEFAULT_TIMEOUT,
                                           client_context)
    responses = []
    for rsp in call.responseList:
        responses.append(rsp)
    return responses


def run_transaction_log_command_impl(cmd, responses, json_output, beautify,
                                     schema_file):
    rv = '""'

    limit_size_enforced = False
    limit_size = 0
    client_ids = []
    object_types = []
    object_ids = []
    vdb_ids = []

    index = 0
    while index + 1 < len(cmd):
        if 'limit_size' in cmd[index]:
            limit_size = int(cmd[index + 1])
            limit_size_enforced = True
        elif 'client' in cmd[index]:
            client_ids.append(cmd[index + 1])
        elif 'object_type' in cmd[index]:
            object_types.append(cmd[index + 1])
        elif 'object_id' in cmd[index]:
            object_ids.append(cmd[index + 1])
        elif 'vdb_id' in cmd[index]:
            try:
                vdb_ids.append(int(cmd[index + 1]))
            except ValueError:
                sys.stderr.write("Invalid vdb_id provided %s" % cmd[index + 1])
                return False, rv
        index = index + 2

    received_txlogs = {}
    received_txlogs['txlog'] = []
    include = True
    txlog = {}
    for rsp in responses:
        for txlog_unit_serialized in rsp.txn_log_unit:
            txlog_unit = nestdb_pb2.TransactionLogUnitMsg()
            txlog_unit.ParseFromString(txlog_unit_serialized)
            if txlog_unit.HasField('unix_time'):
                txlog['timestamp'] = unix_time_ms_to_datetime(
                    txlog_unit.unix_time)
            if txlog_unit.HasField('client_id'):
                txlog['client_id'] = txlog_unit.client_id
                if client_ids and txlog['client_id'] not in client_ids:
                    include = False
            if txlog_unit.HasField('transaction_id'):
                txlog['transaction_id'] = txlog_unit.transaction_id
            if 'vdb_id' not in txlog:
                txlog['vdb_id'] = {}
            for vdb_id in txlog_unit.vdb_id:
                if vdb_id not in txlog['vdb_id']:
                    txlog['vdb_id'][vdb_id] = {}
                if 'updated_objects' not in txlog['vdb_id'][vdb_id]:
                    txlog['vdb_id'][vdb_id]['updated_objects'] = {}
                if 'deleted_objects' not in txlog['vdb_id'][vdb_id]:
                    txlog['vdb_id'][vdb_id]['deleted_objects'] = {}
                if ('idempotent_updated_objects' not in
                        txlog['vdb_id'][vdb_id]):
                    txlog['vdb_id'][vdb_id][
                        'idempotent_updated_objects'] = {}
            for objects_bunch in txlog_unit.updated:
                if not objects_bunch.HasField('object_type'):
                    print("Updated objects msg is missing object_type")
                    continue
                for vdb_id in txlog_unit.vdb_id:
                    msg_type = get_msg_type_from_object_type(
                        get_object_type_val(schema_file,
                                            objects_bunch.object_type))
                    if (msg_type not in
                            txlog['vdb_id'][vdb_id]['updated_objects']):
                        txlog['vdb_id'][vdb_id]['updated_objects'][
                            msg_type] = []
                    for object_id in objects_bunch.object_id:
                        readable_id = get_human_readable_id(
                            get_object_type_key_type(schema_file, msg_type),
                            object_id)
                        if readable_id not in txlog['vdb_id'][vdb_id][
                                'updated_objects'][msg_type]:
                            txlog['vdb_id'][vdb_id]['updated_objects'][
                                msg_type].append(readable_id)
            for objects_bunch in txlog_unit.deleted:
                if not objects_bunch.HasField('object_type'):
                    print("Updated objects msg is missing object_type")
                    continue
                for vdb_id in txlog_unit.vdb_id:
                    msg_type = get_msg_type_from_object_type(
                        get_object_type_val(schema_file,
                                            objects_bunch.object_type))
                    if (msg_type not in
                            txlog['vdb_id'][vdb_id]['deleted_objects']):
                        txlog['vdb_id'][vdb_id]['deleted_objects'][
                            msg_type] = []
                    for object_id in objects_bunch.object_id:
                        readable_id = get_human_readable_id(
                            get_object_type_key_type(schema_file, msg_type),
                            object_id)
                        if readable_id not in txlog['vdb_id'][vdb_id][
                                'deleted_objects'][msg_type]:
                            txlog['vdb_id'][vdb_id]['deleted_objects'][
                                msg_type].append(readable_id)
            for objects_bunch in txlog_unit.idempotent:
                if not objects_bunch.HasField('object_type'):
                    print("Updated objects msg is missing object_type")
                    continue
                for vdb_id in txlog_unit.vdb_id:
                    msg_type = get_msg_type_from_object_type(
                        get_object_type_val(schema_file,
                                            objects_bunch.object_type))
                    if (msg_type not in txlog['vdb_id'][vdb_id][
                            'idempotent_updated_objects']):
                        txlog['vdb_id'][vdb_id]['idempotent_updated_objects'][
                            msg_type] = []
                    for object_id in objects_bunch.object_id:
                        readable_id = get_human_readable_id(
                            get_object_type_key_type(schema_file, msg_type),
                            object_id)
                        if readable_id not in txlog['vdb_id'][vdb_id][
                                'idempotent_updated_objects'][msg_type]:
                            txlog['vdb_id'][vdb_id][
                                'idempotent_updated_objects'][msg_type].append(
                                    readable_id)
            if txlog_unit.HasField('put_request'):
                if 'requests' not in txlog:
                    txlog['requests'] = []
                txlog['requests'].append({'put': deserialize_put_request(
                    schema_file, txlog_unit.put_request, beautify)})
            if txlog_unit.HasField('delete_request'):
                if 'requests' not in txlog:
                    txlog['requests'] = []
                txlog['requests'].append({'delete': deserialize_delete_request(
                    schema_file, txlog_unit.delete_request, beautify)})
            if txlog_unit.HasField('flush_request'):
                if 'requests' not in txlog:
                    txlog['requests'] = []
                txlog['requests'].append({'flush': deserialize_flush_request(
                    schema_file, txlog_unit.flush_request)})
            if txlog_unit.HasField('patch_request'):
                if 'requests' not in txlog:
                    txlog['requests'] = []
                txlog['requests'].append({'patch': deserialize_patch_request(
                    schema_file, txlog_unit.patch_request, beautify)})
            if txlog_unit.HasField('modify_request'):
                if 'requests' not in txlog:
                    txlog['requests'] = []
                txlog['requests'].append({'modify': deserialize_modify_request(
                    schema_file, txlog_unit.modify_request, beautify)})
            if txlog_unit.HasField('copy_request'):
                if 'requests' not in txlog:
                    txlog['requests'] = []
                txlog['requests'].append({'copy': deserialize_copy_request(
                    schema_file, txlog_unit.copy_request, beautify)})
            if txlog_unit.HasField('flush_by_block_id_request'):
                if 'requests' not in txlog:
                    txlog['requests'] = []
                txlog['requests'].append({
                    'flush_by_block_id': deserialize_flush_by_block_id_request(
                        schema_file, txlog_unit.flush_by_block_id_request)})
            if txlog_unit.HasField('create_vdb_request'):
                if 'requests' not in txlog:
                    txlog['requests'] = []
                txlog['requests'].append(
                    {'create_vdb': deserialize_create_vdb_request(
                        schema_file, txlog_unit.create_vdb_request)})
            if txlog_unit.HasField('delete_vdb_request'):
                if 'requests' not in txlog:
                    txlog['requests'] = []
                txlog['requests'].append(
                    {'delete_vdb': deserialize_delete_vdb_request(
                        schema_file, txlog_unit.delete_vdb_request)})
            if txlog_unit.HasField('transaction_end'):
                assert txlog_unit.transaction_end
                if include and vdb_ids:
                    if set(vdb_ids).isdisjoint(list(txlog['vdb_id'].keys())):
                        include = False
                if include:
                    if object_types:
                        matched_obj_types = set()
                        for vdb_id in txlog['vdb_id']:
                            matched_obj_types.update(
                                list(txlog['vdb_id'][vdb_id][
                                    'updated_objects'].keys()))
                            matched_obj_types.update(
                                list(txlog['vdb_id'][vdb_id][
                                    'idempotent_updated_objects'].keys()))
                            matched_obj_types.update(
                                list(txlog['vdb_id'][vdb_id][
                                    'deleted_objects'].keys()))
                        matched_obj_types_copy = set(matched_obj_types)
                        for matched_obj_type in matched_obj_types_copy:
                            matched_obj_types.add(
                                matched_obj_type.split(".")[-1])
                        if set(object_types).isdisjoint(matched_obj_types):
                            include = False
                        elif object_ids:
                            ids = []
                            for obj_type in object_types:
                                for vdb_id in txlog['vdb_id']:
                                    if obj_type in txlog['vdb_id'][vdb_id][
                                            'updated_objects']:
                                        ids.extend(
                                            txlog['vdb_id'][vdb_id][
                                                'updated_objects'][obj_type])
                                    if obj_type in txlog['vdb_id'][vdb_id][
                                            'deleted_objects']:
                                        ids.extend(
                                            txlog['vdb_id'][vdb_id][
                                                'deleted_objects'][obj_type])
                                    if obj_type in txlog['vdb_id'][vdb_id][
                                            'idempotent_updated_objects']:
                                        ids.extend(
                                            txlog['vdb_id'][vdb_id][
                                                'idempotent_updated_objects'][
                                                    obj_type])
                            if set(object_ids).isdisjoint(ids):
                                include = False
                if include:
                    received_txlogs['txlog'].append(txlog)
                include = True
                txlog = {}

    cli_output = {}
    cli_output['txlog'] = []
    if limit_size_enforced:
        txlog_lines = []
        txlog_lines_size = 0
        json_output_size = len(dump_json(cli_output))
        for txlog in reversed(received_txlogs['txlog']):
            if json_output is False:
                this_txlog_lines = []
                this_txlog_size = 0
                header_string = "%s\tClient:%s\tTxn_id:%s" % (
                    txlog['timestamp'], txlog['client_id'],
                    txlog['transaction_id'])
                this_txlog_lines.append(header_string)
                this_txlog_size += (len(header_string) + 1)

                if 'requests' in txlog:
                    requests_str = "\t\trequests:"
                    this_txlog_lines.append(requests_str)
                    this_txlog_size += (len(requests_str) + 1)
                    for request in txlog['requests']:
                        request_str = "\t\t\t%s" % str(request)
                        this_txlog_lines.append(request_str)
                        this_txlog_size += (len(request_str) + 1)

                updated_types = ['updated_objects', 'deleted_objects',
                                 'idempotent_updated_objects']
                for vdb_id in txlog['vdb_id']:
                    updated_type_str = "\t\tvdb_id: %s" % str(vdb_id)
                    this_txlog_lines.append(updated_type_str)
                    this_txlog_size += (len(updated_type_str) + 1)
                    for updated_type in updated_types:
                        updated_type_str = "\t\t\t%s: %s" % (updated_type, str(
                            txlog['vdb_id'][vdb_id][updated_type]))
                        this_txlog_lines.append(updated_type_str)
                        this_txlog_size += (len(updated_type_str) + 1)

                if txlog_lines_size + this_txlog_size > limit_size:
                    break

                txlog_lines_size += this_txlog_size
                txlog_lines.append("\n".join(this_txlog_lines))
            else:
                txlog_json = dump_json(cli_output)
                if json_output_size + len(txlog_json) > limit_size:
                    break
                json_output_size += len(txlog_json)
                cli_output['txlog'].insert(0, txlog)

        for txlog_line in reversed(txlog_lines):
            print(txlog_line)
    else:
        for txlog in received_txlogs['txlog']:
            if json_output is False:
                print(("%s\tClient:%s\tTxn_id:%s" % (
                    txlog['timestamp'], txlog['client_id'],
                    txlog['transaction_id'])))
                if 'requests' in txlog:
                    print("\t\trequests:")
                    for request in txlog['requests']:
                        print(("\t\t\t%s" % str(request)))
                updated_types = ['updated_objects', 'deleted_objects',
                                 'idempotent_updated_objects']
                for vdb_id in txlog['vdb_id']:
                    print(("\t\tvdb_id: %s" % str(vdb_id)))
                    for updated_type in updated_types:
                        print(("\t\t\t%s: %s" % (updated_type, str(
                            txlog['vdb_id'][vdb_id][updated_type]))))
        if json_output is not False:
            cli_output = received_txlogs

    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return True, rv


def run_transaction_log_command(cmd, json_output=None, beautify=False,
                                schema_file=DEFAULT_SCHEMA,
                                endpoint=DEFAULT_ENDPOINT):
    responses = []
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            responses = get_transaction_logs(nsxRpcClient)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'

    return run_transaction_log_command_impl(
                cmd, responses, json_output, beautify, schema_file)


def create_put_msg(schema_file, put_object, object_types, vdbs,
                   current_request_vdbs):
    msg_type = put_object["object_type"]
    if (len(object_types) != 0 and msg_type not in object_types):
        return None

    current_vdbs = put_object["vdb_id"]
    if len(current_vdbs) != 0:
        del current_request_vdbs[:]
        current_request_vdbs.extend(current_vdbs)
    if (len(vdbs) != 0 and
            set(vdbs).isdisjoint(set(current_request_vdbs))):
        return None

    put_msg = nestdb_pb2.PutMsg()

    object_type = get_object_type(
        schema_file,
        extract_object_type_enum_name(msg_type))
    put_msg.object.object_type = object_type

    object_instance = get_proto_object(schema_file, msg_type)
    populate_proto_object(put_object["object"], object_instance)
    put_msg.object.object_body = object_instance.SerializeToString()

    if len(vdbs) == 0:
        put_msg.vdb_id.extend(put_object["vdb_id"])
    else:
        for vdb_id in (set(vdbs) & set(current_request_vdbs)):
            put_msg.vdb_id.append(vdb_id)

    return put_msg


def create_put_request_msg(schema_file, request_msg, object_types, vdbs,
                           current_request_vdbs):
    msg = nestdb_pb2.PutRequestMsg()
    msg.sync = request_msg["sync"]

    for put_object in request_msg["objects"]:
        put_msg = create_put_msg(schema_file, put_object, object_types, vdbs,
                                 current_request_vdbs)
        if put_msg is not None:
            msg.put.add().CopyFrom(put_msg)

    if len(msg.put) != 0:
        return msg
    else:
        return None


def create_delete_msg(schema_file, delete_object, object_types, vdbs,
                      current_request_vdbs):
    msg_type = delete_object["object_type"]
    if (len(object_types) != 0 and msg_type not in object_types):
        return None

    current_vdbs = delete_object["vdb_id"]
    if len(current_vdbs) != 0:
        del current_request_vdbs[:]
        current_request_vdbs.extend(current_vdbs)
    if (len(vdbs) != 0 and
            set(vdbs).isdisjoint(set(current_request_vdbs))):
        return None

    delete_msg = nestdb_pb2.DeleteMsg()

    object_type = get_object_type(
        schema_file,
        extract_object_type_enum_name(msg_type))
    delete_msg.object.object_type = object_type

    object_instance = get_proto_object(schema_file, msg_type)
    populate_proto_object(delete_object["embedded_id"], object_instance)
    delete_msg.object.embedded_id = object_instance.SerializeToString()

    delete_msg.delete_dependents = delete_object["delete_dependents"]

    if len(vdbs) == 0:
        delete_msg.vdb_id.extend(delete_object["vdb_id"])
    else:
        for vdb_id in (set(vdbs) & set(current_request_vdbs)):
            delete_msg.vdb_id.append(vdb_id)

    return delete_msg


def create_delete_request_msg(schema_file, request_msg, object_types, vdbs,
                              current_request_vdbs):
    msg = nestdb_pb2.DeleteRequestMsg()
    msg.sync = request_msg["sync"]

    for delete_object in request_msg["objects"]:
        delete_msg = create_delete_msg(schema_file, delete_object,
                                       object_types, vdbs,
                                       current_request_vdbs)
        if delete_msg is not None:
            msg.delete.add().CopyFrom(delete_msg)

    if len(msg.delete) != 0:
        return msg
    else:
        return None


def create_flush_msg(schema_file, flush_object, object_types, vdbs,
                     current_request_vdbs):
    msg_type = flush_object["object_type"]
    if (len(object_types) != 0 and msg_type not in object_types):
        return None

    current_vdbs = flush_object["vdb_id"]
    if len(current_vdbs) != 0:
        del current_request_vdbs[:]
        current_request_vdbs.extend(current_vdbs)
    if (len(vdbs) != 0 and
            set(vdbs).isdisjoint(set(current_request_vdbs))):
        return None

    flush_msg = nestdb_pb2.FlushMsg()

    object_type = get_object_type(
        schema_file,
        extract_object_type_enum_name(msg_type))
    flush_msg.object_type = object_type

    if len(vdbs) == 0:
        flush_msg.vdb_id.extend(flush_object["vdb_id"])
    else:
        for vdb_id in (set(vdbs) & set(current_request_vdbs)):
            flush_msg.vdb_id.append(vdb_id)

    return flush_msg


def create_flush_request_msg(schema_file, request_msg, object_types, vdbs,
                             current_request_vdbs):
    msg = nestdb_pb2.FlushRequestMsg()
    msg.sync = request_msg["sync"]

    for flush_object in request_msg["objects"]:
        flush_msg = create_flush_msg(schema_file, flush_object, object_types,
                                     vdbs, current_request_vdbs)
        if flush_msg is not None:
            msg.flush.add().CopyFrom(flush_msg)

    if len(msg.flush) != 0:
        return msg
    else:
        return None


def create_patch_msg(schema_file, patch_object, object_types, vdbs,
                     current_request_vdbs):
    msg_type = patch_object["object_type"]
    if (len(object_types) != 0 and msg_type not in object_types):
        return None

    current_vdbs = patch_object["vdb_id"]
    if len(current_vdbs) != 0:
        del current_request_vdbs[:]
        current_request_vdbs.extend(current_vdbs)
    if (len(vdbs) != 0 and
            set(vdbs).isdisjoint(set(current_request_vdbs))):
        return None

    patch_msg = nestdb_pb2.PatchMsg()

    object_type = get_object_type(
        schema_file,
        extract_object_type_enum_name(msg_type))
    patch_msg.object_type = object_type

    op_object_instance = get_proto_object(schema_file,
                                          msg_type.replace('Msg', 'OpMsg'))
    populate_proto_object(patch_object["op_msg"], op_object_instance)
    patch_msg.op_msg = op_object_instance.SerializeToString()

    val_object_instance = get_proto_object(schema_file, msg_type)
    populate_proto_object(patch_object["value"], val_object_instance)
    patch_msg.value = val_object_instance.SerializeToString()

    if len(vdbs) == 0:
        patch_msg.vdb_id.extend(patch_object["vdb_id"])
    else:
        for vdb_id in (set(vdbs) & set(current_request_vdbs)):
            patch_msg.vdb_id.append(vdb_id)

    return patch_msg


def create_patch_request_msg(schema_file, request_msg, object_types, vdbs,
                             current_request_vdbs):
    msg = nestdb_pb2.PatchRequestMsg()
    msg.sync = request_msg["sync"]

    for patch_object in request_msg["patches"]:
        patch_msg = create_patch_msg(schema_file, patch_object, object_types,
                                     vdbs, current_request_vdbs)
        if patch_msg is not None:
            msg.patch.add().CopyFrom(patch_msg)

    if len(msg.patch) != 0:
        return msg
    else:
        return None


def create_copy_msg(schema_file, copy_object, object_types, vdbs,
                    current_request_vdbs):
    msg_type = copy_object["object_type"]
    if (len(object_types) != 0 and msg_type not in object_types):
        return None

    current_vdbs = copy_object["dst_vdb_id"]
    if len(current_vdbs) != 0:
        del current_request_vdbs[:]
        current_request_vdbs.extend(current_vdbs)
    if (len(vdbs) != 0 and
            set(vdbs).isdisjoint(set(current_request_vdbs))):
        return None

    if (len(vdbs) != 0 and
            (("src_vdb_id" in copy_object and
              copy_object["src_vdb_id"] not in vdbs) or
             ("src_vdb_id" not in copy_object and 1 not in vdbs))):
        return None

    copy_msg = nestdb_pb2.CopyMsg()

    object_type = get_object_type(
        schema_file,
        extract_object_type_enum_name(msg_type))
    copy_msg.object.object_type = object_type

    object_instance = get_proto_object(schema_file, msg_type)
    populate_proto_object(copy_object["embedded_id"], object_instance)
    copy_msg.object.embedded_id = object_instance.SerializeToString()

    if len(vdbs) == 0:
        if "src_vdb_id" in copy_object:
            copy_msg.src_vdb_id = copy_object["src_vdb_id"]
        copy_msg.dst_vdb_id.extend(copy_object["dst_vdb_id"])
    else:
        if "src_vdb_id" in copy_object:
            copy_msg.src_vdb_id = copy_object["src_vdb_id"]
        for vdb_id in (set(vdbs) & set(current_request_vdbs)):
            copy_msg.dst_vdb_id.append(vdb_id)

    return copy_msg


def create_copy_request_msg(schema_file, request_msg, object_types, vdbs,
                            current_request_vdbs):
    msg = nestdb_pb2.CopyRequestMsg()
    msg.sync = request_msg["sync"]

    for copy_object in request_msg["objects"]:
        copy_msg = create_copy_msg(schema_file, copy_object, object_types,
                                   vdbs, current_request_vdbs)
        if copy_msg is not None:
            msg.copy.add().CopyFrom(copy_msg)

    if len(msg.copy) != 0:
        return msg
    else:
        return None


def create_modify_request_msg(schema_file, request_msg, object_types, vdbs,
                              current_request_vdbs):
    msg = nestdb_pb2.ModifyRequestMsg()
    msg.sync = request_msg["sync"]

    for modify_object in request_msg["modify"]:
        for modify_type, modify_value in list(modify_object.items()):
            if modify_type == 'put':
                put_msg = create_put_msg(schema_file, modify_value,
                                         object_types, vdbs,
                                         current_request_vdbs)
                if put_msg is not None:
                    msg.modify.add().put.CopyFrom(put_msg)
            elif modify_type == 'delete':
                delete_msg = create_delete_msg(schema_file, modify_value,
                                               object_types, vdbs,
                                               current_request_vdbs)
                if delete_msg is not None:
                    msg.modify.add().delete.CopyFrom(delete_msg)
            elif modify_type == 'patch':
                patch_msg = create_patch_msg(schema_file, modify_value,
                                             object_types,
                                             vdbs, current_request_vdbs)
                if patch_msg is not None:
                    msg.modify.add().patch.CopyFrom(patch_msg)
            elif modify_type == 'flush':
                flush_msg = create_flush_msg(schema_file, modify_value,
                                             object_types,
                                             vdbs, current_request_vdbs)
                if flush_msg is not None:
                    msg.modify.add().flush.CopyFrom(flush_msg)
            elif modify_type == 'copy':
                copy_msg = create_copy_msg(schema_file, modify_value,
                                           object_types,
                                           vdbs, current_request_vdbs)
                if copy_msg is not None:
                    msg.modify.add().copy.CopyFrom(copy_msg)

    if len(msg.modify) != 0:
        return msg
    else:
        return None


def replay_txlog(txlog, vdbs, object_types, nsxRpcClient, schema_file):
    """
    Parses the txlog object. It checks whether the txlog satisfies the
    specified object_types and vdbs. If it does, then recreate
    the requests in the txlog and sends the requests to the NestDB server.
    """
    requests = txlog['requests']
    skip_txlog = False

    replay_requests = []
    replay_request_type = ""
    current_request_vdbs = [1]

    # All requests are of the same rpc operation type:
    # Put, Flush, Patch, Delete Modify, Create Vdb, Delete Vdb
    for request in requests:
        for request_type, request_msg in list(request.items()):
            replay_request_type = request_type
            if request_type == 'put':
                put_replay_request = create_put_request_msg(
                    schema_file, request_msg, object_types, vdbs,
                    current_request_vdbs)
                if put_replay_request is not None:
                    replay_requests.append(put_replay_request)
            elif request_type == 'delete':
                delete_replay_request = create_delete_request_msg(
                    schema_file, request_msg, object_types, vdbs,
                    current_request_vdbs)
                if delete_replay_request is not None:
                    replay_requests.append(delete_replay_request)
            elif request_type == 'patch':
                patch_replay_request = create_patch_request_msg(
                    schema_file, request_msg, object_types, vdbs,
                    current_request_vdbs)
                if patch_replay_request is not None:
                    replay_requests.append(patch_replay_request)
            elif request_type == 'flush':
                flush_replay_request = create_flush_request_msg(
                    schema_file, request_msg, object_types, vdbs,
                    current_request_vdbs)
                if flush_replay_request is not None:
                    replay_requests.append(flush_replay_request)
            elif request_type == 'copy':
                copy_replay_request = create_copy_request_msg(
                    schema_file, request_msg, object_types, vdbs,
                    current_request_vdbs)
                if copy_replay_request is not None:
                    replay_requests.append(copy_replay_request)
            elif request_type == 'modify':
                modify_replay_request = create_modify_request_msg(
                    schema_file, request_msg, object_types, vdbs,
                    current_request_vdbs)
                if modify_replay_request is not None:
                    replay_requests.append(modify_replay_request)
            elif request_type == 'create_vdb':
                replay_request = nestdb_pb2.CreateVDbRequestMsg()
                replay_request.vdb_name = request_msg["vdb_name"]
                replay_requests.append(replay_request)
            elif request_type == 'delete_vdb':
                replay_request = nestdb_pb2.DeleteVDbRequestMsg()
                replay_request.vdb_name = request_msg["vdb_name"]
                replay_requests.append(replay_request)

    if len(replay_requests) == 0:
        return True

    client_context = nsxRpcClient.CreateClientContext()
    client_context.SetClientId(txlog['client_id'])

    try:
        if request_type == 'put':
            writer, reader = nsxRpcClient.Put(None, DEFAULT_TIMEOUT,
                                              client_context)
            for replay_request in replay_requests:
                writer.next(replay_request)
            writer.done()
            rsp = next(reader.responseList)
        elif request_type == 'delete':
            writer, reader = nsxRpcClient.Delete(None, DEFAULT_TIMEOUT,
                                                 client_context)
            for replay_request in replay_requests:
                writer.next(replay_request)
            writer.done()
            rsp = next(reader.responseList)
        elif request_type == 'patch':
            writer, reader = nsxRpcClient.Patch(None, DEFAULT_TIMEOUT,
                                                client_context)
            for replay_request in replay_requests:
                writer.next(replay_request)
            writer.done()
            rsp = next(reader.responseList)
        elif request_type == 'flush':
            writer, reader = nsxRpcClient.Flush(None, DEFAULT_TIMEOUT,
                                                client_context)
            for replay_request in replay_requests:
                writer.next(replay_request)
            writer.done()
            rsp = next(reader.responseList)
        elif request_type == 'copy':
            writer, reader = nsxRpcClient.Copy(None, DEFAULT_TIMEOUT,
                                               client_context)
            for replay_request in replay_requests:
                writer.next(replay_request)
            writer.done()
            rsp = next(reader.responseList)
        elif request_type == 'modify':
            writer, reader = nsxRpcClient.Modify(None, DEFAULT_TIMEOUT,
                                                 client_context)
            for replay_request in replay_requests:
                writer.next(replay_request)
            writer.done()
            rsp = next(reader.responseList)
        elif request_type == 'create_vdb':
            call = nsxRpcClient.CreateVDb(replay_request, DEFAULT_TIMEOUT,
                                          client_context)
        elif request_type == 'delete_vdb':
            call = nsxRpcClient.DeleteVDb(replay_request, DEFAULT_TIMEOUT,
                                          client_context)
    except NsxRpcCallError as e:
        return False

    return True


def run_replay_command_impl(cmd, nsxRpcClient, json_output, schema_file):
    rv = '""'
    if len(cmd) < 1:
        return False, rv
    try:
        txlogs = json.load(open(cmd[0]))
    except json.scanner.JSONDecodeError as e:
        sys.stderr.write("Unable to decode json object from input. %s\n" %
                         e.msg)
        return False, rv

    object_types = []
    vdbs = []
    clients = []
    max_txn_id = None
    min_txn_id = None

    curr_index = 1
    while curr_index < len(cmd) - 1:
        if cmd[curr_index] == "object_type":
            try:
                object_types.append(
                    find_proto_msg_type(schema_file, cmd[curr_index + 1]))
            except:
                sys.stderr.write("Invalid object type provided")
                return False, rv
        elif cmd[curr_index] == "client":
            clients.append(cmd[curr_index + 1])
        elif cmd[curr_index] == "vdb_id":
            vdbs.append(cmd[curr_index + 1])
        elif cmd[curr_index] == "max_txn_id":
            if max_txn_id is not None:
                sys.stderr.write("Multiple max_txn_id provided\n")
                return False, rv
            max_txn_id = int(cmd[curr_index + 1])
        elif cmd[curr_index] == "min_txn_id":
            if min_txn_id is not None:
                sys.stderr.write("Multiple min_txn_id provided\n")
                return False, rv
            min_txn_id = int(cmd[curr_index + 1])
        curr_index = curr_index + 2

    cli_output = {}
    success = True

    for txlog in txlogs['txlog']:
        # Filter out txlog that does not match condition
        if max_txn_id is not None:
            if txlog["transaction_id"] > max_txn_id:
                break
        if min_txn_id is not None:
            if txlog["transaction_id"] < min_txn_id:
                continue
        # TODO(haroldl): This currently does not capture flush requests
        # that did not change anything. This results in the
        # flush request getting filtered out.
        if len(vdbs) != 0:
            if set(vdbs).isdisjoint(list(txlog['vdb_id'].keys())):
                continue
        if len(clients) != 0:
            if txlog["client_id"] not in clients:
                continue
        # To avoid parsing multiple times, pass object_types
        # to utility method.
        # Also pass vdbs to utility method so that
        # vdb
        if not replay_txlog(txlog, vdbs, object_types, nsxRpcClient,
                            schema_file):
            success = False
            break

    if json_output is False:
            print("Successfully replayed transaction log")
    elif json_output is not False:
        rv = json.dumps({'status': 'Success' if success else 'Fail'}, indent=4,
                        separators=(', ', ' : '))
        if json_output:
            print(rv)

    return success, rv


def run_replay_command(cmd, json_output=None,
                       schema_file=DEFAULT_SCHEMA,
                       endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDb_Stub, endpoint) as nsxRpcClient:
            return run_replay_command_impl(
                cmd, nsxRpcClient, json_output, schema_file)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


def run_metrics_command_impl(cmd, nsxRpcClient, json_output):
    rv = '""'

    prefix = ""
    if len(cmd) > 0:
        prefix = cmd[0]
    msg = nestdb_pb2.GetMetricsRequestMsg()
    msg.prefix = prefix

    try:
        rsp = nsxRpcClient.GetMetrics(msg, DEFAULT_TIMEOUT)
    except NsxRpcCallError as e:
        sys.stderr.write("NsxRpcClient encountered an error: "
                         "%s, code: %s, message: %s.\n" %
                         (e.name, e.code, e.message))
        return False, rv

    cli_output = {}
    for metric in rsp.metric:
        if json_output is False:
            print(("%s\t%s\n" % (metric.key, metric.value.replace(u'\xb5',
                                                                  'u'))))
        else:
            cli_output[metric.key] = metric.value
    if json_output is not False:
        rv = dump_json(cli_output)
        if json_output:
            print(rv)
    return True, rv


def run_metrics_command(cmd, json_output=None, endpoint=DEFAULT_ENDPOINT):
    try:
        with NsxRpcClient(nestdb_pb2.NestDbOps_Stub, endpoint) as nsxRpcClient:
            return run_metrics_command_impl(cmd, nsxRpcClient, json_output)
    except Exception as e:
        sys.stderr.write("NsxRpcClient encountered an error: %s\n" % str(e))
        return False, '""'


class NestDbCLI(cmd.Cmd):
    """NestDb CLI Shell"""

    def set_args(self, json_output, beautify, schema_file, server):
        self.prompt = '(nestdb) '
        self.json_output = json_output
        self.beautify = beautify
        self.schema_file = schema_file
        self.server = server
        self.object_types = []
        self.previous_command_return_code = True

    def default(self, line):
        print(('Command not recognized: {}'.format(line)))
        print("Do 'help' for available commands.")
        return False

    def emptyline(self):
        pass

    def do_EOF(self, line):
        print()
        return True

    def help_EOF(self):
        usage = "\n".join(nestdbcli_commands['EOF']['usage'])
        description = nestdbcli_commands['EOF']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def do_quit(self, line):
        return True

    def help_quit(self):
        usage = "\n".join(nestdbcli_commands['quit']['usage'])
        description = nestdbcli_commands['quit']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def do_server(self, line):
        if line:
            self.server = line.split()[0]
        print(("NestDb CLI's server endpoint: {}".format(self.server)))
        self.previous_command_return_code = True

    def help_server(self):
        usage = "\n".join(nestdbcli_commands['server']['usage'])
        description = nestdbcli_commands['server']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def do_output_format(self, line):
        if not line:
            print(("Current formatting for command outputs: '{}'".format(
                'json' if self.json_output else 'plain')))
        else:
            self.json_output = (line == 'json')
            print(("Output format set to: {}".format(
                'json' if self.json_output else 'plain')))
        self.previous_command_return_code = True

    def help_output_format(self):
        usage = "\n".join(nestdbcli_commands['output_format']['usage'])
        description = nestdbcli_commands['output_format']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_output_format(self, text, line, begin_index, end_index):
        sub_commands = ['json', 'plain']
        completions = []
        word_at_cursor = 1 if not text else 0
        if len(line[:end_index].split()) + word_at_cursor <= 2:
            if not text:
                completions = sub_commands
            else:
                completions = [sub_command for sub_command in sub_commands if
                               sub_command.startswith(text)]
        return completions

    def do_cli_schema(self, line):
        if line:
            self.schema_file = line.split()[0]
            # Reset descriptor_pool to force the next cli command that
            # uses the schema to load the descriptor pool.
            global descriptor_pool
            descriptor_pool = None
        print(("NestDb CLI's local schema file: {}".format(self.schema_file)))
        self.previous_command_return_code = True

    def help_cli_schema(self):
        usage = "\n".join(nestdbcli_commands['cli_schema']['usage'])
        description = nestdbcli_commands['cli_schema']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def do_schema(self, line):
        if not line:
            print("ERROR: not enough arguments. Do 'help schema' for usage "
                  "instructions.")
            self.previous_command_return_code = False
            return
        cmd = line.split()
        if cmd[0] == 'load' and len(cmd) >= 2:
            self.schema_file = cmd[1]
        self.previous_command_return_code = run_schema_command(
            cmd, self.json_output, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run schema command.")

    def help_schema(self):
        usage = "\n".join(nestdbcli_commands['schema']['usage'])
        description = nestdbcli_commands['schema']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_schema(self, text, line, begin_index, end_index):
        sub_commands = ['load', 'show', 'version']
        completions = []
        word_at_cursor = 1 if not text else 0
        if len(line[:end_index].split()) + word_at_cursor <= 2:
            if not text:
                completions = sub_commands
            else:
                completions = [sub_command for sub_command in sub_commands if
                               sub_command.startswith(text)]
        return completions

    def do_get(self, line):
        cmd = line.split(None, 1)
        if len(cmd) < 1:
            print("ERROR: not enough arguments. Do 'help get' for usage "
                  "instructions.")
            self.previous_command_return_code = False
            return

        # Extract vdb_id token from embedded id, if exists.
        if len(cmd) > 1:
            optional_params = cmd[1].rsplit(None, 1)
            try:
                int(optional_params[-1])
                cmd = [cmd[0]] + optional_params
            except:
                pass

        self.previous_command_return_code = run_get_command(
            cmd, self.json_output, self.beautify,
            self.schema_file, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run get command.")

    def complete_get(self, text, line, begin_index, end_index):
        return self.__complete_object_type_for_first_argument(
            text, line, begin_index, end_index)

    def help_get(self):
        usage = "\n".join(nestdbcli_commands['get']['usage'])
        description = nestdbcli_commands['get']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def do_put(self, line):
        params = line.strip().split("vdb_list")
        cmd = params[0].split(None, 1)
        if len(cmd) < 2:
            print("ERROR: not enough arguments. Do 'help put' for usage "
                  "instructions.")
            self.previous_command_return_code = False
            return

        # Parse vdb_list
        if len(params) > 1 and len(params[1]) > 0:
            try:
                cmd.extend([int(x) for x in params[1].split()])
            except ValueError:
                sys.stderr.write(
                    "Invalid vdb_id in vdb_list: %s\n" % params[1])
                self.previous_command_return_code = False
                return

        self.previous_command_return_code = run_put_command(
            cmd, self.json_output,
            self.schema_file, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run put command.")

    def help_put(self):
        usage = "\n".join(nestdbcli_commands['put']['usage'])
        description = nestdbcli_commands['put']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_put(self, text, line, begin_index, end_index):
        return (
            self.__complete_object_type_for_first_arg_and_vdb_list_after_json(
                text, line, begin_index, end_index))

    def do_delete(self, line):
        params = line.strip().split("vdb_list")
        cmd = params[0].split(None, 1)
        if len(cmd) < 2:
            print("ERROR: not enough arguments. Do 'help delete' for usage "
                  "instructions.")
            self.previous_command_return_code = False
            return

        # Parse vdb_list
        if len(params) > 1 and len(params[1]) > 0:
            try:
                cmd.extend([int(x) for x in params[1].split()])
            except ValueError:
                sys.stderr.write(
                    "Invalid vdb_id in vdb_list: %s\n" % params[1])
                self.previous_command_return_code = False
                return

        self.previous_command_return_code = run_delete_command(
            cmd, self.json_output, self.schema_file,
            self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run delete command.")

    def help_delete(self):
        usage = "\n".join(nestdbcli_commands['delete']['usage'])
        description = nestdbcli_commands['delete']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_delete(self, text, line, begin_index, end_index):
        return (
            self.__complete_object_type_for_first_arg_and_vdb_list_after_json(
                text, line, begin_index, end_index))

    def do_flush(self, line):
        params = line.strip().split("vdb_list")
        cmd = params[0].split(None, 1)
        if len(cmd) < 1:
            print("ERROR: not enough arguments. Do 'help flush' for usage "
                  "instructions.")
            self.previous_command_return_code = False
            return

        # Parse vdb_list
        if len(params) > 1 and len(params[1]) > 0:
            try:
                cmd.extend([int(x) for x in params[1].split()])
            except ValueError:
                sys.stderr.write(
                    "Invalid vdb_id in vdb_list: %s\n" % params[1])
                self.previous_command_return_code = False
                return

        self.previous_command_return_code = run_flush_command(
            cmd, self.json_output, self.schema_file, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run flush command.")

    def help_flush(self):
        usage = "\n".join(nestdbcli_commands['flush']['usage'])
        description = nestdbcli_commands['flush']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_flush(self, text, line, begin_index, end_index):
        completions = []
        word_at_cursor = 1 if not text else 0
        if len(line[:end_index].split()) + word_at_cursor <= 2:
            completions = self.__fill_object_type(text)
        elif len(line[:end_index].split()) + word_at_cursor == 3:
            if not text or 'vdb_list'.startswith(text):
                completions = ["vdb_list"]
        return completions

    def do_copy(self, line):
        cmd = line.split(None, 1)
        if len(cmd) < 2:
            print("ERROR: not enough arguments. Do 'help copy' for usage "
                  "instructions.")
            self.previous_command_return_code = False
            return

        cmds = [cmd[0]]
        start_idx = cmd[1].find('{')
        end_idx = cmd[1].rfind('}')

        if start_idx == -1 or end_idx == -1:
            print("ERROR: Unable to find embedded id. "
                  "Do 'help copy' for usage instructions.")
            self.previous_command_return_code = False
            return
        cmds.append(cmd[1][start_idx: end_idx + 1])

        vdb_params = cmd[1][end_idx + 1:]

        # Parse src_vdb and dst_vdb_list
        if (vdb_params.count("src_vdb_id") > 1 or
                vdb_params.count("dst_vdb_list") > 1 or
                (vdb_params.count("src_vdb_id") == 0 and
                 vdb_params.count("dst_vdb_list") == 0 and
                 len(vdb_params) > 0)):
            print("ERROR: Incorrect format for vdb parameter "
                  "Do 'help copy' for usage instructions.")
            self.previous_command_return_code = False
            return

        src_vdb_id = None
        if vdb_params.count("src_vdb_id") == 1:
            src_vdb_param = vdb_params.split(
                "src_vdb_id", 1)[1].split("dst_vdb_list")[0]
            try:
                src_vdb_id = int(src_vdb_param)
            except ValueError:
                sys.stderr.write(
                    "Invalid src_vdb_id provided: %s\n" % src_vdb_param)
                self.previous_command_return_code = False
                return
        dst_vdb_list = []
        if vdb_params.count("dst_vdb_list") == 1:
            dst_vdb_params = vdb_params.split(
                "dst_vdb_list", 1)[1].split("src_vdb_id")[0].split()
            try:
                if len(dst_vdb_params) == 0:
                    raise ValueError
                dst_vdb_list.extend([int(x) for x in dst_vdb_params])
            except ValueError:
                sys.stderr.write(
                    "Invalid dst_vdb_list provided: %s\n" % dst_vdb_params)
                self.previous_command_return_code = False
                return

        cmds.extend([src_vdb_id, dst_vdb_list])

        self.previous_command_return_code = run_copy_command(
            cmds, self.json_output, self.schema_file,
            self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run delete command.")

    def help_copy(self):
        usage = "\n".join(nestdbcli_commands['copy']['usage'])
        description = nestdbcli_commands['copy']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_copy(self, text, line, begin_index, end_index):
        completions = []
        word_at_cursor = 1 if not text else 0
        has_json = (line.count('{') > 0 and line.count('}') > 0 and
                    line.find('{') < line.find('}') and
                    line.rfind('{') < line.rfind('}') and
                    len(line[:line.find('{')].strip().split()) == 2)

        if has_json:
            try:
                json.loads(line[line.find('{'):line.rfind('}') + 1])
            except:
                has_json = False

        if has_json:
            vdb_params = line[line.rfind('}') + 1:]
            has_src_dst_vdbs = (
                'dst_vdb_list' in vdb_params and 'src_vdb_id' in vdb_params)

        if len(line[:end_index].split()) + word_at_cursor <= 2:
            completions = self.__fill_object_type(text)
        elif (has_json and len(vdb_params) > 0 and
              vdb_params[0] == ' ' and not has_src_dst_vdbs):
            vdb_tokens = vdb_params.split()
            sub_commands = ['src_vdb_id', 'dst_vdb_list']
            if len(vdb_params.strip()) == 0:
                completions = sub_commands
            elif len(vdb_tokens) == 1 and vdb_params[-1] != ' ':
                completions = [sub_command for sub_command in sub_commands if
                               sub_command.startswith(vdb_tokens[0])]
            elif ('src_vdb_id' in vdb_params and
                  ((len(vdb_tokens) == 2 and vdb_params[-1] == ' ') or
                   (len(vdb_tokens) == 3 and
                    "dst_vdb_list".startswith(vdb_tokens[-1])))):
                completions = ['dst_vdb_list']
            elif ('dst_vdb_list' in vdb_params and
                  ((len(vdb_tokens) >= 2 and vdb_params[-1] == ' ') or
                   (len(vdb_tokens) >= 3 and
                    "src_vdb_id".startswith(vdb_tokens[-1])))):
                completions = ['src_vdb_id']
        return completions

    def do_query(self, line):
        cmd = line.split()
        if len(cmd) < 2:
            print("ERROR: not enough arguments. Do 'help query' for usage "
                  "instructions.")
            self.previous_command_return_code = False
            return

        # Attempt to extract vdb_id, if exists
        vdb_id = None
        json_params = " ".join(cmd[1:])
        try:
            vdb_id = int(cmd[-1])
            json_params = " ".join(cmd[1:-1])
        except:
            pass

        try:
            json_strs = list(load_jsons(json_params))
        except ValueError as e:
            sys.stderr.write("%s\n" % str(e))
            self.previous_command_return_code = False
            return
        if len(json_strs) < 2:
            print("ERROR: not enough arguments. Do 'help query' for usage "
                  "instructions.")
            self.previous_command_return_code = False
            return
        cmds = [cmd[0], json.dumps(json_strs[0]), json.dumps(json_strs[1])]
        if len(json_strs) > 2:
            cmds.append(json.dumps(json_strs[2]))
        else:
            cmds.append(None)
        if vdb_id is not None:
            cmds.append(vdb_id)
        self.previous_command_return_code = run_query_command(
            cmds, self.json_output, self.beautify, self.schema_file,
            self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run query command.")

    def help_query(self):
        usage = "\n".join(nestdbcli_commands['query']['usage'])
        description = nestdbcli_commands['query']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_query(self, text, line, begin_index, end_index):
        return self.__complete_object_type_for_first_argument(
            text, line, begin_index, end_index)

    def do_replay(self, line):
        """Replay the requests in a given transaction log file (in json format)
           that match the given conditions"""
        self.previous_command_return_code = run_replay_command(
            line.split(), self.json_output, self.schema_file, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run replay command.")

    def help_replay(self):
        usage = "\n".join(nestdbcli_commands['replay']['usage'])
        description = nestdbcli_commands['replay']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_replay(self, text, line, begin_index, end_index):
        completions = []
        arg_names = ['client', 'object_type', 'vdb_id', 'min_txn_id',
                     'max_txn_id']
        cursor_is_empty = 1 if not text else 0
        cursor_has_word = 1 if not cursor_is_empty else 0
        words_before_cursor = line[:end_index].split()
        if (len(words_before_cursor) + 1 + cursor_is_empty) % 2 == 0:
            if 'min_txn_id' in words_before_cursor:
                arg_names.remove('min_txn_id')
            if 'max_txn_id' in words_before_cursor:
                arg_names.remove('max_txn_id')
            if not text:
                completions = arg_names
            else:
                completions = [arg_name for arg_name in arg_names if
                               arg_name.startswith(text)]
        elif (len(words_before_cursor) + cursor_is_empty >= 3 and
                words_before_cursor[-1-cursor_has_word] == 'object_type'):
            completions = self.__fill_object_type(text)
        return completions

    def do_transaction_log(self, line):
        """Read transaction logs and output those that match the given
        select conditions"""
        self.previous_command_return_code = run_transaction_log_command(
            line.split(), self.json_output, self.beautify, self.schema_file,
            self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run txn_log command.")

    def help_transaction_log(self):
        usage = "\n".join(nestdbcli_commands['transaction_log']['usage'])
        description = nestdbcli_commands['transaction_log']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_transaction_log(self, text, line, begin_index, end_index):
        completions = []
        arg_names = ['client', 'object_type', 'object_id', 'vdb_id']
        cursor_is_empty = 1 if not text else 0
        cursor_has_word = 1 if not cursor_is_empty else 0
        words_before_cursor = line[:end_index].split()
        if (len(words_before_cursor) + cursor_is_empty) % 2 == 0:
            if len(words_before_cursor) + cursor_is_empty == 2:
                arg_names.append('limit_size')
            if not text:
                completions = arg_names
            else:
                completions = [arg_name for arg_name in arg_names if
                               arg_name.startswith(text)]
        elif (len(words_before_cursor) + cursor_is_empty >= 3 and
                words_before_cursor[-1-cursor_has_word] == 'object_type'):
            completions = self.__fill_object_type(text)
        return completions

    def do_metrics(self, line):
        self.previous_command_return_code = run_metrics_command(
            line.split(), self.json_output, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run metrics command.")

    def help_metrics(self):
        usage = "\n".join(nestdbcli_commands['metrics']['usage'])
        description = nestdbcli_commands['metrics']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def do_object_type(self, line):
        self.previous_command_return_code = run_object_type_command(
            line.split(), self.json_output, self.schema_file)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run object_type command.")

    def help_object_type(self):
        usage = "\n".join(nestdbcli_commands['object_type']['usage'])
        description = nestdbcli_commands['object_type']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_object_type(self, text, line, begin_index, end_index):
        sub_commands = ['list', 'show']
        completions = []
        word_at_cursor = 1 if not text else 0
        if len(line[:end_index].split()) + word_at_cursor <= 2:
            if not text:
                completions = sub_commands
            else:
                completions = [sub_command for sub_command in sub_commands if
                               sub_command.startswith(text)]
        elif (len(line[:end_index].split()) + word_at_cursor <= 3 and
                line.split()[1] == 'show'):
            completions = self.__fill_object_type(text)
        return completions

    def do_attribute(self, line):
        cmd = line.split()
        if len(cmd) < 4:
            print("ERROR: not enough arguments. Do 'help attribute' for "
                  "usage instructions.")
            self.previous_command_return_code = False
            return
        if cmd[0] == 'query':
            cmds = cmd
        elif cmd[0] == 'update':
            cmds = [cmd[0], cmd[1]]
            start_idx = line.find('{')
            end_idx = line.rfind('}')
            if start_idx == -1 or end_idx == -1:
                print("ERROR: Unable to find embedded id. "
                      "Do 'help attribute' for usage instructions.")
                self.previous_command_return_code = False
                return
            cmds.append(line[start_idx: end_idx + 1])
            cmds += [x for x in line[end_idx + 1:].split() if x != "vdb_list"]
        else:
            print("ERROR: Unknown command. Do 'help attribute' for usage "
                  "instructions.")
            self.previous_command_return_code = False
            return

        self.previous_command_return_code = run_attribute_command(
            cmds, self.json_output, self.beautify,
            self.schema_file, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run attribute command.")

    def help_attribute(self):
        usage = "\n".join(nestdbcli_commands['attribute']['usage'])
        description = nestdbcli_commands['attribute']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_attribute(self, text, line, begin_index, end_index):
        sub_commands = ['query', 'update']
        completions = []
        word_at_cursor = 1 if not text else 0
        if len(line[:end_index].split()) + word_at_cursor <= 2:
            if not text:
                completions = sub_commands
            else:
                completions = [sub_command for sub_command in sub_commands if
                               sub_command.startswith(text)]
        elif len(line[:end_index].split()) + word_at_cursor <= 3:
            completions = self.__fill_object_type(text)
        elif 'update' in line:
            has_json = (line.count('{') > 0 and line.count('}') > 0 and
                        line.find('{') < line.find('}') and
                        line.rfind('{') < line.rfind('}'))
            if has_json:
                try:
                    json.loads(line[line.find('{'):line.rfind('}') + 1])
                except:
                    has_json = False
            if (has_json and
                    len(line[line.rfind('}') + 1:].split()) +
                    word_at_cursor == 3):
                if not text or 'vdb_list'.startswith(text):
                    completions = ["vdb_list"]
        return completions

    def do_vdb(self, line):
        cmd = line.split()
        if (len(cmd) < 1 or
                (len(cmd) < 2 and (cmd[0] == 'create' or cmd[0] == 'delete'))):
            print("ERROR: not enough arguments. Do 'help vdb' for "
                  "usage instructions.")
            self.previous_command_return_code = False
            return
        self.previous_command_return_code = run_vdb_command(
            cmd, self.json_output, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run vdb command.")

    def help_vdb(self):
        usage = "\n".join(nestdbcli_commands['vdb']['usage'])
        description = nestdbcli_commands['vdb']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_vdb(self, text, line, begin_index, end_index):
        sub_commands = ['create', 'delete', 'get']
        completions = []
        word_at_cursor = 1 if not text else 0
        if len(line[:end_index].split()) + word_at_cursor <= 2:
            if not text:
                completions = sub_commands
            else:
                completions = [sub_command for sub_command in sub_commands if
                               sub_command.startswith(text)]
        return completions

    def do_certificate(self, line):
        cmd = line.split()
        if (len(cmd) < 2 or
            (cmd[0] != 'add' and cmd[0] != 'delete' and cmd[0] != 'get') or
                (len(cmd) < 3 and cmd[0] == 'add')):
            print("ERROR: Incorrect arguments. Do 'help certificate' for "
                  "usage instructions.")
            self.previous_command_return_code = False
            return
        self.previous_command_return_code = run_certificate_command(
            cmd, self.json_output, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run certificate command.")

    def help_certificate(self):
        usage = "\n".join(nestdbcli_commands['certificate']['usage'])
        description = nestdbcli_commands['certificate']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_certificate(self, text, line, begin_index, end_index):
        sub_commands = ['add', 'delete', 'get']
        completions = []
        word_at_cursor = 1 if not text else 0
        if len(line[:end_index].split()) + word_at_cursor <= 2:
            if not text:
                completions = sub_commands
            else:
                completions = [sub_command for sub_command in sub_commands if
                               sub_command.startswith(text)]
        return completions

    def do_merkle_tree(self, line):
        cmd = line.strip().split()
        if len(cmd) < 1:
            print("ERROR: not enough arguments. Do 'help merkle_tree' for "
                  "usage instructions.")
            self.previous_command_return_code = False
            return

        self.previous_command_return_code = run_merkle_tree_command(
            cmd, self.json_output, self.server)[0]
        if not self.previous_command_return_code:
            print("ERROR: failed to run flush command.")

    def help_merkle_tree(self):
        usage = "\n".join(nestdbcli_commands['merkle_tree']['usage'])
        description = nestdbcli_commands['merkle_tree']['description']
        print(("Usage:-\n" "%s\n" "Description:-\n" "%s" %
              (usage, description)))

    def complete_flush(self, text, line, begin_index, end_index):
        completions = []
        word_at_cursor = 1 if not text else 0
        if len(line[:end_index].split()) + word_at_cursor <= 2:
            completions = self.__fill_object_type(text)
        elif len(line[:end_index].split()) + word_at_cursor == 3:
            if not text or 'vdb_list'.startswith(text):
                completions = ["vdb_list"]
        return completions

    def __complete_object_type_for_first_argument(
            self, text, line, begin_index, end_index):
        completions = []
        word_at_cursor = 1 if not text else 0
        if len(line[:end_index].split()) + word_at_cursor <= 2:
            completions = self.__fill_object_type(text)
        return completions

    def __fill_object_type(self, text):
        "Function handler to help fill object type"
        if not self.object_types:
            self.object_types = get_msg_types(self.schema_file)
            self.object_types.sort()
        if not text:
            return self.object_types
        else:
            return [obj_type for obj_type in self.object_types
                    if obj_type.startswith(text)]

    def __complete_object_type_for_first_arg_and_vdb_list_after_json(
            self, text, line, begin_index, end_index):
        completions = []
        word_at_cursor = 1 if not text else 0
        has_json = (line.count('{') > 0 and line.count('}') > 0 and
                    line.find('{') < line.find('}') and
                    line.rfind('{') < line.rfind('}') and
                    len(line[:line.find('{')].strip().split()) == 2)

        if has_json:
            try:
                json.loads(line[line.find('{'):line.rfind('}') + 1])
            except:
                has_json = False

        if has_json:
            vdb_params = line[line.rfind('}') + 1:]

        if len(line[:end_index].split()) + word_at_cursor <= 2:
            completions = self.__fill_object_type(text)
        elif (has_json and len(vdb_params) > 0 and
              'vdb_list' not in vdb_params and vdb_params[0] == ' '):
            if (len(vdb_params.strip()) == 0 or
                    'vdb_list'.startswith(vdb_params[1:])):
                completions = ['vdb_list']
        return completions


def get_help_commands():
    commands_of_interest = ['schema', 'get', 'put', 'delete', 'flush', 'query',
                            'transaction_log', 'metrics', 'object_type',
                            'attribute', 'vdb']
    help_commands = []
    for command in commands_of_interest:
        help_commands.extend(
            ["- " + x for x in nestdbcli_commands[command]['usage']])
    return help_commands


def main():
    commands = get_help_commands()
    parser = argparse.ArgumentParser(
        description='Nestdb CLI Client',
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        '--cmd', nargs='+',
        help='Runs a CLI command where CMD is one of:\n%s' %
        "\n".join(commands))
    parser.add_argument(
        '--json', nargs='?', dest='json_output', default=False, const=True,
        help='Outputs in json format (default: %(default)s)')
    parser.add_argument(
        '--schema', dest='schema_file',
        default=DEFAULT_SCHEMA,
        help='Location of the nestdb schema file (default: %(default)s)')
    parser.add_argument(
        '--schema_prefix', dest='schema_prefix',
        default=DEFAULT_SCHEMA_PREFIX,
        help='nestdb schema file\'s prefix (default: %(default)s)')
    parser.add_argument(
        '--client_id', dest='client_id',
        default=DEFAULT_CLIENT_ID,
        help='client_id to use to connect to the NestDb server '
        '(default: %(default)s)')
    parser.add_argument(
        "--server", dest='server',
        default=DEFAULT_ENDPOINT,
        help='Endpoint address of Nestdb server (default: %(default)s)')
    parser.add_argument(
        "--beautify", nargs='?', dest='beautify',
        default=False, const=True,
        help='Convert known message type to string (default: %(default)s')
    args = parser.parse_args()

    set_schema_prefix(args.schema_prefix)

    set_client_id(args.client_id)

    nestdb_cli = NestDbCLI()
    nestdb_cli.set_args(args.json_output, args.beautify, args.schema_file,
                        args.server)
    if args.cmd:
        nestdb_cli.onecmd(" ".join(args.cmd))
        sys.exit(0 if nestdb_cli.previous_command_return_code else 1)
    else:
        print((
            "json_output={} schema_file={} schema_prefix={} server={}".format(
                str(args.json_output), args.schema_file, schema_prefix,
                args.server)))
        nestdb_cli.cmdloop()


if __name__ == "__main__":
    main()
