# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import json
import sys
import traceback

from sha.contrib.metric.edge_base import EdgeBase
from sha.contrib.metric.utils import run_command, timeout_cache

class TlsStats(EdgeBase):

    def get_tls_stats(self):
        tls_stats_list = []

        # Get TLS stats
        # Get TLS stats via curl command
        cmd_get_stats = ['sudo', 'ip', 'netns', 'exec', 'tlr_sr', 'curl', 'http://127.0.0.1:9001/nsx/get/stats?export=1']
        try:
            output = self.run_command_with_time_check(cmd_get_stats, timeout=10)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.debug("Failed to get TLS stats: {}, {}".format(e, trace))
            output = None
        if output is None:
            tls_stats_list = None
        else:
            tls_stats_list = json.loads(output)
        self._logger.debug("TLS stats({}): {}".format(sys.getsizeof(tls_stats_list), tls_stats_list))
        return tls_stats_list

    def get_tls_bypassed_conn(self):
        tls_debug_count = []

        # Get debug count via edge backend command
        cmd_get_debug_count = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'fw/get_debug_count']
        try:
            output = self.run_command_with_time_check(cmd_get_debug_count, timeout=10)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.debug("Failed to get debug count: {}, {}".format(e, trace))
            output = None
        if output is not None:
            tls_debug_count = json.loads(output)
        self._logger.debug("TLS debug count({}): {}".format(sys.getsizeof(tls_debug_count), tls_debug_count))

        tls_bypassed_conn = {}
        for debug_count in tls_debug_count:
            tls_bypassed_conn[debug_count.get('ifuuid', 0)] = debug_count.get('TLS bypassed connections', 0)

        return tls_bypassed_conn;

    @timeout_cache(10)
    def usage(self, *args, **kwargs):
        # Get TLS rule stats via edge backend command
        cmd_get_rule_stats = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'fw/show', 'tlsrulestat']
        try:
            output = self.run_command_with_time_check(cmd_get_rule_stats, timeout=10)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.debug("Failed to get TLS rule stats: {}, {}".format(e, trace))
            output = None
        if output is None:
            tls_rule_stats = None
        else:
            tls_rule_stats = json.loads(output)
        self._logger.debug("TLS rule stats({}): {}".format(sys.getsizeof(tls_rule_stats), tls_rule_stats))

        tls_bypassed_conn = {}
        try:
            tls_bypassed_conn = self.get_tls_bypassed_conn()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.debug("Failed to get TLS bypassed connection on {}: {}".format(e, trace))

        tls_stats_list = []
        try:
            tls_stats_list = self.get_tls_stats()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.debug("Failed to get TLS stats on {}: {}".format(e, trace))

        lr_to_if = {}
        for rule_stats in tls_rule_stats:
            lr_uuid = rule_stats.get('mp-router-id', None)
            if_uuid = rule_stats.get('ifuuid', None)
            if_type = rule_stats.get('iftype', None)
            if lr_uuid is None or if_uuid is None or if_type != 'UPLINK':
                continue
            lr_to_if[lr_uuid] = if_uuid

        for lr_tls_stats in tls_stats_list:
            lr_uuid = lr_tls_stats.get('lr_uuid', None)
            if lr_uuid:
                if_uuid = lr_to_if.get(lr_uuid, None)
                if if_uuid:
                    lr_stats = lr_tls_stats.get('stats', {})
                    if 'bypassed_due_to_rules' in lr_stats:
                        lr_stats['bypassed_due_to_rules'] += tls_bypassed_conn.get(if_uuid, 0)

        return {'tls-stats-list': tls_stats_list, \
                'rule-stats': tls_rule_stats}
