# Copyright (C) 2019-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import threading
import time
import traceback

import sha.config
import sha.preconfig
from sha.core.plugin.plugin_status_collection import PluginStatusCollection
from sha.constants import CHECK_INTERVAL, ENABLE, NORMAL, ERROR
from sha.config import GLOBAL_MINIMUM_CHECKING_INTERVAL
from sha.core.collector import Collector
from sha.core.whiteboard import DataWhiteboard
from sha.utils import sleep_and_check_running


class LoopNotifier(threading.Thread):

    def __init__(self, plugin, *args, **kwargs):
        super().__init__(name='nsx-sha-loop-notifier', *args, **kwargs)
        self._is_running = False
        self._logger = sha.preconfig.get_sub_logger(__name__, s2comp="loop-notifier-[{}]".format(plugin.name))
        self._plugin = plugin
        self.check_interval = GLOBAL_MINIMUM_CHECKING_INTERVAL
        self.registered = False

    def __repr__(self):
        return 'LoopNotifier - is_running:{}, plugin:{}'.format(self._is_running, self._plugin)

    def unregister_metrics(self):
        if self.registered:
            self.registered = False
            Collector.unregister_metric(self._plugin.id)

    def register_metrics(self, check_interval):
        if not self.registered or self.check_interval != check_interval:
            self.unregister_metrics()
            Collector.register_metric(self._plugin.id, self._plugin.loop_checked_metric_names, check_interval)
            self.check_interval = check_interval
            self.registered = True

    def get_metrics(self):
        """ Get metrics from data whiteboard:
            {
                metric_name1: {'value': value1, 'timestamp': timestamp1},
                metric_name2: {'value': value2, 'timestamp': timestamp2},
                metric_name3: None   # when no data in whiteboard
            }
        """
        if not self._plugin.metric_names \
                and not self._plugin.prefix_metric_names:
            return {}

        return DataWhiteboard.get_metrics(self._plugin.id,
                                          self._plugin.metric_names,
                                          prefix_metrics=self._plugin.prefix_metric_names)

    def run(self):
        self._logger.info('starting loop notify: {}'.format(self._plugin.name))
        self._is_running = True
        self.init_delay()

        while self._is_running:
            try:
                profile = self._plugin.load_profile()

                if not profile[ENABLE]:
                    PluginStatusCollection.update_plugin_status(self._plugin.id, NORMAL, 'Plugin is disabled.')
                    self._logger.debug('plugin:{} is disabled.'.format(self._plugin.name))
                    # If registered, unregister metrics
                    self.unregister_metrics()
                    continue

                unreached_msg = self._plugin.check_pre_condition()
                if unreached_msg is not None:
                    PluginStatusCollection.update_plugin_status(self._plugin.id, NORMAL,
                                                                'Plugin is not executed as: {}'.format(unreached_msg))
                    # If registered, unregister metrics
                    self.unregister_metrics()
                    continue

                # Register metrics if need
                if self._plugin.loop_checked_metric_names:
                    self.register_metrics(profile[CHECK_INTERVAL])
                else:
                    self.check_interval = profile[CHECK_INTERVAL]

                # Get metrics
                metrics = self.get_metrics()
                if metrics is None:
                    self._logger.warning("Failed in getting metrics from whiteboard, metrics: {}, perfix_metrics: {}"
                                         .format(self._plugin.metric_names, self._plugin.prefix_metric_names))
                    continue

                self._logger.debug('Get metrics: {}.'.format(metrics))

                self._plugin.execute(metrics)

                PluginStatusCollection.update_plugin_status(self._plugin.id, NORMAL, '')
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning(
                    "Failed in LoopNotifier for {}, Exception:{}, trace:{}".format(self._plugin.name, e, trace))
                PluginStatusCollection.update_plugin_status(
                    self._plugin.id, ERROR, "Failed in running by loop notifier, error:" + str(e))
            finally:
                try:
                    sleep_and_check_running(self, max(self.check_interval, sha.config.GLOBAL_MINIMUM_CHECKING_INTERVAL))
                except Exception as e:
                    trace = traceback.format_exc()
                    self._logger.warning("Plugin {} is interrupted from sleep, Exception:{}, trace:{}"
                                         .format(self._plugin.name, e, trace))

        self.unregister_metrics()
        self._logger.warning("stopped loop notifier of {}".format(self._plugin.name))

    def stop(self):
        self._logger.warning("going to stop loop notifier of {}".format(self._plugin.name))
        self._is_running = False

    @property
    def plugin(self):
        return self._plugin

    def init_delay(self):
        try:
            delay = self._plugin.delay_before_first_check
            self._logger.info("Delay {}s before first check round, plugin:{}.".format(delay, self._plugin.name))
            sleep_and_check_running(self, delay)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Plugin {} is interrupted from init_delay sleep, Exception:{}, trace:{}"
                                 .format(self._plugin.name, e, trace))

    @property
    def is_running(self):
        return self._is_running


