# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
from sys import getsizeof
from threading import Lock

from sha.utils import total_size

from sha import config

_logger = config.get_logger(__name__)


class PluginData:
    """
    Structure for plugin data which will be stored in whiteboard finally.
    For each plugin, the max size of its plugin data is 50 KBytes.
    """

    # max data size per plugin is 50KB
    MAX_DATA_SIZE = 50 * 1000

    def __init__(self, id, data=None):
        self._id = id
        self._data = data if data is not None else {}
        self._current_size = total_size(data)
        self._access_lock = Lock()

    def __repr__(self):

        return 'PluginData - plugin id:{}, data_size:{}, data:{}'.format(self._id, self._current_size, self._data)

    def clear(self):
        with self._access_lock:
            self._data.clear()
            self._refresh_size()

    def set(self, key, value):
        with self._access_lock:
            self._validate(key, value)
            self._data[key] = value
            self._refresh_size()
            _logger.debug('PluginData - plugin id:{}, set key:{}, value:{}'.format(self._id, key, value))

    def get(self, key):
        return self._data.get(key, None)

    def delete(self, key):
        with self._access_lock:
            del self._data[key]
            self._refresh_size()

    @property
    def interfaces(self):
        return {'set_plugin_data': self.set,
                'get_plugin_data': self.get,
                'clear_plugin_data': self.clear,
                'delete_plugin_data': self.delete}

    @property
    def data(self):
        return self._data

    @staticmethod
    def create_metric_name(id):
        return "nsx.sha.plugin-data.plugin-id-{}".format(id)

    def _validate(self, key, value):
        """
        1. key can only be str, value can only be str, int, float or boolean.
        2. there will be an up-limit (50KB) for plugin_data size for each plugin.
        """

        if not isinstance(key, str):
            raise Exception("Invalid key type:{}. The type of key must be str.".format(type(key)))

        if value is not None and not isinstance(value, (int, float, str, bool)):
            raise Exception("Invalid value type:{}. "
                            "The type of value must belong to set: (int, float, str, bool).".format(type(key)))

        k_size = getsizeof(key)
        v_size = getsizeof(value)
        if k_size + v_size + self._current_size > self.MAX_DATA_SIZE:
            raise Exception("Set data is failed (key:{}, value:{}). The total data size will exceed up-limit:{}. "
                            "key_size:{}, value_size:{}, current_data_size:{}"
                            .format(key, value, self.MAX_DATA_SIZE, k_size, v_size, self._current_size))

    @property
    def metric_key(self):
        return self.create_metric_name(self._id)

    def _refresh_size(self):
        self._current_size = total_size(self._data)

