# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import copy
import threading
import traceback

from sha.core.plugin.plugin_status import PluginStatus
from sha.constants import ERROR, INSTALL_FAIL

from sha.contrib.action.remedy import Remedy
from sha.core.whiteboard._data_whiteboard import DataWhiteboard
from sha.preconfig import get_sub_logger

_logger = get_sub_logger(__name__, s2comp="plugin-status-collection")

class PluginStatusCollection:

    plugin_status_list = {}
    lock = threading.RLock()

    @classmethod
    def add_plugin_status(cls, plugin_id, plugin_name):
        with cls.lock:
            exist_status = cls.plugin_status_list.get(plugin_id, None)
            if exist_status is not None:
                exist_status.name = plugin_name
                _logger.warn("Plugin status is already in collection, no need to add again. exist_status:{}"
                                 .format(exist_status))
                return

            cls.plugin_status_list[plugin_id] = PluginStatus(plugin_id, plugin_name)

    @classmethod
    def get_plugin_status(cls, plugin_id):
        return cls.plugin_status_list.get(plugin_id, None)

    @classmethod
    def remove_plugin_status(cls, plugin_id):
        with cls.lock:
            plugin_status = cls.plugin_status_list.pop(plugin_id, None)
            if plugin_status is not None:
                _logger.info("Removed plugin status:{}".format(plugin_status))

    @classmethod
    def get_all_plugin_status(cls):
        # refresh profile info, in case some plugin's profile is out of date.
        # e.g. UplinkStatusMonitor (not check by interval) or other plugins whose check interval is large.
        cls._refresh_profile_info()
        status_list = copy.deepcopy(cls.plugin_status_list)
        cls._add_rmdy_result(status_list)
        return status_list

    @classmethod
    def _add_rmdy_result(cls, status_list):
        for plugin_id, status in status_list.items():
            res = cls._get_rmdy_result(plugin_id)
            if not res:
                continue
            if status.detail:
                status.detail += ', '
            status.detail += res

    @staticmethod
    def _get_rmdy_result(plugin_id):
        name = Remedy.create_prefix_metric_name(plugin_id)
        metrics = DataWhiteboard.get_metrics(plugin_id, metrics=[], prefix_metrics=[name])
        if metrics is None:
            _logger.warning('PluginStatusCollection - Failed in getting metrics from whiteboard, '
                            'metric_name:{}, plugin_id:{}'.format(name, plugin_id))
            return ''

        results = []
        for m_val in metrics.values():
            results.extend(map(str, m_val['value']))

        if not results:
            return ''

        return ', '.join(str(r) for r in results)

    @classmethod
    def update_plugin_status(cls, plugin_id, status, detail=None, profile=None):
        with cls.lock:
            plugin_status = cls.get_plugin_status(plugin_id)
            if plugin_status:
                plugin_status.status = status
                if detail is not None:
                    plugin_status.detail = detail
                if profile is not None:
                    plugin_status.profile = copy.deepcopy(profile)

    @classmethod
    def update_plugin_status_profile(cls, plugin_id, profile=None):
        with cls.lock:
            plugin_status = cls.get_plugin_status(plugin_id)
            if plugin_status:
                plugin_status.profile = copy.deepcopy(profile)

    @classmethod
    def _refresh_profile_info(cls):
        from sha.core.profile._resolver import Resolver
        for status in cls.plugin_status_list.values():
            if status.status == INSTALL_FAIL:
                # For those install_fail plugin, no need to refresh its profile
                continue

            plugin_id = status.plugin_id
            try:
                profile = Resolver.get(plugin_id)
            except Exception as e:
                cls.update_plugin_status(plugin_id, ERROR, "Failed in loading profile." + str(e), {})
                trace = traceback.format_exc()
                _logger.warn("PluginStatusCollection - Failed in load profile for plugin id {}, trace:{}"
                                 .format(plugin_id, trace))
                continue
            # TODO update logic here, perform update if any diff exist or error.
            cls.update_plugin_status_profile(plugin_id, profile)
