# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential


def clear_data(last_remedy_time, started_by_init_timestamp):
    """
    clear plugin data only keep last_remedy_time
    """
    clear_plugin_data()
    set_plugin_data('last_remedy_time', last_remedy_time)
    set_plugin_data('started_by_init_timestamp', started_by_init_timestamp)


def nestdb_remediation():
    """
    Perform remediation according to metrics.

    For the remediation logic explanation, please refer to below flowchart of below FS:
    https://confluence.eng.vmware.com/display/NSBU/SHA+Distributed+Remediation#SHADistributedRemediation-RemediationNestDBPluginWorkflow
    """
    profile = data_store['profile']
    last_remedy_time = get_plugin_data('last_remedy_time')
    min_interval = profile['MIN_INTERVAL_BETWEEN_TWO_REMEDIATION']
    interval = time.monotonic() - last_remedy_time \
        if last_remedy_time is not None else min_interval + 1

    logger.debug('nestdb_remediation - profile:{}, data_store:{}'.format(profile, data_store))
    logger.debug('nestdb_remediation - plugin data - init_nestdb_is_started_timestamp:{}, crash_before_disk_full:{}, '
                'crash_by_disk_full:{}, tried_count:{}, last_remedy_time:{}, watchdog_is_running_before:{}'
                .format(get_plugin_data('init_nestdb_is_started_timestamp'), get_plugin_data('crash_before_disk_full'),
                        get_plugin_data('crash_by_disk_full'), get_plugin_data('tried_count'),
                        get_plugin_data('last_remedy_time'), get_plugin_data('watchdog_is_running_before')))

    if interval <= min_interval:
        logger.debug('Not remedy, because the interval:{} is smaller than '
                     'MIN_INTERVAL_BETWEEN_TWO_REMEDIATION:{}'.format(interval, min_interval))
        return

    if data_store['metrics']['nsx.nestdb.nestdb-status'] is None:
        logger.info('Metric {} is None from whiteboard'.format('nsx.nestdb.nestdb-status'))
        return

    nestdb_status = data_store['metrics']['nsx.nestdb.nestdb-status']

    started_by_init = None
    started_by_init_timestamp = None

    # if it's none, means SHA did not receive / lost nestdb init script notification due to some reason.
    # e.g. SHA restarted.
    if data_store['metrics']['nsx.metrics.nestdb-init.nestdb_is_started'] is not None:
        init_nestdb_is_started = data_store['metrics']['nsx.metrics.nestdb-init.nestdb_is_started']

        started_by_init = init_nestdb_is_started['value']['val']
        started_by_init_timestamp = init_nestdb_is_started['timestamp']
        if started_by_init is False:
            clear_data(last_remedy_time, started_by_init_timestamp)
            logger.debug('Not remedy, because nestdb is stopped by init script.')
            return

        if get_plugin_data('started_by_init_timestamp') is None \
                or started_by_init_timestamp != get_plugin_data('started_by_init_timestamp'):
            clear_data(last_remedy_time, started_by_init_timestamp)
            logger.debug('Clear history data, as nestdb process is started/stopped by init script.')

        if nestdb_status['timestamp'] < init_nestdb_is_started['timestamp']:
            logger.debug('Not remedy, as nestdb_status is out of date.')
            return

    if nestdb_status['value']['watchdog_is_running']:
        clear_data(last_remedy_time, started_by_init_timestamp)
        set_plugin_data('watchdog_is_running_before', True)
        logger.debug('Not remedy, because the watchdog is running.')
        return

    if nestdb_status['value']['nestdb_is_running']:
        clear_data(last_remedy_time, started_by_init_timestamp)
        logger.debug('Not remedy, because nestdb is running.')
        return

    # started_by_init is None indicates that not receive nestdb_init notification.
    # This should be caused by SHA start later than nestdb or SHA restartaed.
    #
    # Though without valid value of started_by_init, should continue (not return)
    # when nestdb is up in precheck and down in current check.
    if started_by_init is None and not get_plugin_data('watchdog_is_running_before'):
        logger.debug('Not remedy, because watchdog of nestdb is not running before '
                    'and did not receive nestdb_init either.')
        return

    if nestdb_status['value']['disk_is_full']:
        # nestdb is down and disk is full now;
        # check whether nestdb is down before disk full, if yes, should not remedy; otherwise, set flag and wait disk
        # available then remedy.
        crash_before_disk_full = get_plugin_data('crash_before_disk_full')

        if crash_before_disk_full:
            logger.debug('Not set flag, because nestdb crashed before disk full.')
            return

        if not get_plugin_data('crash_by_disk_full'):
            set_plugin_data('crash_by_disk_full', True)
            logger.info('Not remedy because disk is full, only set the flag crash_by_disk_full '
                        'and wait to do remediation when disk change to not full')

        return

    if not get_plugin_data('crash_by_disk_full'):
        # nestdb is down and disk is not full now;
        # if not crash by disk full, then nestdb is down by other reason.
        if not get_plugin_data('crash_before_disk_full'):
            set_plugin_data('crash_before_disk_full', True)
            logger.info('Not remedy, because the nestdb down is not caused by disk full.')

        return

    logger.info('Nestdb remediation start.')

    try:
        res = Remedy(cmd='RESTART_NESTDB', plugin_id=data_store['plugin_id'])
        success = res.is_succeed
        err_msg = res.err_msg
    except Exception as e:
        success = False
        err_msg = str(e)

    last_remedy_time = time.monotonic()
    if success:
        clear_data(last_remedy_time, started_by_init_timestamp)
        logger.info('Nestdb remediation end - succeeded.')
        return

    tried_count = get_plugin_data('tried_count') + 1 if get_plugin_data('tried_count') is not None else 1
    max_try_count = profile['MAX_TRY_COUNT_FOR_A_CRASH']
    if tried_count == max_try_count:
        clear_data(last_remedy_time, started_by_init_timestamp)
        logger.info('Nestdb remediation end - failed with error:{}.'.format(err_msg))
        logger.info('Stop trying for remediation as tried count reached max limit:{}.'.format(max_try_count))
        return

    set_plugin_data('tried_count', tried_count)
    set_plugin_data('last_remedy_time', last_remedy_time)
    logger.info('Nestdb remediation end - failed with error:{}, '
                'will retry, current tried_count:{}.'.format(err_msg, tried_count))


nestdb_remediation()
